/*
 * File dbtools.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL),
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef DBTOOLS_H_
#define DBTOOLS_H_

/**
 * \file dbtools.h
 * \brief Header file for different modules. Generally covers
 * database routines.
 *
 * This header contains declarations of functions, structures
 * and constants related to TINE database file access.
 * This includes csv file access as well as address resolution
 * and caching routines.
 *
 * Functions found here are implemented in
 * <ol>
 * <li>dblib.c</li>
 * <li>srvdbase.c</li>
 * <li>others (not yet completely analyzed)</li>
 * </ol>
 *
 * \todo list all modules for which this is a header.
 */

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define EQPDB_FILE_TEMP "eqpdbase.tmp"
#define FECDB_FILE_TEMP "fecaddr.tmp"

#define FECXML_FILE "fec.xml"

#define DEVICES_MANIFEST "devmf.csv"
#define PROPERTIES_MANIFEST "prpmf.csv"

#ifndef MAX_NUM_USERS
# define MAX_NUM_USERS 100
#endif

#ifndef MAX_NUM_USERGROUPS
# define MAX_NUM_USERGROUPS 10
#endif

#ifndef MAX_NUM_IPNETS
# define MAX_NUM_IPNETS 100
#endif

/**
 * \internal
 * \brief A structure which contains information about a column in a CSV file.
 *
 * This structure is used internally by CSVDBStruct.
 */
typedef struct
{
  char *nam;          /**< column name */
  char *def;          /**< default value */
  int  pos;           /**< column index position (to be determined) */
  int  fmt;           /**< value format */
  int  siz;           /**< number of element of specified format */
  BYTE *ptr;          /**< pointer to container position (initially relative pointer) */
  int (*fcn)(char *); /**< optional dispatch function (called when value has been determined) */
  BYTE *start;        /**< pointer starting position in container  (initially 0) */
} CSVCOLUMN;

#ifndef _CSVCOLUMN
# define _CSVCOLUMN CSVCOLUMN
#endif

#define CSVDB_PREALLOC    0x01
#define CSVDB_INITIALIZED 0x02

/**
 * \internal
 * \brief Container for the contents of a CSV file.
 *
 * This structure has two purposes. It serves as
 * <ul>
 * <li>a container</li>
 * <li>its own single iterator</li>
 * </ul>
 *
 * \see CSVCOLUMN
 * \todo redevelopment of csv library desirable
 */
typedef struct CSVDBStruct
{
  _CSVCOLUMN *col;                   /**< pointer to CSVCOLUMN list */
  int ncols;                         /**< num columns */
  int siz;                           /**< size of db file excluding comments => num rows - header */
  void (*ini)(struct CSVDBStruct *); /**< optional initializer routine */
  int  (*itr)(struct CSVDBStruct *); /**< optional iterator */
  void (*exi)(struct CSVDBStruct *); /**< optional finalizer routine */
  int cnt;                           /**< counter */
  BYTE *tgt;                         /**< pointer to the container */
  int tgt_siz;                       /**< size of container (e.g. sizeof(some typedef) */
  int tgt_flgs;                      /**< container flags (e.g. CSVDB_PREALLOC) */
} CSVDB;

#define USERS_ALLOWED_SETNAME "users_allowed"
#define USERS_DENIED_SETNAME "users_denied"
#define GROUPS_ALLOWED_SETNAME "groups_allowd"
#define NETS_ALLOWED_SETNAME "nets_allowed"
#define NETS_DENIED_SETNAME "nets_denied"
#define NETS_NETWORK_SETNAME "nets_network"

typedef struct
{
	int (* processOpeningTag)( char * tag );
	int (* processClosingTag)( char * tag );
	int (* processData)( char * data );
	void * currentDataSection;
	int currentDataType;
  int currentDataCapacity;
	void * currentData;
} XML_DBHANDLER;

typedef struct listItem
{
  struct listItem * previous;
  struct listItem * next;
  void * data;
  int dataBufferAllocated;
} DBLNKLIST_ITEM;

typedef struct
{
  int nElements;
  DBLNKLIST_ITEM * first;
  DBLNKLIST_ITEM * current;
  DBLNKLIST_ITEM * last;
} DBLNKLIST;

TINE_EXPORT DBLNKLIST *XMLListCreate( void );
TINE_EXPORT void XMLListDelete( DBLNKLIST * list );
TINE_EXPORT void XMLListAdd( DBLNKLIST * list, void * data );
TINE_EXPORT void XMLListAddCopy( DBLNKLIST * list, void * data, int sizeOfData );
TINE_EXPORT void XMLListGoToFirst( DBLNKLIST * list );
TINE_EXPORT void XMLListGoToLast( DBLNKLIST * list );
TINE_EXPORT void XMLListRemoveCurrent( DBLNKLIST * list );
TINE_EXPORT void *XMLListGetNext( DBLNKLIST * list );
TINE_EXPORT void *XMLListGetPrevious( DBLNKLIST * list );
TINE_EXPORT int XMLListHasNext( DBLNKLIST * list );
TINE_EXPORT int XMLListHasPrevious( DBLNKLIST * list );
TINE_EXPORT void XMLSetCurrentHandler( XML_DBHANDLER * xmlHandler);
TINE_EXPORT void XMLRestoreHandler(void);
int dbpush( void * p );
int dbpop( void ** p );

typedef struct
{
  char name[SET_NAME_SIZE];
  DBLNKLIST *members;
} xmlSETIS;

typedef struct
{
  char almTag[32];
  char dataFormat[64];
  char almText[64];
  char devText[64];
  char datText[64];
  char urlText[64];
  UINT32 code;
  UINT32 mask;
  UINT32 severity;
  UINT32 system;
  UINT32 dataSize;
} xmlADS;

typedef struct
{
  UINT32 hstIndex;
  char hstDevName[DEVICE_NAME_SIZE];
  UINT32 hstHeartbeat;
  UINT32 hstPollRate;
  UINT32 hstArchRate;
  char hstTolerance[64];
  UINT32 hstDepthShort;
  char hstDepthLong[64];
  char hstRemoteServer[PROPERTY_REDIR_SIZE];
  char hstFilterString[256];
} xmlHIS;

typedef struct
{
  char almDevName[DEVICE_NAME_SIZE];
  char almValNormal[64];
  UINT32 almSevHigh;
  UINT32 almSevLow;
  UINT32 almSevHighWarn;
  UINT32 almSevLowWarn;
  UINT32 almSystem;
  UINT32 almValMask;
  UINT32 almCntThreshold;
  float  almValHigh;
  float  almValLow;
  float  almValHighWarn;
  float  almValLowWarn;
  int  code;
  int  codeHigh;
  int  codeHighWarn;
  int  codeLow;
  int  codeLowWarn;
  char tagNormal[ALARM_TAG_SIZE];
  char tagHigh[ALARM_TAG_SIZE];
  char tagHighWarn[ALARM_TAG_SIZE];
  char tagLow[ALARM_TAG_SIZE];
  char tagLowWarn[ALARM_TAG_SIZE];
  char almFilterString[256];
} xmlAWIS;

typedef struct
{
  char name[PROPERTY_NAME_SIZE];
  char description[128];
  char redirection[PROPERTY_REDIR_SIZE];
  char deviceSet[SET_NAME_SIZE];
  char access[ACCESS_STR_SIZE];
  char dataInType[64];
  char dataOutType[64];
  char xEgu[16];
  char yEgu[16];
  UINT32 dataInSize;
  UINT32 dataOutSize;
  SINT32 id;
  UINT32 deadband;
  float  xMax;
  float  xMin;
  float  yMax;
  float  yMin;
  DBLNKLIST *histories;
  DBLNKLIST *almwatches;
  DBLNKLIST *sets;
} xmlPRPIS;

typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char desc[DEVICE_DESC_SIZE];
  char loc[DEVICE_LOCATION_SIZE];
  char redirection[DEVICE_REDIR_SIZE];
  char propertySet[SET_NAME_SIZE];
  char region[TAG_NAME_SIZE];
  UINT32 number;
  UINT32 mask;
  UINT32 offline;
  float zpos;
  DBLNKLIST *sets;
} xmlDEVIS;

typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char target[DEVICE_NAME_SIZE];
} xmlAIS;

typedef struct
{
  char name[EQM_NAME_SIZE];
  char context[CONTEXT_NAME_SIZE];
  char serverName[EXPORT_NAME_SIZE];
  char serverGroup[EXPORT_NAME_SIZE];
  char serverGroupDevPrefix[XFIX_SIZE];
  char serverGroupDevPostfix[XFIX_SIZE];
  char subsystem[SUBSYSTEM_NAME_SIZE];
  char serverMaster[EXPORT_NAME_SIZE];
  char slaveMaster[EXPORT_NAME_SIZE];
  char region[TAG_NAME_SIZE];
  UINT32  serverGroupIndex;
  UINT32  deviceSpace;
  DBLNKLIST *devices;
  DBLNKLIST *properties;
  DBLNKLIST *sets;
  DBLNKLIST *alarmdefs;
} xmlXIS;

typedef struct
{
  char name[FEC_NAME_SIZE];
  char context[CONTEXT_NAME_SIZE];
  char description[FEC_DESC_SIZE];
  char hardware[FEC_HDW_SIZE];
  char location[FEC_LOCATION_SIZE];
  char subsystem[SUBSYSTEM_NAME_SIZE];
  char historyHome[HISTORY_HOME_SIZE];
  int  portOffset;
  int  fecMtu;
  int  fecBurstLimit;
  int  fecCycleDelay;
  int  fecWorkAreaSize;
  DBLNKLIST *eqms;
  DBLNKLIST *sets;
  DBLNKLIST *alias;
  ALMSETIS *almsettings;
} xmlFIDS;

extern XML_DBHANDLER DocumentHandler,FecHandler,EqmHandler,DeviceHandler,AlmSettingsHandler;
extern XML_DBHANDLER PropertyHandler,AlmDefHandler,NameSetHandler,MemberHandler,AliasHandler,AlmWatchHandler,HistoryHandler;
TINE_EXPORT int XMLReadFile(char *filepath, char *filename, DBLNKLIST *fecList);
extern DBLNKLIST *gFecInfoList;
extern int isUsingXMLFecDatabase;
extern int fecXmlDone;
int populateFIDS(DBLNKLIST *fecList,FIDS **fids);
int populateXIS(char *fecname,char *eqmname,DBLNKLIST *fecList,XIS **xis);
int populateDEVIS(char *fecname,char *eqmname,DBLNKLIST *fecList,DEVIS **devis);
int populatePRPDEVIS(char *fecname,char *eqmname,char *prpname,DBLNKLIST *fecList,DEVIS **devis);
int populateDEVPRPIS(char *fecname,char *eqmname,char *lstname,DBLNKLIST *fecList,NAME64 **prps);
int populateNameList(DBLNKLIST *fecList,char *fecName,char *eqmName,char *devName,char *prpName,char *listName,NAME64 **lst);
int populateAIS(char *fecname,DBLNKLIST *fecList,ALIAS_TABLE **ais);
void freeFecListMemory( DBLNKLIST *fecList );
extern CSVDB csvHstMfDb;

#if !defined(UNIX) && !defined(VXWORKS)
TINE_EXPORT char *strtok_r(char *s, const char *delim, char **ptrptr);
#endif
TINE_EXPORT char *strtrm(char *s);
TINE_EXPORT char *strntrm(char *s,int maxlen);
TINE_EXPORT int strtobool(const char *str);
TINE_EXPORT int csvAssignValue(void *ptr,int fmt,int siz,char *str);
int _csvAssignValue(void *ptr, int fmt, int siz, char *str, int internal);
TINE_EXPORT int csvGetValue(void *ptr, int fmt, int siz, char *str);
int _csvGetValue(void *ptr, int fmt, int siz, char *str, int internal);
TINE_EXPORT int csvReadFile(char *path,char *file,CSVDB *db,void **target);
TINE_EXPORT int csvWriteFile(char *path,char *file,CSVDB *db);
TINE_EXPORT void csvStdInit(struct CSVDBStruct *db);
TINE_EXPORT int  csvStdIterator(struct CSVDBStruct *db);
TINE_EXPORT void csvStdExit(struct CSVDBStruct *db);
TINE_EXPORT int GetCsvColumnIndex(const char *csvHeader,const char *columnTag);
TINE_EXPORT char *GetCsvColumnString(int col,const char *str);
TINE_EXPORT int GetCsvDataBaseSize(const char *filename);
TINE_EXPORT char *csvGetLine(char *s, int size, FILE *stream);
int prepIncomingAddress(int pktlen,char *addrstream,FecAddrStruct *fec,ExpDataStruct *srv);
int queryNameServerForFec(char *fecName,FecAddrStruct *fec);
int GetDataBaseSize(char *filename);
char *makeDbFileName(char *eqm,char *fname,int usesubdir);
int getFecIndex(char *fecName);
int locateExport(char *expName,char *context,char *device,int *rpcidx);
int asyncRelocateExport(ConTblEntry *c);
int GetAddressFromNameFile(char *filepath,char *context,char *expName, char *fecName,char *devName,int *rpcidx);
int GetAddressFromNameServer(char *context,char *expName, char *fecName,int *rpcidx);
int addNameServerToTables(void);
int appendAddressToCache(char *fecName,char *ip, short portOffset);
int appendServerToCache(char *fecName,char *context,char *expName, char *eqmName);
int addServerToGroupCacheFile(char *filepath,char *ctx,char *grp,char *srv);
int addDeviceToMemberCacheFile(char *filepath,char *ctx,char *srv,char *dev);
int queryNameServerForAddress(char *ctx,char *exp,FecAddrStruct *fec,ExpDataStruct *srv);
int csvnamcmp(char *name,char *str,int col,int limit);
int findcol(char *str,char *tag);
int getColumnStr(int col,char *str,char *buf,int bufsiz);
char *colptr(int col,char *str);
void BytesToHex(BYTE *b,char *s,int n);
void HexToBytes(char *s,BYTE *b,int nibs);
unsigned int HexToDec(char *str);
int GetNameServerAddress(void);
int getFecName(void);
int reserveFecAddressSpace(void);
int rmvRegUserLst(char *eqm,NAME16 *userlist,int listsize);
int getRegUsrsFromFile(char *eqm,char *dev,char *prp,AclInfoType *acl,char *fn,char *accstr,int *accflg);
int getAllowedUsers(char *eqm);
int getAllowedNets(const char *eqm);
int getDeniedNets(void);
void dumpRegisteredUsers(void);
void dumpRegisteredGroups(void);
void dumpRegisteredAdmins(void);
int getRegisteredExports(char *filename,char *eqm);
void setFailoverConfigInExportList(ExportListStruct *el, char *master, char *slaveMaster);
int outputConsumerList(void);
int outputContractList(void);
int outputEqmList(void);
int outputEqmProperties(char *eqm);
int outputEqmDevices(char *eqm);
int outputContractStats(int i);
int outputConnectionStats(int k);
int outputServerStats(void);
int outputServerTime(void);
int outputSettings(void);
int dumpSettings(void);
int dumpEnv(void);
int dumpSocketSettings(void);
int dumpBucketList(void);
int dumpContractStats(int i);
int dumpConnectionStats(int i);
int dumpConsumerStats(int i);
int dumpServerStats(void);
int dumpServerTime(void);
int dumpVersion(void);
int dumpBuildInfo(void);
void fixTineHome(void);
int isSystemContextDecoration(char *dec);
int SaveStockPropertyValues(ExportListStruct *el,char *prpName);
int RestoreStockPropertyValues(ExportListStruct *el,char *prpName);
TINE_EXPORT void isEquipmentNameServer(int value);
TINE_EXPORT void isGroupNameServer(int value);
TINE_EXPORT int GetHostFromFecName(char *fecName,char *hostname);
TINE_EXPORT int GetServerHostAddress(const char *context,const char *expName,char *host,char *addr,char *fec);
TINE_EXPORT void FlushFecAddrCache(void);
TINE_EXPORT void flushFecAddrCache(void);
TINE_EXPORT void flushRedirectedLinkList(void);
int removeRedirectedLink(RdrTblEntry *r,int cc);
void flushWcLst(void);
TINE_EXPORT int GetENSAddr(char *name,char *addr);
TINE_EXPORT int ToggleNameServer(void);
TINE_EXPORT int GetEquipmentModuleList(NAME8 *lst,int *lstSize);
TINE_EXPORT int wc_stricmp(const char *wcstr,const char *str);
TINE_EXPORT int IgnoreFecXmlDatabase(void);
int makeDirectoryPath(char *dname,int log);
TINE_EXPORT int MakeDirectoryPath(char *dname,int log);
int zerocmp(void *v,int cnt);

extern TINE_EXPORT int gScanForNameLists;
extern TINE_EXPORT int gScanForDeviceLists;
extern TINE_EXPORT int gIgnoreVolatileAddressCache;
extern TINE_EXPORT int gIgnoreAddressCacheOnUnknownAddress;
extern int gIsEquipmentNameServer;
extern int gIsGroupNameServer;
extern FILE *dbgfp;
#define DBGBUF_SIZE 256
extern char dbgbuf[DBGBUF_SIZE];
extern int dbaseloaded;
extern int MaxTblAllocationSize;
extern char tineHomePath[STD_PATH_LENGTH];
extern char fecDbPath[STD_PATH_LENGTH];
extern int FecNameRegistered;
extern int ExportsRegistered;
extern int gotNameServerAddress;
extern AclInfoType gDeniedUsersList;
extern int isNameServerRequest;
extern int NameServerLoaded;
extern int MaxNumAllowedUsers;
extern int MaxNumUserGroups;
extern int MaxNumAllowedIPNets;
extern int gAllowNetworkAddressResolution;
extern int gEnsAddrResolutionInProgess;
extern int fecAddrAcqId; /* async link id for addr re-acquisition */
void reAcqAddr(ConTblEntry *c);
void rmvRdrInfo(RdrTblEntry *r,int cc);
int addStandAloneEntryToAddressCache(char *ctxName,char *expName,char *eqmName,int *fecIdx,int *srvIdx);
int getProtocolFromAddressString(char *strAdr);
int getInetFamilyFromAddressString(char *strAdr);
int getFilesOfType(const char *path, const char *pattern, NAME64 *lst, int lstsiz);
FILE *getMfFile(char *eqm, char *mode, char *fileName);
#ifndef FS_RDONLY
extern char gAddrCacheFilePath[128];
int isAddressInCacheFile(char *filepath,int srvidx,int fecidx);
int addAddressToCacheFile(char *filepath,int srvidx,int fecidx);
void establishAddrCacheFilePath(void);
int fixDbFileName(char *ffn,char *eqm,char *file,int ismff);
void resetOutputBlocks(int fecidx);
int writeStructCacheFile(const char *key,const char *tag);
int readStructCacheFile(const char *key, const char *tag, DTYPE *dout);
int dumpDevicesManifest(char* eqm, int to_stdout);
int dumpPropertiesManifest(char* eqm, int to_stdout);
#endif
#ifndef SMALL_TINE_LIB
int populateADS(char *fecname,char *eqmname,DBLNKLIST *fecList,ADSIS **adsis);
int populateAWS(char *fecname,DBLNKLIST *fecList,AWSIS **awsis);
int populateAlmSettings(char *fecname,DBLNKLIST *fecList,ALMSETIS **asis);
int addNetworkEntryToAddressCache(int *fecIdx,int *srvIdx);
int GetGroupMembersFromFile(const char *filepath,const char *ctx,const char *srv,NAME64 **memLst,int *lstSize);
int grpFileHasDevice(const char *filepath,const char *ctx,const char *mem,const char *tgt);
#endif

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* DBTOOLS_H_ */
