/*
 * File errors.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**< \file errors.h
 *
 * system error code definitions
 *
 */
#ifndef ERRORS_H_
#define ERRORS_H_

#ifndef NO_DOXYGEN
#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

extern TINE_EXPORT char erlst[][32];
extern char logLevelList[][8];
extern TINE_EXPORT char *errLongMsg[];

TINE_EXPORT char *cc2str(int code);
TINE_EXPORT char *logLevel2str(int code);
TINE_EXPORT char *cc2longString(int code);
TINE_EXPORT int isTimeout(int code);
int isCoerciveStatus(int code);

#endif /* NO_DOXYGEN */

#define TINE_LOGLEVEL_INFO  0
#define TINE_LOGLEVEL_WARN  1
#define TINE_LOGLEVEL_ERROR 2
#define TINE_LOGLEVEL_FATAL 3

#define CE_REDIRECTED  0x8000
#define CE_SENDDATA    0x4000
#define CE_RENEGOTIATE 0x2000
#define CE_SENDCODE    0x1000

#define HAS_DATA(c) ((c) == 0 || (c) & CE_SENDDATA)

#define BASE_CC(c) ((c) & 0xfff)

/** \def Error Number Definitions */
enum TErrorList
{
  operation_success       = 0,  /**< successful completion */
  illegal_line            = 1,  /**< illegal line number */
  illegal_format          = 2,  /**< illegal format or data type */
  illegal_arithmetic      = 3,  /**< illegal arithmetic expression */
  ambiguous               = 4,  /**< ambigous input information */
  illegal_delimiter       = 5,  /**< illegal delimiter in expression */
  zero_divide             = 6,  /**< divide by zero */
  work_area_full          = 7,  /**< working area buffer is full */
  non_existent            = 8,  /**< requested target does not exist */
  invalid_transport       = 9,  /**< transport medium is not valid */
  data_size_mismatch      = 10, /**< incoming data payload larger than requested ? */

  no_data_in_range        = 11, /**< no data found in input range */
  un_allocated            = 12, /**< requested target is not allocated in the given context */
  no_such_line            = 13, /**< expected line not found in database */
  illegal_data_size       = 14, /**< give data size is illegal or not valid */
  io_error                = 15, /**< bus or network io error */
  illegal_context         = 16, /**< given context is illegal */
  runtime_error           = 17, /**< runtime error */
  system_error            = 18, /**< system internal error */
  hardware_busy           = 19, /**< hardware is busy */
  argument_list_error     = 20, /**< input argument or parameter is invalid (alias for invalid_parameter */
  invalid_parameter       = 20, /**< input argument or parameter is invalid */

  file_error              = 21, /**< file access error */
  use_stream_transport    = 22, /**< datagram message size exceeds the allowable mtu (reserved) */
  dimension_error         = 23, /**< array dimension error */
  square_root_negative    = 24, /**< square root of a negative number */
  buffer_too_small        = 25, /**< supplied buffer is too small to handle the request */
  string_too_long         = 26, /**< supplied string is too long to buffer and cannot be truncated */
  ipx_socket_error        = 27, /**< ipx socket error (reserved) */
  net_read_error          = 28, /**< network read error */
  not_ready               = 29, /**< call to equipment module dispatch is not ready (internal, reserved) */
  invalid_transport_size  = 30, /**< total message size exceeds the configured allowable message size */

  log_negative            = 31, /**< log of a negative or zero number */
  device_not_connected    = 32, /**< requested device is not connected */
  unauthorised_action     = 33, /**< requested operation is not authorized */
  hardware_error          = 34, /**< hardware error */
  illegal_equipment_number= 35, /**< requested device does not have a table entry (alias for illegal_device_number) */
  illegal_device_number   = 35, /**< requested device does not have a table entry */
  illegal_device          = 35, /**< requested device is illegal */
  illegal_property        = 36, /**< requested property is illegal */
  out_of_range            = 37, /**< given value is out of range */
  not_implemented         = 38, /**< requested operation has not been implemented */
  no_such_computer        = 39, /**< requested host computer does not exist */
  struct_sealed           = 40, /**< tagged structure has already been sealed (reserved) */

  syntax_error            = 41, /**< expression syntax error (unused) */
  no_such_file            = 42, /**< requested file does not exist */
  file_already_exists     = 43, /**< requested file already exists */
  no_file_space           = 44, /**< not enough capacity in pre-allocated random access file to complete the requested operation */
  link_not_open           = 45, /**< presistent link has timed out (alias for link_timeout) */
  link_timeout            = 45, /**< presistent link has timed out */
  remitted_data_lost      = 46, /**< returned only a portion of the requested data set (reserved) */
  end_of_file             = 47, /**< attempt to read past the end of a file (unused) */
  archive_busy            = 48, /**< archive operation is in progress */
  server_name_in_use      = 49, /**< device server name is already being used */
  no_such_column          = 50, /**< expected column not found in database */

  out_of_client_memory    = 51, /**< a client call could not allocate sufficient memory */
  mcast_access_required   = 52, /**< access requires CM_NETWORK which was not supplied! (contract coercion) */
  address_unresolved      = 53, /**< requested target address could not be resolved */
  invalid_property        = 54, /**< requested property is not properly formed */
  address_unknown         = 55, /**< requested target address could not be found */
  name_unknown            = 56, /**< requested name could not be found */
  invalid_keyword         = 57, /**< input globals keyword is invalid */
  invalid_link            = 58, /**< input link id is invalid */
  string_expected         = 59, /**< database file read encountered an end-of-line while expecting a string */       
  out_of_local_memory     = 60, /**< a local call to allocated memory was not successful */

  out_of_shared_memory    = 61, /**< a call to obtain a shared-memory buffer was not successful */
  invalid_structure_tag   = 62, /**< the given structure tag is not in the structure registry */
  invalid_index           = 63, /**< the index given does not reference a table entry */
  illegal_equipment_name  = 64, /**< the equipment name given is not valid */
  link_error              = 65, /**< internal error processing data link (reserved) */
  code_failure            = 66, /**< internal failure due to unhandled exceptions, unavailable resource, OS failure, etc. */
  non_existent_server     = 67, /**< requested server does not exist */
  function_deprecated     = 68, /**< requested action or functionality has been deprecated */
  not_supported           = 69, /**< requested action or functionality not supported */
  illegal_character       = 70, /**< error parsing input */
  parsing_error           = 70, /**< error parsing input */

  illegal_operator        = 71, /**< input arithmetic or logical operator is not allowed */
  not_allowed             = 72, /**< requested operation is not allowed */
  illegal_read_write      = 73, /**< requested access is not allowed (typically CA_WRITE) */
  out_of_server_memory    = 74, /**< a server call could not allocate sufficient memory */
  database_not_loaded     = 75, /**< requested operation requires a database which has not been loaded */
  illegal_command         = 76, /**< requested command is illegal */
  resources_exhausted     = 77, /**< requested operation would exceed the configured capacity of an allocation table (clients, contracts, connections, etc.) */
  header_not_found        = 78, /**< expected csv file does not have a header */
  sedac_error             = 79, /**< sedac bus error */
  reacquire_address       = 80, /**< signal to reacquire a server's address from ENS (reserved, internal) */

  semaphore_error         = 81, /**< a semaphore or mutex could not be obtained or released */
  driver_not_installed    = 82, /**< request requires a hardware driver which is not installed */
  port_not_available      = 83, /**< request requires access to a port whichis not available */
  async_access_required   = 84, /**< access to selected property should be asynchronous (contract coercion) */
  semaphore_busy          = 85, /**< a semahpre or mutex could not be obtained within the requested time */
  non_existent_elem       = 86, /**< the requested equipment modules does not exist in the registry */
  non_existent_fec        = 87, /**< the requested FEC does not exist */
  non_existent_client     = 88, /**< (unused) */
  cannot_lock             = 89, /**< not able to obtain an access lock or lock resources */
  not_running             = 90, /**< the requested resource is not running */

  not_posted              = 91, /**< a dispatch routine did not post a reply */
  not_accepted            = 92, /**< the requested operation was refused */
  operation_timeout       = 93, /**< the requested operation did not complete within the time limit specified */
  illegal_protocol        = 94, /**< the incoming tine protocol is not recognized (reserved, internal) */
  gpib_error              = 95, /**< gpib bus error */
  rs232_error             = 96, /**< rs232 bus error */
  operation_busy          = 97, /**< the requested operation is busy and needs more time to complete */
  connection_timeout      = 98, /**< a synchronous link did not return within the time limit specified */
  illegal_mode            = 99, /**< the requested access mode is not valid */
  not_owner              = 100, /**< caller is not the owner of the requested resource */

  not_defined            = 101, /**< the requested operation has not yet been defined */
  net_write_error        = 102, /**< unable to send data on the network */
  invalid_data           = 103, /**< the input data are not valid */
  software_error         = 104, /**< general software error */
  access_denied          = 105, /**< call was refused by the security subsystem */
  tcp_not_supported      = 106, /**< an attempt to use TCP/IP access was made and is not supported on this host (reserved) */
  ipx_not_supported	     = 107, /**< an attempt to use IPX access was made and is not supported on this host (reserved) */
  host_not_resolved	     = 108, /**< the requested host could not be resolved */
  tcp_connect_error	     = 109, /**< unable to connect to a TCP/IP socket (reserved) */
  tcp_socket_error 	     = 110, /**< general TCP/IP socket error (reserved) */

  configuration_error    = 111, /**< resource configuration error (e.g. redirection information) */
  invalid_return_code    = 112, /**< indicates a systematic error in processing the equipment dispatch handler (reserved) */
  link_blacklisted 	     = 113, /**< requested link parameters cannot succeed (due to initial return of non_existent_elem, address_unknown, etc.) */
  link_exists      	     = 114, /**< requested link already exists in the link table indicating a dependent link (reserved,internal) */
  max_alarms_exceeded    = 115, /**< attempt to add an alarm to the local alarm system whose capacity has been reached */
  not_exported           = 116, /**< (unused) */
  is_switching           = 117, /**< switching transition is in progress (unused) */
  at_limit               = 118, /**< setting/value is at a maximum or minimum or range exceeds storage (local history) */
  get_subscription_id    = 119, /**< used by access mode CM_NETWORK (reserved, internal) */
  renegotiate_contract   = 120, /**< used to repackage output data when few bytes are returned than requested (reserved, internal) */

  server_redirection     = 121, /**< used to supply redirection information (reserved, internal) */
  value_too_high         = 122, /**< returned value is too high */
  value_too_low          = 123, /**< returned value is too low  */
  warn_too_high          = 124, /**< returned value is approaching too high  */
  warn_too_low           = 125, /**< returned value is approaching too low  */
  completion_waiting     = 126, /**< dependent link is waiting for parent to complete */
  cannot_select          = 127, /**< select() operation failed (reserved) */
  has_query_function     = 128, /**< query call to PROPERTIES or DEVICES has completed but indicated a hierachical preference (reserved) */
  not_signalled          = 129, /**< contract is waiting to call the equipment module dispatch routine (reserved, internal) */
  call_redirection       = 130, /**< nominal alias for server_redirection */

  udp_socket_error 	     = 131, /**< general UDP socket error (reserved) */
  mutex_error            = 132, /**< a mutex could not be obtained or released */
  data_not_local         = 133, /**< indicates that a device within a wildcard call needs to span multiple servers (reserved, internal) */
  name_already_exists    = 134, /**< a structure or bitfield field name already exists */
  already_assigned       = 135, /**< the requested resource has already been assigned */
  invalid_field          = 136, /**< a structure or bitfield field is invalid */
  mcast_init_error       = 137, /**< could not initialize a multicast socket (reserved) */
  invalid_mcast_group    = 138, /**< multicast group address is invalid */
  non_existent_function  = 139, /**< an equipment module's dispatch routine is not registered */
  property_is_mca        = 140, /**< requested property is a multi-channel array  (contract coercion) */

  invalid_name           = 141, /**< input name is invalid */
  invalid_value          = 142, /**< input value is invalid */
  device_error           = 143, /**< requested device is in an error state */
  device_not_ready       = 144, /**< requested device is not ready */
  device_busy            = 145, /**< requested device is busy */
  invalid_interval       = 146, /**< requested polling interval is invalid (contract coercion) */
  notification_pending   = 147, /**< a link is pending notification (reserved, internal) */
  thread_error           = 148, /**< a thread operation failed */
  is_terminating         = 149, /**< the device server is terminating (reserved, internal) */
  name_server_unknown    = 150, /**< the equipment name server address is unknown (reserved, internal) */

  lock_required          = 151, /**< an access lock is required to process the operation */
  not_initialized        = 152, /**< server is not yet initialized */
  invalid_structure_size = 153, /**< the input structure size does not match the size in the structure registry */
  canbus_error           = 154, /**< can bus error */
  profibus_error         = 155, /**< profibus error */
  data_stale             = 156, /**< contract's data are stale (reserved) */
  has_bitfield_tag       = 157, /**< reqested contract needs to request a bitfield (reserved, internal) */
  not_locked             = 158, /**< the requested resource does not have an access lock (reserved) */
  lock_expired           = 159, /**< the requested access lock has expired (reserved) */

  not_registered         = 160, /**< the requested property (e.g. SystemScheduleProperty()) has not been registered */
  non_existent_property  = 161, /**< the requested property does not exist */
  memory_overwrite       = 162, /**< a heap error within a dispatch routine has been detected */
  server_idle            = 163, /**< the server is in an idle state */
  not_applicable         = 164, /**< the requested opertion is not applicable in the current context */
  access_locked          = 165, /**< access temporarily suspended due to access lock */
  invalid_reference      = 166, /**< invalid data reference handle */
  has_structure_tag      = 167, /**< reqested contract needs to request a structure (reserved, internal) */
  warn_disk_space        = 168, /**< warn that disk space is low (local alarm system) */
  low_disk_space         = 169, /**< signal that disk space is very low (local alarm system)  */

  information_static     = 170, /**< signal that static information is being 'polled'  (contract coercion) */
  reset_mca_property     = 171, /**< signal to re-establish all mca single element links */
  record_too_long        = 172, /**< record length is too long to process within kernel */
  unix_socket_error      = 173, /**< an error condition on a unix (pipe) socket */
  pipe_error             = 173, /**< an error condition on a unix (pipe) socket */
  shm_error              = 174, /**< an error occured accessing shared memory */
  tcp_connection_closed  = 175, /**< tcp peer has closed the connection */
  service_unavailable    = 176, /**< requested service is unavailable */
  device_offline         = 177, /**< requested device is offline */
  out_of_sequence        = 178, /**< data value is out of sequence */
  invalid_datarequest    = 179, /**< requested data length or type is invalid (contract coercion) */

  property_is_mca_elevated = 180, /**< requested property is a multi-channel array with elevated format */
  name_not_found         = 181, /**< named resource was not found in database */
  out_of_tolerance       = 182, /**< data out of tolerance */
  file_not_open          = 183, /**< read/write on file which is not open (re-introduce as some system servers were using it) */
  low_main_memory        = 184, /**< low available main memory detected */
  tcp_access_required    = 185, /**< access requires CM_CONNECT which was not supplied! (contract coercion) */
};

extern int numErr;

typedef struct ErrTableItem
{
  int code;
  char msg[32];
  char msgLong[128];
  struct ErrTableItem *nxt;
} ErrTblItem;
extern ErrTblItem *gRegisteredErrors;
extern int numRegisteredErrors;

typedef struct HashRefXRef
{
  ErrTblItem *ref;
  struct HashRefXRef *next;
} HashRefXRef;
#ifndef ERRORTBL_HASH_SIZE
# define ERRORTBL_HASH_SIZE 103
#endif
#define ERRORTBL_HASH_PAYOFF (ERRORTBL_HASH_SIZE/10)
extern HashRefXRef **RegisteredErrorsXRef;

/**
 * \brief Registers the error code given and associates with the given message strings
 *
 * \param code is the user-specified error code (512 or above) which can be returned from
 * an equipment module or property handler.
 * \param msgShort is a short message string (32 characters or less) supplying a brief 
 * description of the error 
 * \param msgLong is a long message string (128 characters or less) supplying a more
 * descriptive explanation of the error.
 *
 * \return 0 upon success or a TINE error code (e.g. invalid_return_code) upon error.
 */
TINE_EXPORT int RegisterErrorCode(int code, char *msgShort, char *msgLong);
void freeErrorTable();
ErrTblItem *getErrTblItem(int code);
/**
 * \brief Returns the error text associated with an error code
 *
 * \param code is error code (system code or registered code) for which a descriptive message
 * is desired.
 * \param errstr is a string buffer to hold the error message.
 * \param length is the length of the buffer supplied as \em errstr
 *
 * \return the string buffer supplied as \em errstr (for convenience)
 */
TINE_EXPORT char *GetErrorText(int code, char *errstr, int length);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* ERRORS_H_ */
