/*
 * File eztine.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file eztine.h
 * \brief TINE EZ client documentation
 *
 */
#ifndef EZTINE_H_
#define EZTINE_H_

#include "listener.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifdef EZTINE_EXPORTS
# define EXPORT_EZ __declspec(dllexport) 
#endif 
#ifdef EZTINE_IMPORTS
# define EXPORT_EZ __declspec(dllimport) 
#endif
#ifndef EXPORT_EZ
# define EXPORT_EZ 
#endif

#define EZTBL_HASH_SIZE 997
#define EZTBL_KEY_SIZE 320

/**
 * \struct ezResult
 * \brief Defines the 'Result' Object for the EZ TINE interface.
 * 
 * A structure instance is part of a forward/backward linked list.
 * Athough all fields are visible (this is C), the relevant fields
 * to the user are the 'status' and 'dout' fields.
 *
 * The return object can be passed as input to several EZ TINE
 * interface routines.
 */
typedef struct ezResult
{
	char key[EZTBL_KEY_SIZE]; /**< hash list key */
	int idListener;           /**< the table id of a listener (if applicable) */
	int idMonitor;            /**< the supplied monitor notifier id (if applicable) */ 
  int status;               /**< the current status */
  int scratchCapacity;      /**< the capacity in bytes of the scratch area */
  int lengthUpdate;         /**< the array length of the last update */
  time_t timeUpdate;        /**< the time (UTC) of the last update */
  DTYPE dout;               /**< the data output object (volatile) */
  DTYPE din;                /**< the data input object (fixed) */
	NAME64 *devLst;           /**< the channel list (if MCA) */
  BYTE *scratch;            /**< the scratch buffer (used in serveral array returns) */
  BYTE *scratchNames;       /**< the scratch buffer (used in channel names return) */
  struct ezResult *prv;     /**< previous linked list item */
  struct ezResult *nxt;     /**< next linked list item */
} ezResult;

/**
 * \struct ezDblArray
 * \brief Defines the double array type returned by ezGetDblValues()
 * 
 */
typedef struct
{
  double *values;   /**< pointer to an array of doubles (overloading the 'scratch' buffer) */
  int length;       /**< length of the array */
  int status;       /**< call status */
  time_t timeValid; /**< last time (UTC) the values were known to be valid */
} ezDblArray;

/**
 * \struct ezIntArray
 * \brief Defines the double array type returned by ezGetIntValues()
 * 
 */
typedef struct
{
  int *values;      /**< pointer to an array of ints (overloading the 'scratch' buffer) */
  int length;       /**< length of the array */
  int status;       /**< call status */
  time_t timeValid; /**< last time (UTC) the values were known to be valid */
} ezIntArray;

/**
 * \struct ezStrArray
 * \brief Defines the double array type returned by ezGetStrValues() and ezGetChannels()
 * 
 */
typedef struct
{
  char **values;    /**< pointer to an array of strings (overloading the 'scratch' buffer) */
  int length;       /**< length of the array */
  int status;       /**< call status */
  time_t timeValid; /**< last time (UTC) the values were known to be valid */
} ezStrArray;

/**
 * \brief Frees an EZ result object
 * 
 * Calls to ezGet() or ezSet() will create an EZ Result object (if one does not already exist).
 * The result object will reserve memory and occupy a spot in a hash table until it is explicity
 * freed via this call or until the application program ends.
 * A result object is specifically tied to the input parameter(s) 
 * provided in the call to ezGet() or ezSet() and is therefore created only once for a given set
 * of parameters.  
 *  
 * \param res is EZ Result object to be freed.
 *
 * \return 0 upon success or a TINE error code.
 */
EXPORT_EZ int ezFreeResult(ezResult *res);
/**
 * \brief Get the channel names associated with an EZ result object
 * 
 * If the result object is associated with a known multi-channel array, then the channel
 * names are returned.  Otherwise the 'device' name used to establish the result object
 * is returned;
 *  
 * \param res is EZ Result object for which the channel names should be returned.
 *
 * \note The array returned within the ezStrArray object is 'volatile' in the sense
 * that it points to a 'scratch' buffer within the result object.  This array should 
 * consequently be either used or copied prior to any futher call to ezGetDblValues(),
 * ezGetIntValues(), ezGetStrValues(), or ezGetChannels(), which also make use of the
 * scratch buffer.
 *
 * \return an ezStrArray containing the channel names
 */
EXPORT_EZ ezStrArray ezGetChannels(ezResult *res);
/**
 * \brief Gets the resulting data and status according the target parameter(s) given.
 * 
 * The returned 'result' object contains the data returned from the call according to
 * the target's default registered information (concerning data lengths and types).
 * 
 * \param fullNameAndProperty is a single string containing the fully specified
 * context, server name, device name, and property of the the desired target. This 
 * can be either of the form '/context/server/device/property' (i.e. 'property'
 * included in the hierarchy) or '/context/server/device[property]') (i.e. 'property'
 * distinguished via '[]').  The latter is occasionally easier to view and parse when
 * the 'device' name itself contains '/' characters (so as to extend the hierarchy).
 * \param input can be optionally provided in analogy with the printf() family of C
 * functions.  If input is provided, the 'fullNameAndProperty' text should be 
 * extended to include an input'pipe' character '<' followed by '%d' or '%g' to 
 * indicate either integer or float values as further input.  A '%D' can also be
 * supplied, indicating that a pointer to a tine DTYPE object will be input (which
 * can then provide any complex data input required).  
 *
 * \note Most 'get' calls do not require input, so this function can be used in its
 * most simple form in the vast majority of cases.  Those 'get' calls which do take
 * input typically require one or two integer or float values, making simple 
 * 'printf' style input a viable option.
 *
 * \b example:
 *
 * \include eg_ezGet.c
 *
 * \return an ezResult object containing the results
 *
 * \see ezGetDblValues(), ezGetIntValues(), ezGetStrValues()
 */
EXPORT_EZ ezResult *ezGet(char *fullNameAndProperty,...);
/**
 * \brief Sets the input data on the target given. 
 * 
 * The returned 'result' object contains any data returned from the call according to
 * the target's default registered information (concerning data lengths and types).
 * Thus this call is also a 'set/get' operation if the target support this.  The
 * user is free to ignore any returned data if the purpose of the operation was only
 * a 'set'.  This call will apply 'WRITE' access to the 'set' operation (as opposed
 * to a call to ezGet() with the same parameters as ezSet().
 * 
 * \param fullNameAndProperty is a single string containing the fully specified
 * context, server name, device name, and property of the the desired target. This 
 * can be either of the form '/context/server/device/property' (i.e. 'property'
 * included in the hierarchy) or '/context/server/device[property]') (i.e. 'property'
 * distinguished via '[]').  The latter is occasionally easier to view and parse when
 * the 'device' name itself contains '/' characters (so as to extend the hierarchy).
 * \param input can be optionally provided in analogy with the printf() family of C
 * functions.  If input is provided, the 'fullNameAndProperty' text should be 
 * extended to include an input'pipe' character '<' followed by '%d' or '%g' to 
 * indicate either integer or float values as further input.  A '%D' can also be
 * supplied, indicating that a pointer to a tine DTYPE object will be input (which
 * can then provide any complex data input required).  
 *
 *
 * \b example:
 *
 * \include eg_ezSet.c
 *
 * \return an ezResult object containing the results
 */
EXPORT_EZ ezResult *ezSet(char *fullNameAndProperty,...);

/**
 * \brief Returns an ezDblArray according to the EZ result object passed.
 * 
 * Calls to ezGet() or ezSet() will create an EZ Result object (if one does not already exist).
 * The result object contains the returned data (in the 'dout' field) which may or may not
 * refer to a double array.  The 'dout' object can be examined to determine the data type 
 * returned if necessary. Otherwise a call to ezGetDblValues() will attempt to convert the
 * returned data to a double array.  If this is not possible then an empty array is returned and
 * the 'status' field of the result object passed is set to indicate the problem as well as the 
 * 'status' field of the ezDblArray.
 *
 * \note compound data types such as 'DBLDBL' or 'USTRING', etc. which do contain at least one
 * data field which can be converted to a double result in the first (or primary) number field
 * being used to establish the double array.  
 * In such cases, the ezDblArray returned contains an array of doubles but with the 'status'
 * field set to 'ambiguous'.
 *  
 * \param res is EZ Result object whose data should be used to form an ezDblArray.
 *
 * \note The array returned within the ezDblArray object is 'volatile' in the sense
 * that it points to a 'scratch' buffer within the result object.  This array should 
 * consequently be either used or copied prior to any futher call to ezGetDblValues(),
 * ezGetIntValues(), or ezGetStrValues(), which also make use of the
 * scratch buffer.
 * On the other hand, repeated calls to ezGetDblValues() using the same result object will 
 * only overwrite the contents of the existing double array, which might be the desired 
 * result.
 *
 * \return ezDblArray (which could be zero length if there is a problem)
 *
 * \see ezGetIntValues(), ezGetStrValues(), ezGetChannels()
 */
EXPORT_EZ ezDblArray ezGetDblValues(ezResult *res);
/**
 * \brief Returns an ezIntArray according to the EZ result object passed.
 * 
 * Calls to ezGet() or ezSet() will create an EZ Result object (if one does not already exist).
 * The result object contains the returned data (in the 'dout' field) which may or may not
 * refer to an integer array.  The 'dout' object can be examined to determine the data type 
 * returned if necessary. Otherwise a call to ezGetIntValues() will attempt to convert the
 * returned data to an integer array.  If this is not possible then an empty array is returned and
 * the 'status' field of the result object passed is set to indicate the problem as well as the 
 * 'status' field of the ezIntArray.
 *
 * \note compound data types such as 'DBLDBL' or 'USTRING', etc. which do contain at least one
 * data field which can be converted to an integer result in the first (or primary) number field
 * being used to establish the double array.
 * In such cases, the ezIntArray returned contains an array of doubles but with the 'status'
 * field set to 'ambiguous'.
 *  
 * \param res is EZ Result object whose data should be used to form an ezIntArray.
 *
 * \note The array returned within the ezIntArray object is 'volatile' in the sense
 * that it points to a 'scratch' buffer within the result object.  This array should 
 * consequently be either used or copied prior to any futher call to ezGetDblValues(),
 * ezGetIntValues(), or ezGetStrValues(), which also make use of the
 * scratch buffer.
 * On the other hand, repeated calls to ezGetIntValues() using the same result object will 
 * only overwrite the contents of the existing double array, which might be the desired 
 * result.
 *
 * \return ezIntArray (which could be zero length if there is a problem)
 *
 * \see ezGetDblValues(), ezGetStrValues(), ezGetChannels()
 */
EXPORT_EZ ezIntArray ezGetIntValues(ezResult *res);
/**
 * \brief Returns an ezStrArray according to the EZ result object passed.
 * 
 * Calls to ezGet() or ezSet() will create an EZ Result object (if one does not already exist).
 * The result object contains the returned data (in the 'dout' field) which may or may not
 * refer to a string-type array.  The 'dout' object can be examined to determine the data type 
 * returned if necessary. Otherwise a call to ezGetStrValues() will attempt to convert the
 * returned data to a string array.  If this is not possible then an empty array is returned and
 * the 'status' field of the result object passed is set to indicate the problem as well as the 
 * 'status' field of the ezStrArray.
 *
 * \param res is EZ Result object whose data should be used to form an ezStrArray.
 *
 * \note The array returned within the ezStrArray object is 'volatile' in the sense
 * that it points to a 'scratch' buffer within the result object.  This array should 
 * consequently be either used or copied prior to any futher call to ezGetDblValues(),
 * ezGetIntValues(), or ezGetStrValues(), which also make use of the
 * scratch buffer.
 * On the other hand, repeated calls to ezGetStrValues() using the same result object will 
 * only overwrite the contents of the existing double array, which might be the desired 
 * result.
 *
 * \return ezStrArray (which could be zero length if there is a problem)
 *
 * \see ezGetIntValues(), ezGetDblValues(), ezGetChannels()
 */
EXPORT_EZ ezStrArray ezGetStrValues(ezResult *res);
/**
 * \brief Returns a character string containing the returned data values.
 * 
 * Calls to ezGet() or ezSet() will create an EZ Result object (if one does not already exist).
 * This routine returns a string containing the returned data (one element per line of output).
 *
 * \param res is EZ Result object whose data should be dumped to a string.
 *
 * \return a string representation of the data.
 */
EXPORT_EZ char *ezToString(ezResult *res);
/**
 * \brief attaches a monitor 'notifier' function to the given result object
 * 
 * Calls to ezGet() or ezSet() will create an EZ Result object (if one does not already exist).
 * This routine will attach a notifier funtion of the caller's chosing to the internal
 * data listener attached to the result object.  The listener will call the supplied
 * notifier function when new data have arrived, which could server a signal to make a fresh
 * call to ezGet() (as opposed to polling from within the application).
 *
 * \param res is EZ Result object to which the notifier should be attached.
 * \param nf is the caller supplied notifier function, which must have the prototype
 * 'void function(int id)'.
 * \param nid is the caller supplied notification 'id' which will be returned in the
 * call to the monitor notifier.
 *
 *
 * \b example:
 *
 * \include eg_ezAddMonitor.c
 *
 * \return 0 upon success otherwise a TINE error code.
 */
EXPORT_EZ int ezAddMonitor(ezResult *res,void (*nf)(int), int nid);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif


#endif /* EZTINE_H_ */
