/*
 * File hstlib.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _HSTLIB_H_
#define _HSTLIB_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "prolog.h"
#include "srvcore.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MAX_HISTORY_RECORDS
# define MAX_HISTORY_RECORDS 100
#endif
#define HISTORY_RECORD_CAPACITY_ADD 100
#define POI_RANGE_FACTOR 0.10
#define POI_TOLERANCE_FACTOR 10.0
#define USE_POI_RANGE     0x0001
#define USE_POI_TOLERANCE 0x0002
#define USE_POI_NONE      0x0004

#define DATATIME_NOW (time(NULL)-1+(time_t)gDataTimeStampOffset)

#define LHS_APPLY_ARCHIVE_NONE 0
#define LHS_APPLY_ARCHIVE_THIS 1
#define LHS_APPLY_ARCHIVE_TRANSITION 2  

/**
 * structure used in local histories of variable length data types
 */
typedef struct
{
  int capacity; /**< capacity of data block */
  int length;   /**< length of data in data block */
  BYTE *data;   /**< pointer to single entry data block */
} DataPoolEntry;

/**
 * History Table Entry structure 
 */
typedef struct
{
  CONTRACT c;       /**< the entry contract */
  int  pollingRate; /**< polling interval in msec */
  int  archiveRate; /**< archive interval in msec (sampling interval if LTS depth == 0) */
  int  depthShort;  /**< for short term storage */
  int  depthLong;   /**< for long term storage (months) */
  int  depthDays;   /**< for long term storage (days) */
  int  dataPtr;     /**< current valid data pointer */
  int  heartbeat;   /**< archive heartbeat in sec */
  float pTolerance; /**< percent tolerance */
  float aTolerance; /**< absolute tolerance */
  float range;      /**< value range from property settings */
  int  recordIndex; /**< local index identifier */
  int  recordLength;/**< (primary) record length in bytes */
  int  recordInitialized; /**< initialization flag */
  int  isScheduled; /**< schedule flag */
  int  keepPointsOfInterest; /**< poi flag */
  int  useStandardSet; /**< use std. pre-allocated files */
  int  hasSavedFiles; /**< long-term archive files found in the 'saved' area */
  int  storeSummary; /**< store data summary if TRUE */
  int  hasDataPool; /**< for variable length formats */
  int  poolEntryCapacity; /**< for variable length formats */
  int  recordCapacity; /**< record capacity for standard files (full byte capacity including header and prefix) */
  int  lastAccessSts; /**< return code of last eqm access */
  int  lastArchiveSts; /**< return code of last archive write */
  double lastAccessTime; /**< timestamp of last eqm access */
  double lastArchiveTime; /**< timestamp of last archive */
  char archiveFileName[32]; /**< current working file name */
  char rmvFileName[32]; /**< file for standard rotation removal */
  double *timestamp; /**< timestamp array */
  SINT32 *sysstamp; /**< sysstamp array */
  SINT32 *usrstamp; /**< usrstamp array */
  UINT16 *memidx;   /**< members index list if necessary */
  PrpRedirBlk *prd; /**< re-director to CHS if not NULL */
  BYTE *data;       /**< pointer to data block */
# ifndef SMALL_TINE_LIB
  FilterLink *fltr; /**< assigned filter */
# endif
} HstTblEntry;

typedef struct HistRecInputStruct
{
  CONTRACT c;
  int  pollingRate; /**< polling rate in msec */
  int  archiveRate; /**< archive rate in msec */
  int  depthShort;  /**< for short term storage */
  int  heartbeat;   /**< archive heartbeat in sec */
  int  recordIndex; /**< local index identifier */
  int  poolEntryCapacity; /**< for variable length formats */
  char exportTarget[EXPORT_NAME_SIZE];
  char redirection[PROPERTY_REDIR_SIZE];/**< re-director to CHS if not NULL */
  char fmt[64];     /**< format descriptor */
  char depthLong[64]; /**< for long term storage (parsed as mm.dd) */
  char tolerance[64]; /**< parseable tolerance string */
# ifndef SMALL_TINE_LIB
  char filterString[256]; /* supplied filter */
# endif
} HRIS;

/**
 * \struct HRS
 * \brief history record structure used in stock properties
 *
 */
typedef struct
{
  char property[PROPERTY_NAME_SIZE]; /**< property name of the record */
  char device[DEVICE_NAME_SIZE]; /**< device name of the record */
  int  datasiz;     /**< call size of the record */
  int  datafmt;     /**< calling format of the record */
  int  pollingRate; /**< polling rate in msec */
  int  archiveRate; /**< archive rate in msec */
  int  depthShort;  /**< for short term storage */
  int  depthLong;   /**< for long term storage */
  int  heartbeat;   /**< archive heartbeat in sec */
  float pTolerance; /**< percent tolerance */
  float aTolerance; /**< absolute tolerance */
} HRS; 
#define HRS_SIZE (64*2+9*4)

#define HRS_DEFAULT_ARCHIVE_RATE 1000
#define HRS_DEFAULT_ABS_TOLERANCE 0
#define HRS_DEFAULT_DEPTHLONG 1
#define HRS_DEFAULT_DEPTHSHORT 600
#define HRS_DEFAULT_HEARTBEAT 900
#define HRS_DEFAULT_POLL_RATE 1000
#define HRS_DEFAULT_REL_TOLERANCE 10

#define HRS_FREE_RECORD_INDEX_OFFSET 100
#define HRS_REDIRECTION_FLAG -1

#define MAKELONGDEPTH(m,d) ((m) < 0 ? -1 : (((d)&0x1f)<<16) + ((m)&0xffff))
#define HSTLONGMONTHS(t) ((t) < 0 ? -1 : (t) & 0xffff)
#define HSTLONGDAYS(t) ((t) < 0 ? -1 : ((t)>>16) & 0x1f)

/**
 * \struct HistorySpecification
 * \brief Defines a TINE History Specification object
 *
 * \note the pTolerance (relative or percent) field
 * below is the multiplication factor applied to the last 
 * archived data and is so incorprated into the tolerance.
 * As such it should be a value between 0 and 1
 * (i.e. '.1' is 10 %).
 * Values greater than 1 are automatically divided by 100 %
 * (the assumption being the the caller was thinking in units
 * of '%') and the ensuing fractional number will be seen in
 * any queries and cache files.
 */
typedef struct
{
  int  pollingRate;      /**< polling rate in msec */
  int  archiveRate;      /**< archive rate in msec */
  int  depthShort;       /**< for short term storage */
  int  depthLong;        /**< for long term storage */
  int  heartbeat;        /**< archive heartbeat in sec */
  float pTolerance;      /**< relative (percent) tolerance */
  float aTolerance;      /**< absolute tolerance */
  char *rhsServerName;   /**< Remote Server Name */
  char *rhsPropertyName; /**< Remote Property Name */
} HistorySpecification;

extern HistorySpecification defaultHSpec;

typedef struct
{
  double timeStamp;
  SINT32 sysStamp;
  SINT32 usrStamp;
} HstHdr;

#ifdef VXWORKS
# define S_USERWRITE (0644)
#endif
#define HISTORY_FILENAME "history.csv"
#define HISTORY_MANIFEST "hstmf.csv"
#define HISTORYHOME_MANIFEST "historymf.csv"

#define HISTORY_BUFFER_SIZE 1024
#define STS_DEADBAND 1

extern double gHistTimeStamp;
extern int gHistSysStamp;
extern int gHistUsrStamp;
extern int hstdbg;
extern int gHstMfDumped;
extern char arcDbPath[HISTORY_HOME_SIZE];
extern HstTblEntry *hstTbl;
extern int nHistoryRecords;
extern TINE_EXPORT int minHistoryPollingRate;
extern TINE_EXPORT int maxHistoryRecords;
extern int gScavangerThreadBusy;
extern int gIsLclHstLateAddition;
extern int gHstDeactivateOnError;
extern int gHistoryCycleNeedsScheduler;
extern int history_home_done;

int scavengeOlderLTSFiles(unsigned long minDiskSpaceInBlocks,time_t currentFileTime);
void validateHistoryHomePath(void);
void getHistoryFilterString(FilterLink *fltr, char * filterString);
int makeFilesForRecord(int index);
int getHstTblEntryFromRecordIndex(int idx);
int isInHistoryTable(HstTblEntry *hst,int nentries);
int getHistoryTableIndex(HstTblEntry *hst,int nentries);
char *SystemGetLastErrorString(void);
int addHistoryItem(int i, char *eqm,char *prp,char *dev,char *tag,int len,int fmt,int idx,
                   HistorySpecification *hspec,int hstFlags);
void freeHstMemory(void);
int hasHstFileAlarm(void);

int prepIncomingHstFilter(BYTE *filterstream, FilterStruct *flst);
int prepOutgoingHstFilter(BYTE *filterstream, int length);

TINE_EXPORT void SetHistoryDeactivateOnError(int value);
TINE_EXPORT int GetHistoryDeactivateOnError(void);

TINE_EXPORT void SetExcludePointsOfInterest(int value);
TINE_EXPORT int GetExcludePointsOfInterest(void);

/* useMinimalStorage */
/**
 * \brief Instructs the local history server to use a simple UTC timestamp 
 * (4-byte integer) when saving data to disk.
 *
 * By default, the local history subsystem will timestamp the data records stored to
 * disk with a higher resolution timestamp than the traditional UTC variable. The
 * data system stamps will likewise be stored with the data record.
 * This results in more disk space being used by an archive record, as the
 * record prefix is now 16 bytes per archived record instead of 4.  This can
 * be controlled by setting this variable to TRUE (small prefix) or FALSE (large prefix).
 *
 * \b Default: FALSE
 */
extern TINE_EXPORT int useMinimalStorage;
/* useMonthlyHistoryFiles */
/**
 * \brief Instructs the local history server to keep monthly archive files instead of daily
 * files.
 *
 * The local history subsystem maintains one file per record per day.  In some cases it 
 * might be advantageous to keep one file per record per month, for instance when many
 * small records are being maintained.  You can influence this behavior be setting this
 * flag to TRUE.
 */
extern TINE_EXPORT int useMonthlyHistoryFiles;

/* AppendHistoryInformation() */
/**
 * \brief Inserts a local history element into the local history server
 * 
 * A server can instruct the local history server to keep a history of the given
 * property by utilizing this call.  The local history server will periodically call
 * the property as specified according to the following input parameters.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param len is the length of the local history call.
 * \param fmt is the TINE format of the local history call.
 * \param idx is the local history index to be identified with this local history element 
 *  (Note: this must be unique with this server process and within the history data repository).
 * \param hspec is a pointer to a HistorySpecification structure, where the relevant filtering
 *  criteria are given.
 *  
 * \note In lieu of calling AppendHistoryInformation(), it is recommended to make use of the
 * startup configuration file 'history.csv', in which all local history critieria can be 
 * introduced.  This has the advantage that no information is hard-coded.  Indeed by introducing
 * a 'history.csv' file one can immediately instruct a server to maintain local history 
 * information (even if the server is developed and maintained by someone else!).
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_AppendHistoryInformation.c
 */
TINE_EXPORT int AppendHistoryInformation(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec);

/* AppendHistoryInformationEx() */
/**
 * \brief Inserts a local history element into the local history server
 * 
 * A server can instruct the local history server to keep a history of the given
 * property by utilizing this call.  The local history server will periodically call
 * the property as specified according to the following input parameters. The extended version
 * allows passing the additional parameter 'hstFlags' which can be used to direct the style
 * of points-of-interest (USE_POI_NONE, USE_POI_RANGE, USE_POI_TOLERANCE). 
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param len is the length of the local history call.
 * \param fmt is the TINE format of the local history call.
 * \param idx is the local history index to be identified with this local history element 
 *  (Note: this must be unique with this server process and within the history data repository).
 * \param hspec is a pointer to a HistorySpecification structure, where the relevant filtering
 *  criteria are given.
 * \param hstFlags gives additional storage directives. For example if no 'points of interest'
 *  companion files are desired, setting hstFlags = USE_POI_NONE will instruct the subsystem not 
 *  to maintain them.  If this parameter is 0, then the default criteria are used, and the call
 *  collapses to AppendHistoryInformation().
 *  
 * \note In lieu of calling AppendHistoryInformation(), it is recommended to make use of the
 * startup configuration file 'history.csv', in which all local history critieria can be 
 * introduced.  This has the advantage that no information is hard-coded.  Indeed by introducing
 * a 'history.csv' file one can immediately instruct a server to maintain local history 
 * information (even if the server is developed and maintained by someone else!).
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_AppendHistoryInformationEx.c
 */
TINE_EXPORT int AppendHistoryInformationEx(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec,int hstFlags);

/* ApplyHistoryFilter() */
/**
 * \brief Applies a history filter to an existing local history record.
 * 
 * A server can filter out long term storage by applying a filter condition which must be 
 * satisfied.
 *
 * \param idx is the local history record index of the history element to which the filter should
 *  by applied.
 *  (Note: this must already be allocated).
 * \param parsableFilterString is a parsable filter string defining the filter condition.
 *  This should be of the form /<context>/<server/<device>[<property>]<comparator>\<value\>
 *  where <comparator> is one of '=', '!=', '>', or '<' and \<value\> is the filter's
 *  threshold value.  
 *  Multiple comparator/value combinations can be ANDed (&&) or ORed (||) to refine the 
 *  filter condition.
 *  The filter targer address should deliver (or be able to deliver) a single 
 *  numeric or string name value.  If the filter string cannot be parsed or the target address
 *  does not exists, then no filter will be established.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 */
TINE_EXPORT int ApplyHistoryFilter(int idx,char *parsableFilterString);

/**
 * \brief Inserts local history data into the local history ring buffer
 * 
 * A server can insert data into the local history ring buffer directly with this call.
 * The inserted data behave in exactly the same manner as if added via a call the the
 * designated equipment module from the local history subsystem.
 *
 * The information provided in the call MUST correspond to a registered local history element or
 * the call will fail. The server developer can ensure that history data are inserted ONLY via 
 * this call (and not via the local history subsystem) by trapping the CA_HIST bit in the access
 * flag within the equipment module dispatch and returning an error condition such as 'not_implemented'.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history record is being kept.
 * \param dev is the device name for which a history record is being kept.
 * \param dout is a reference to a DTYPE object specifying the array length, data format,
 *  timestamp and other data stamps of the history record along with the accompanying data.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_AppendHistoryData.c
 */
TINE_EXPORT int AppendHistoryData(char *eqm,char *prp,char *dev,DTYPE *dout);

int LoadHistoryInformation(char *fname,char *expName,char *eqmName);
int getRemoteHistoryServer(char *eqm,char *prp);
unsigned long getAvailableDiskSpaceInBlocks(char *path);
void historyInit(char *expName,char *eqmName);
void historyCycle(int isScheduled);
TINE_EXPORT int setSTSData(char *eqm,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize);

#define HST_SEEK_CLOSEST_OLD 0x01
#define HST_USE_POI_FILENAME 0x10

typedef struct HstReq
{
  char *eqm; /**< target equipment module */
  char *prp; /**< target property */
  char *dev; /**< target device name */
  int index; /**< specific index */
  int fieldindex; /**< specific field index (if complex data type) */
  int sample; /**< specific raster sample */
  int options; /**< specific options */
} HstReqType;
              
typedef struct HstRng
{
  double starttime; /**< time range start */
  double stoptime; /**< time range stop */
  float lwrLimit;  /**< lower value limit (if filter given) */
  float uprLimit;  /**< upper value limit (if filter given) */
  int startstamp; /**< system/user stamp range start */
  int stopstamp; /**< system/user stamp range stop */
  int userstamp; /**< TRUE => filter on user-stamp instead of system stamp */
} HstRngType;          

TINE_EXPORT int SaveShortTermData(char *eqm, char *prp, char *dev, HstRngType rng);

int loadLTSData(HstTblEntry *hst,char *fn,HstReqType *req, HstRngType rng,BYTE *data,int fmt,char *tag,int len);
int getLTSData(HstReqType *req, HstRngType rng,BYTE *data,int fmt,char *tag,int len);
int getSTSData(HstReqType *req, HstRngType rng,BYTE *data,int fmt,char *tag,int len);
int getRBSummary(int idx,double sampleInterval,DSUMMARY *ds);
int getSTSDataSummary(char *eqm,char *prp, char *dev,double sampleInterval,DSUMMARY *ds);
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound);
int isInRange(DTYPE *d,float lwrLimit,float uprLimit);
time_t getStartTimeFromDepth(char *eqm,char *prp,char *dev,int depth);
int dumpHistoryManifest(char *eqm,int to_stdout);
int isPrpInHistoryTable(char *prp);

/* GetLastStoredData() */
/**
 * \brief Retreives the last stored data set from disk for the given input parameters
 * 
 * A server can retrieve the last stored data set from disk for any given local history
 * property.  This is frequently useful to insure that a server restart will reload data 
 * various sets, for instance when data are collected according to intermittent triggers, 
 * etc.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param idx is the targeted array index (if the history element represents an array)
 *  identifying the beginning of the range of data to retrieve.
 * \param data is the buffer to receive the data.  It must be pre-allocated to at least the
 *  array length specified in the 'len' parameter.
 * \param fmt is the TINE format of the local history call.
 * \param len is the length of the local history call.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_getLastStoredData.c
 */
TINE_EXPORT int GetLastStoredData(char *eqm,char *prp, char *dev,int idx,BYTE *data,int fmt,int len);
#define getLastStoredData(a1,a2,a3,a4,a5,a6,a7) GetLastStoredData(a1,a2,a3,a4,a5,a6,a7)

/**
* \brief Retreives the last stored data set from disk for the given input parameters (extended call).
*
* A server can retrieve the last stored data set from disk for any given local history
* property.  This is frequently useful to insure that a server restart will reload data
* various sets, for instance when data are collected according to intermittent triggers,
* etc.
* The extended call can also return the timestamp of the last stored data if required.
*
* \param eqm is the 6-character local equipment identifier name, which is internal to server.
* \param prp is the requested property for which a history is to be kept
* \param dev is the device name to be associated with the property name supplied as the second
*  parameter.
* \param idx is the targeted array index (if the history element represents an array)
*  identifying the beginning of the range of data to retrieve.
* \param data is the buffer to receive the data.  It must be pre-allocated to at least the
*  array length specified in the 'len' parameter.
* \param fmt is the TINE format of the local history call.
* \param len is the length of the local history call.
* \param tstamp (out) will contain the TINE data timestamp of the last stored data if 
*  the call is successful and a non-NULL reference pointer is passed.
*
* \return 0 upon success, otherwise a TINE error code.
*/
TINE_EXPORT int GetLastStoredDataEx(char *eqm, char *prp, char *dev, int idx, BYTE *data, int fmt, int len, double *tstamp);

TINE_EXPORT int GetLocalHistoryData(char *eqm,char *prp, char *dev,DTYPE *dout,DTYPE *din);
int HISTQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut);

/* SetMaximumHistoryTableRecords() */
/**
 * \brief Sets the maximum local history table size
 * 
 * The local history subsystem makes use of a table containing all relevant history contracts.
 * The maximum size of the this table is governed by this routine.
 *
 * \param value is the desired maximum history table size (default is 100).
 *  
 */
TINE_EXPORT void SetMaximumHistoryTableRecords(int value);

/* SetPointsOfInterestRangeFactor() */
/**
 * \brief Sets the definitive property range (maximum-minimum) factor which determines whether
 * a value is consider a point of interest.
 * 
 * If the change in a property's readback value is greater than the value's display range times
 * this range factor, then the value is considered a point of interest and therefore always
 * returned in archive calls over the relevant time range.
 *
 * \param value is the desired display range factor (default is 0.25).
 *  
 */
TINE_EXPORT void SetPointsOfInterestRangeFactor(float value);

/* SetPointsOfInterestToleranceFactor() */
/**
 * \brief Sets the definitive property tolerance factor which determines whether
 * a value is consider a point of interest.
 * 
 * If the change in a property's readback value is greater than the value's tolerance (if non-zero)
 * times this range factor, then the value is considered a point of interest and therefore always
 * returned in archive calls over the relevant time range.
 *
 * \param value is the desired display range factor (default is 10).
 *  
 */
TINE_EXPORT void SetPointsOfInterestToleranceFactor(float value);

/* SetUseStandardHistoryFiles() */
/**
 * \brief Instructs the server to use the so-called 'standard' set of 
 * local history files.
 * 
 * The 'standard' set of local history files is pre-allocated on the local file system
 * to be able to contain a 'worst-case' respository of random-access local history
 * records according to the assigned polling interval and data size of the record
 * (i.e. a null result is assumed for any would-be filtering action).  The initial allocation of
 * such a file set is of course time-consuming if a depth of several months is required
 * and the records are large.  However the subsequent data retrieval is very fast.
 *
 * This routine will enforce the usage of the standard local history file set.
 *
 * \param value is the desired value of this setting (default: FALSE).
 *  
 */
TINE_EXPORT void SetUseStandardHistoryFiles(int value);

/**
 * \brief Returns TRUE of FALSE depending on whether the given history record index 
 * makes use of the so-called 'standard' set of local history files.
 *
 * The 'standard' set of local history files is pre-allocated on the local file system
 * to be able to contain a 'worst-case' respository of random-access local history
 * records according to the assigned polling interval and data size of the record
 * (i.e. a null result is assumed for any would-be filtering action).  The initial allocation of
 * such a file set is of course time-consuming if a depth of several months is required
 * and the records are large.  However the subsequent data retrieval is very fast.
 *
 * \param index is the history record index for which the setting is desired.
 *
 * \return TRUE of FALSE depending on whether the given history record index 
 * makes use of the so-called 'standard' set of local history files.
 * 
 */
TINE_EXPORT int GetUseStandardHistoryFiles(int index);

/* SetMinimumDiskSpaceInBlocks() */
/**
 * \brief Sets the amount of free disk space the local history subsystem should try to maintain.
 * 
 * Normally the local history subsystem maintains a history repository based on 
 * the assigned depth in months of the history records.
 * In addition, one can make use of this call to establish a minimum floor to the 
 * amount of free disk space which should be maintained on the server.
 * If this floor is reached, then the oldest history files will be removed until 
 * the disk free space is once again above the assigned floor.
 *
 * \param value is the desired minimum free disk space (in blocks) to be
 * maintained on the local history partition.
 *  
 */
TINE_EXPORT void SetMinimumDiskSpaceInBlocks(UINT32 spaceInBlocks);

/* SetStandardSpillOverSize(int value) */
/**
 * \brief Sets the amount of additional space (number of addition records) to
 *  allocate as a worst case when using the standard local history data file set.
 *
 * The local history standard file set is pre-allocated to manage a 'worst-case'
 * number of stored records based on all filtering criteria coming up negative and
 * making use of the configured history record polling  interval and record size.
 * If it is anticipated that some local history data will be 
 * scheduled, then an additional spill-over space should be applied.
 *
 * \param value is the desired additional spill over size (default 1000).
 *  
 * \see GetStandardSpillOverSize
 */
TINE_EXPORT void SetStandardSpillOverSize(int value);

/* GetStandardSpillOverSize(int value) */
/**
 * \brief Gets the amount of additional space (number of addition records) to
 *  allocate as a worst case when using the standard local history data file set.
 *
 * The local history standard file set is pre-allocated to manage a 'worst-case'
 * number of stored records based on all filtering criteria coming up negative and
 * making use of the configured history record polling  interval and record size.
 *
 * \return the additional spill over size configured (default 1000).
 *  
 * \see SetStandardSpillOverSize
 */
TINE_EXPORT int GetStandardSpillOverSize(void);

/* SetHistoryStaticFilesRepository() */
/**
 * \brief Sets the directory location where static local history files are maintained.
 * 
 * Normally the local history subsystem maintains a volatile history repository 
 * where history files older than the long term depth are regularly removed from 
 * the disk.
 * If it is desired to keep certain local history files from being automatically 
 * removed, such files can be copied by hand into the designated static files 
 * repository.  
 *
 * By default, this will be a sub-directory 'SAVED' located within the normal
 * history files repository given by e.g. the HISTORY_HOME environment variable or
 * otherwise set via API.
 * The location of the static repository can otherwise be set via this API call.
 *
 * \param path is the desired static history files repository location.
 *  
 * \see GetHistoryStaticFilesRepository(), SetHistoryFilesRepository(), GetHistoryFilesRepository()
 */
TINE_EXPORT void SetHistoryStaticFilesRepository(char *path);

/* GetHistoryStaticFilesRepository() */
/**
 * \brief Returns the directory location where static local history files are maintained.
 * 
 * Normally the local history subsystem maintains a volatile history repository 
 * where history files older than the long term depth are regularly removed from 
 * the disk.
 * If it is desired to keep certain local history files from being automatically 
 * removed, such files can be copied by hand into the designated static files 
 * repository.  
 *
 * By default, this will be a sub-directory 'SAVED' located within the normal
 * history files repository given by e.g. the HISTORY_HOME environment variable or
 * otherwise set via API.
 * The location of the static repository can be retrieved via this API call.
 *
 * \return the configured static history files repository location.
 *  
 * \see SetHistoryStaticFilesRepository(), SetHistoryFilesRepository(), GetHistoryFilesRepository()
 */
TINE_EXPORT char *GetHistoryStaticFilesRepository(void);

/* SetHistoryFilesRepository() */
/**
 * \brief Sets the directory location where local history files are maintained.
 * 
 * Typically the location for the local history repository is established by  
 * the environment variable HISTORY_HOME.
 * If not set, the default location for local history files is given by an
 * adjacent directory to the FEC_HOME location given by the sub-directory name
 * HISTORY.
 *
 * \param path is the desired history files repository location.
 *  
 * \see GetHistoryFilesRepository()
 */
TINE_EXPORT void SetHistoryFilesRepository(char *path);

/* SetHistoryFilesRepository() */
/**
 * \brief Returns the directory location where local history files are maintained.
 * 
 * Typically the location for the local history repository is established by  
 * the environment variable HISTORY_HOME.
 * If not set, the default location for local history files is given by an
 * adjacent directory to the FEC_HOME location given by the sub-directory name
 * HISTORY.
 *
 * \return the configured history files repository location.
 *  
 * \see SetHistoryFilesRepository()
 */
TINE_EXPORT char *GetHistoryFilesRepository(void);

extern int gAllowOverloadedConcurrentLocalHistories;
TINE_EXPORT void SetAllowOverloadedConcurrentLocalHistories(int value);
TINE_EXPORT int GetAllowOverloadedConcurrentLocalHistories(void);

/**
 * \internal
 */
TINE_EXPORT int FindAvailableHistoryIndex(char *eqm,char *prp,char *dev,int fmt,int siz);
#define findAvailableHistoryIndex(e,p,d,f,s) FindAvailableHistoryIndex(e,p,d,f,s)
int findRecordIndexInHstManifest(char *eqm,char *prp,char *dev,int fmt,int siz);

/**
 * \internal
 */
TINE_EXPORT int getFormattedHistoryDataPoint(DTYPE *dst,DTYPE *src);

extern int gLclHstLateAdditionWindow;
TINE_EXPORT void SetHistoryLateAddtionWindow(int value);
TINE_EXPORT int GetHistoryLateAdditionWindow(void);

TINE_EXPORT void SetAutoRemoveUnsynchronizedHistories(int value);

TINE_EXPORT int GetAutoRemoveUnsynchronizedHistories(void);

int getHstRecIdx(char *eqm, char *prp, char *dev, int *index, int *index_err);
int checkForStandardFiles(void);
#if !defined(SMALL_TINE_LIB) && !defined(VXWORKS)
extern unsigned long fsBlocksFree;
int checkForStandardFilesEnvironment(void);
int getOldestFile(char *path,char *filePattern,char *fileName,time_t fileTime);
void doLtsScavenger(void);
#endif
#ifndef SMALL_TINE_LIB
extern int gHstHomeMfUpdated;
extern NAME32 *outdatedFiles;
extern int nOutdatedFiles;
extern int nStdDailyFiles;
extern int canRescanOutdatedFiles;
extern float gPoiRangeFactor;
extern float gPoiToleranceFactor;
# define checkIsPoi(CF) \
  for (p=0; !poi && p<(int)hst->c.EqmSizeOut; p+=inc) {\
    t = useRange ? gPoiRangeFactor * hst->range :\
      gPoiToleranceFactor*(hst->aTolerance + hst->pTolerance*fabs((float)((CF *)d1)[p]));\
    if (fabs((float)((CF *)d0)[p] - (float)((CF *)d1)[p]) > t) poi = TRUE;\
  }
int resetStats(char *fname,int recsize);
int scanForOutdatedFiles(int index,int mdepth,int ddepth);
int makeStandardFiles(int index,int mdepth,int mday);
int checkForSavedFiles(int idx);
int isPointOfInterest(HstTblEntry *hst);
void dumpHstManifests(int to_stdout);
void chkMcaIntegrity(HstTblEntry *hst);
int updateHstHomeManifest(void);
#endif
#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _HSTLIB_H_ */
