/*
 * File swaplib.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef SWAPLIB_H_
#define SWAPLIB_H_
/**
 * \internal
 * \file swaplib.h
 * \brief Header for swaplib.c
 * \see swaplib.c
 */
#include "project.def"

#include "tinedefs.h"
#include "tinetype.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \internal
 * \brief Returns the address offset of a field in a structure.
 * 
 * \param struc  Structure name
 * \param field  Field name
 *
 */
#define OFFSETIN(struc,field)   ((size_t)&(((struc *)0L)->field))

#define MEMREV(d,s)\
  for (p1=(BYTE *)(d), p2=p1+(s)-1; p2 > p1; ++p1, --p2) {\
    *p1 ^= *p2; *p2 ^= *p1; *p1 ^= *p2;\
  }

void memrev(void *d, int len);
/* note: SPEEK and IPEEK denote a swap-and-peek. otherwise: the swaps are in-situ */
#if defined(NONALIGNED) || defined(INTEL) || defined(VMS_IEEE)
#  define SwapBytes(dat,size,fmt,dir) ;
#  define SSWAP(w) ((w) = (w))
#  define SPEEK(w) ((w) = (w))
#  define ISWAP(w) ((w) = (w))
#  define IPEEK(w) ((w) = (w))
#  define DSWAP(w) ((w) = (w))
#else
#  ifdef VMS   /* ints unswapped, special swapping for floats, doubles */
#    define SSWAP(w) ((w) = (w))
#    define SPEEK(w) ((w) = (w))
#    define ISWAP(w) ((w) = (w))
#    define IPEEK(w) ((w) = (w))
#  else
#    define SPEEK(w) ((((w)&0xff00)>>8)|(((w)&0xff)<<8))
#    define SSWAP(w) memrev(&(w),2)
#    define IPEEK(w) ((w>>24)|(w<<24)|((w&0xff00)<<8)|((w&0x00ff0000L)>>8))
#    define ISWAP(w) memrev(&(w),4)
#  endif
#  define DSWAP(w) memrev(&(w),8)
#  define SwapBytes(dat,size,fmt,dir) _swapBytes(dat,size,fmt,dir)
#endif
#define FSWAP ISWAP
#define LSWAP DSWAP

int _swapBytes(char *dat, int size, int fmt, int dir);
int cfHistorySwap(char *tag,BYTE *src,BYTE *dst,int dir,int num);

#define STRUCT_LIST_MARKER "_NEXT_"
/**
 * \internal
 * \brief key used to distinguish which servers are using which structures
 * 
 */
struct srvKey
{
  char key[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+8]; /**< in the form: "/context/server" */
  struct srvKey *nxt; 
};

/**
 * \internal
 * \brief Format description of a structure element
 * 
 * To describe a whole structure, structFormat is used as
 * a linked list. In TINE, linked lists are mostly used in
 * reverse order.
 */
struct structFormat
{
  int siz; /**< number of elements. */
  int fmt; /**< TINE CF_ format. */
  int off; /**< offset in the byte stream */
  int addr; /**< Byte offset within the structure. */
  char field[FIELD_NAME_SIZE]; /**< structure field name */
  char stag[TAG_NAME_SIZE]; /**< sub tag if fmt == CF_STRUCT */
  struct structStruct *owner; /**< reference to struct owner of this structFormat */
  struct structFormat *nxt; /**< Pointer to next element. */
};

/**
 * \internal
 * \brief Structure description for tagged structures
 * 
 * A structStruct contains a structFormat list of fields. 
 * It does not only describe a structure, but also describes
 * an array of data of that structure. 
 * At the same time, it is also a buffer.
 * It can also be used as an element of a linked list.
 * 
 * A structStruct is used for tagged structures.
 */
struct structStruct
{
  char name[TAG_NAME_SIZE];/**< structure tag */
  int  rawlen;             /**< raw/network data size of the described structure */
  int  size;               /**< local structure size (memory size) of the described structure */
  int  num;                /**< maximum number of structure sets buffered (array size) */
  int  sealed;             /**< structure has been sealed */
  int  hasExtSpace;        /**< has extended space (uses CF_STRING, CF_IMAGE, etc.) */
  BYTE *sdata;             /**< buffer for data to be sent over the network */
  struct srvKey *key;      /**< list of known servers using this structure */
  struct structFormat *s;  /**< structure definition */
  struct structStruct *next;/**< pointer to the next element in a linked list.*/
};

/**
 * \internal
 * \brief Global list of tagged structures.
 */
extern struct structStruct *structList;

struct structStruct *findstructEx(char *tag, char *key, ...);
TINE_EXPORT struct structStruct* FindStruct(char* tag);
struct structStruct *findstruct(char *tag);
int addtostruct(void *target,int addr,int size, int fmt, char *subtag, char *field);
int getFieldFormat(struct structStruct *s,char *field);
int getFieldSize(struct structStruct *s,char *field);
int addKeyToStruct(struct structStruct *s,char *key);
int isInvalidStructureTag(char *tag);
int getRegisteredStructNames(char *tags,int *ntags,int fmt,int includeKeys,int includeSystem);
int dumpRegisteredStrucures(void);
TINE_EXPORT int fillstruct(char *tag,int addr,int size,int fmt);

/**
 * \brief Adds a field description to a tagged structure.
 * 
 * Use this routine to register the fields of your structure within
 * the structure registry.  The fields can be any TINE format type
 * or another tagger structure (in which case fmt = CF_STRUCT, and
 * field = <parent tag>fieldname).
 *
 * \param tag structure tag name
 * \param addr is the address offset of this field inside the structure.
 * \param size is the array length if the described field. (not the size in bytes)
 * \param fmt format of the new struct element
 * \param field field name
 * 
 * \return 0 or out_of_local_memory if no struct could be found/allocated for tag.
 * 
 * \see SealTaggedStruct
 * 
 * \b Example:
 *
 * \include eg_taggedStruct3.c
 * 
 */
TINE_EXPORT int AddFieldToStruct(char *tag,int addr,int size,int fmt,char *field);
#define addFieldToStruct(a1,a2,a3,a4,a5) AddFieldToStruct(a1,a2,a3,a4,a5)

/**
 * \brief Returns the 'raw' structure length of the tagged structure given.
 * 
 * This routine will retrieve the raw structure length of the tagged structure
 * given from the structure registry.  This length is the length in bytes of the
 * pure byte stream and is NOT necessary identical to the 'sizeof()' call.
 * These two values will only be the same if the registered structure is properly 
 * aligned.
 *
 * \param tag structure tag name
 * 
 * \return the size in bytes of the raw structure length or the negative of a 
 * TINE return code.
 * 
 */
TINE_EXPORT int GetStructSize(char *tag);

/**
 * \brief Returns whether or not the given structure contains only block transferable 
 * memory (i.e. does not contain any variable length format data types).
 * 
 * \param tag structure tag name
 * 
 * \return TRUE if the structure fields are all non variable length types (i.e. no 
 * reference pointers). Otherwise the routine returns FALSE.  It also returns false if
 * the given structure is not registered.
 * 
 */
TINE_EXPORT int StructIsBlittable(char *tag);

TINE_EXPORT struct structStruct *GetStructFromTag(char *tag);
TINE_EXPORT struct structStruct *GetStructFromTagEx(char *tag,char *key);

/**
 * \brief Returns the structure registry status of the given structure
 *
 * \param tag structure tag name
 *
 * \return not_registered if the given structure tag is not found;
 * not_initialized if the structure is found in the registry but is not yet 
 * sealed (registration not complete!); 
 * 0 if the structure is a valid registered structure.
 *
 */
TINE_EXPORT int GetStructStatus(char *tag);

/**
 * \brief Seals a tagged structure (registration finished!).
 * 
 * Call this routine to signal the structure registry that there are
 * no more fields to be added to the tagged structure.
 * 
 * \param tag structure tag name
 * \param size total size of the structure in bytes
 * \param number number of elements of this structure to allocate
 * 
 * \return 0, un_allocated, out_of_local_memory 
 * 
 * \include eg_taggedStruct.c
 * 
 * \see AddFieldToStruct
 */ 
TINE_EXPORT int SealTaggedStruct(char *tag,int size,int number);
#define sealTaggedStruct(a1,a2,a3) SealTaggedStruct(a1,a2,a3)
int _structSwap(struct structFormat *s,BYTE *dat,int offset,int *tail,BYTE *struc,int dir);
int _taggedStructSwap(char *tag,char *key,BYTE *dat,BYTE *struc,int dir,int num,int structsize);
int _systemStructSwap(struct structFormat *sf,BYTE *dat,BYTE *struc,int dir,int num,int structsize);
int initControlStructs(void);
int queryStructFormat(char *tag,BYTE *sfq,int fmt,int *sfqsiz);
TINE_EXPORT int GetStructFormat(char *tag,NAME64DBLDBL *sfn64, int *n);
void freeStruct(struct structFormat *sf);
#define FREESTRUCT(sf) freeStruct(sf), sf = NULL
void freeStructList(int freeMarkedList);
struct structStruct *getTaggedStruct(char *tag);
TINE_EXPORT void flushStructCache(void);

/**
* \internal
* \brief Structure description for PktHdr
* \see PktHdr
*/
extern struct structFormat *PktHdrStructStruct;

/**
* \internal
* \brief Structure description for ReqHdr
* \see ReqHdr
*/
extern struct structFormat *ReqHdrStructStruct;

/**
* \internal
* \brief Structure description for RspHdr
* \see RspHdr
*/
extern struct structFormat *RspHdrStructStruct;

/**
 * \brief Field list of structure CONTRACT.
 */
extern struct structFormat *CONTRACTStruct;
extern struct structFormat *CONTRACT3Struct;

/**
 * \brief Field list of structure ProducerHeaderStruct.
 */
extern struct structFormat *PrdrHdrStruct;
extern struct structFormat *PrdrHdr3Struct;
/**
* \brief Structure description for SubRspHdr
*/
extern struct structFormat *SubRspHdrStruct;
/**
* \brief Structure description for SubReqHdr
*/
extern struct structFormat *SubReqHdrStruct;

/**
 * \brief Field list of structure SubInfoPkt.
 */
extern struct structFormat *SubInfoPktStruct;
extern struct structFormat *SubInfoPkt3Struct;

/**
 * \brief Field list of structure RPCDataStruct.
 */
extern struct structFormat *RPCDataStructStruct;

/**
 * \brief Field list of structure RPCFecStruct.
 */
extern struct structFormat *RPCFecStructStruct;

/**
 * \brief Field list of structure FilterStruct.
 */
extern struct structFormat *FilterStructStruct;

/**
 * \brief Field list of structure ADSStruct.
 */
extern struct structFormat *ADSStructStruct;
extern struct structFormat *ADS3StructStruct;
extern struct structFormat *ADS4StructStruct;

/**
 * \brief Field list of structure AMSStruct.
 */
extern struct structFormat *ALMSStructStruct;
extern struct structFormat *AMS3StructStruct;
extern struct structFormat *AMS4StructStruct;

/**
 * \brief Field list of structure AWSStruct.
 */
extern struct structFormat *AWSStructStruct;

/**
 * \brief Field list of structure AWLQSStruct.
 */
extern struct structFormat *AWLQSStructStruct;

/**
 * \brief Field list of structure PropertyQueryStruct.
 */
extern struct structFormat *PropertyQueryStructStruct;

/**
 * \brief Field list of structure PropertyQueryStructEx.
 */
extern struct structFormat *PropertyQueryStructExStruct;

/**
 * \brief Field list of structure XPropertyQueryStruct.
 */
extern struct structFormat *XPropertyQueryStructStruct;

/**
 * \brief Field list of structure PrpQueryStruct.
 */
extern struct structFormat *PrpQueryStructStruct;

/**
 * \brief Field list of structure ActivityQueryStruct.
 */
extern struct structFormat *ActivityQueryStructStruct;

/**
 * \brief Field list of structure ServerStateStruct.
 */
extern struct structFormat *ServerStateStructStruct;

/**
 * \brief Field list of structure GlbDataHdr3.
 */
extern struct structFormat *GlbDataHdr3Struct;

/**
 * \brief Field list of structure GlbDataHdr4.
 */
extern struct structFormat *GlbDataHdr4Struct;

/**
* \brief Field list of structure GlbDataHdr.
*/
extern struct structFormat *GlbDataHdrStruct;

/**
 * \brief Field list of structure ClientQueryStruct.
 */
extern struct structFormat *LClientQueryStructStruct;
extern struct structFormat *ClnInfoQueryStructStruct;

/**
 * \brief Field list of structure hdrSPECTRUM.
 */
extern struct structFormat *SpectrumHeaderStruct;

/**
* \brief Field list of structure hdrGSPECTRUM.
*/
extern struct structFormat *GSpectrumHeaderStruct;

/**
 * \brief Field list of structure hdrIMAGE.
 */
extern struct structFormat *ImageHeaderStruct;

/**
 * \brief Structure description for type MDA (multi-dimensional array)
 */
extern struct structFormat *MDAHeaderStruct;

/**
 * \brief Field list of structure HRSstruct.
 */
extern struct structFormat *HRSStructStruct;

extern struct structFormat *SummaryStructStruct;

extern struct structFormat *FDSStructStruct;

extern struct structFormat *FASStructStruct;

extern struct structFormat *ConTblInfoStructStruct;

extern struct structFormat *WriteAccessQueryStruct;

extern struct structFormat *ServerSettingsStructStruct;

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* SWAPLIB_H_ */
