/*
 * File threader.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL),
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file threader.h
 *
 * thread-specific definitions and prototypes.
 *
 */
#ifndef THREADER_H_
#define THREADER_H_

#include "tinetype.h"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifdef MULTITHREADED

typedef struct
{
  void (*tsk)(void);
  int rate;
  int hasReference;
  void *elRef;
  void *EqmRef;
} BkgThreadFcn;

#if defined(UNIX)

#define STD_THREAD_PRIORITY 0
#define MIN_THREAD_PRIORITY -99
#define MAX_THREAD_PRIORITY 99
#define THREAD_TYPE void *
#define THREAD_PARAM void *
#include "pthread.h"
#include "semaphore.h"
typedef pthread_mutex_t * MXHANDLE;
typedef sem_t * SEMHANDLE;
typedef pthread_t THRHANDLE;
extern pthread_attr_t gBkgThreadAttr;
void *bkgThreadTask(void *lpvThreadParam);
#define StopThread(t) pthread_cancel((pthread_t)t)
#define ReleaseMutex(mx) pthread_mutex_unlock(mx)
#define ReleaseSemaphore(mx) sem_post(mx)

#elif defined(WIN32)

#define STD_THREAD_PRIORITY THREAD_PRIORITY_NORMAL
#define MIN_THREAD_PRIORITY THREAD_PRIORITY_LOWEST
#define MAX_THREAD_PRIORITY THREAD_PRIORITY_HIGHEST
#define THREAD_TYPE DWORD WINAPI
#define THREAD_PARAM LPVOID
typedef HANDLE MXHANDLE;
typedef HANDLE SEMHANDLE;
typedef HANDLE THRHANDLE;
extern DWORD dwBkgThreadId;
DWORD WINAPI bkgThreadTask(LPVOID lpvThreadParam);
#define StopThread(t) TerminateThread((HANDLE)t,0)
#define bkgThreadResume(t) ResumeThread((HANDLE)t)

#elif defined(VXWORKS)
/* these priorities are relative to the task spawn priority */
#define STD_THREAD_PRIORITY 0
#define MIN_THREAD_PRIORITY -99
#define MAX_THREAD_PRIORITY 99
#define THREAD_TYPE int
#define THREAD_PARAM void *
typedef SEM_ID MXHANDLE;
typedef SEM_ID SEMHANDLE;
typedef int THRHANDLE;
THREAD_TYPE bkgThreadTask(THREAD_PARAM lpvThreadParam);
int vxwReleaseSemaphore(MXHANDLE mx);
#define ReleaseMutex(mx) vxwReleaseSemaphore(mx)
#define ReleaseSemaphore(mx) vxwReleaseSemaphore(mx)
#define StopThread(t) taskDelete(t)

#endif /* UNIX or WIN32 or VXWORKS */

typedef struct
{
  SOCKET sck;
  int type;
  SEMHANDLE sem;
} SckThreadInfo;

THREAD_TYPE sendStreamThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE recvStreamThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE clThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE clnThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE srvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE schThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE sckSendThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE sckRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE glbRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE dcsEvnRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE mcstRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE scvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE reinitThreadTask(THREAD_PARAM lpvThreadParam);
int eqmInitMutexSet(void *lpvMutexParam);
int stkInitMutexSet(void);

extern SEMHANDLE hSendStreamSemaphore;
extern SEMHANDLE hSchedulerSemaphore;
extern SEMHANDLE hNotificationSemaphore;
extern MXHANDLE hSystemInitMutex;
extern MXHANDLE hSystemServerMutex;
extern MXHANDLE hSystemKernelMutex;
extern MXHANDLE hCycleTriggerMutex;
extern MXHANDLE hSystemClientMutex;
extern MXHANDLE hSystemGlobalsUpdateMutex;
extern MXHANDLE hSystemMCastUpdateMutex;
extern MXHANDLE hLinkCbMutex;
extern MXHANDLE hLinkTblMutex;
extern MXHANDLE hXferMutex;
extern MXHANDLE hSndMutex;
extern MXHANDLE hDpdLinkMutex;
extern MXHANDLE hMcaLinkMutex;
extern MXHANDLE hLwdLinkMutex;
extern MXHANDLE hLinkDataMutex;
extern MXHANDLE hLinkLstMutex;
extern MXHANDLE hTcpBucketMutex;
extern MXHANDLE hEnsMutex;
extern MXHANDLE hTxMutex;
extern MXHANDLE hAttrMutex;
extern MXHANDLE hGeneralMutex;
extern MXHANDLE hSrvTblMutex;
extern MXHANDLE hLinkQueueMutex;
extern MXHANDLE hClientTableMutex;
extern MXHANDLE hContractTableMutex;
extern MXHANDLE hLogMutex;
extern MXHANDLE hDbMutex;
extern MXHANDLE hCsvMutex;
extern MXHANDLE hMfMutex;
extern MXHANDLE hHistoryTableMutex;
extern MXHANDLE hHistoryCycleMutex;
extern MXHANDLE hAlmTblMutex;
extern MXHANDLE hStkAccessMutex;

extern int gClnThreadRunning;
extern int gSrvThreadRunning;

extern int srvThreadPriority;
extern int clnThreadPriority;
extern int bkgThreadPriority;

int InitMutex(MXHANDLE *mx);
int InitSystemMutexSet(void);
int CloseSystemMutexSet(void);
int CloseClientMutexSet(void);
int WaitForMutex(MXHANDLE mx,int to);
int ReleaseSystemMutex(MXHANDLE mx);
int CloseSystemMutex(MXHANDLE mx);
int WaitForSemaphore(SEMHANDLE mx,int to);
int ReleaseSystemSemaphore(SEMHANDLE mx);
int FreeSystemSemaphore(SEMHANDLE mx);
int WaitForThreadTermination(THRHANDLE hThr);
int WaitForThreadTerminationEx(THRHANDLE hThr,int timeout);
SEMHANDLE CreateSystemSemaphore(char *name);

THRHANDLE clCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE mcstRecvCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE glbRecvCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE dcsEvnRecvCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE sckRecvCreateThread(SOCKET sck,int type);
THRHANDLE sckSendCreateThread(SOCKET sck,int type,SEMHANDLE *semHandle);
THRHANDLE schCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE recvStreamCreateThread(SOCKET sck,int type);
THRHANDLE bkgCreateThread(BkgThreadFcn *tsk);
THRHANDLE clnCycleCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE srvCycleCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE scavengerCreateThread(THREAD_PARAM lpvThreadParam);

void stopGlobalsThread(SOCKET sck);
char *getSystemMutexName(MXHANDLE mx);

#define WaitForSystemMutex(mx,to) WaitForMutex(mx,to)

#define MTX_PRIMARY 0
#define MTX_SECONDARY 1
#define MTX_CARRIER 2

#define SuspendSystemMutex(mx) ;
#define RestoreSystemMutex(ms) ;

#else /* not MULTITHREADED */

typedef int MXHANDLE;
typedef int SEMHANDLE;
typedef int THRHANDLE;
extern int hSystemInitMutex;
extern int hSystemServerMutex;
extern int hSystemKernelMutex;
extern int hCycleTriggerMutex;
extern int hNotificationSemaphore;
extern int hSystemClientMutex;
extern int hSystemGlobalsUpdateMutex;
extern int hSystemMCastUpdateMutex;
extern int hLinkCbMutex;
extern int hLinkTblMutex;
extern int hXferMutex;
extern int hSndMutex;
extern int hDpdLinkMutex;
extern int hMcaLinkMutex;
extern int hLwdLinkMutex;
extern int hLinkDataMutex;
extern int hLinkLstMutex;
extern int hEnsMutex;
extern int hTxMutex;
extern int hAttrMutex;
extern int hGeneralMutex;
extern int hSrvTblMutex;
extern int hTcpBucketMutex;
extern int hLinkQueueMutex;
extern int hClientTableMutex;
extern int hContractTableMutex;
extern int hLogMutex;
extern int hDbMutex;
extern int hCsvMutex;
extern int hMfMutex;
extern int hHistoryTableMutex;
extern int hHistoryCycleMutex;
extern int hAlmTblMutex;
extern int hStkAccessMutex;
#ifdef UNIX
int checkMutexState(int m);
#endif
extern int hMutexSuspendedState;
#define SuspendSystemMutex(m) hMutexSuspendedState = m, m = 0
#define RestoreSystemMutex(m) m = hMutexSuspendedState
#define ReleaseSystemSemaphore(m) m = 0
#define WaitForSemaphore(m,t) 0
#endif /* MULTITHREADED */

extern TINE_EXPORT int useMultiThreadedBackgroundTasks;
extern TINE_EXPORT int useMultiThreadedEquipmentFunctions;
extern TINE_EXPORT int useMultiThreadedStockFunctions;
extern THRHANDLE hGlbThread;
extern THRHANDLE hDscEvnThread;
extern THRHANDLE hMcstThread;
extern THRHANDLE hSchdThrd;
extern THRHANDLE hScvrThrd;

#define thread_dbglog(txt) {\
    dbglog("[T1]: %.6s %.64s %.64s -> %s",\
    cl->contract.EqmName,cl->contract.EqmDeviceName,\
    cl->contract.EqmProperty,(txt));\
  }
#define MUTEX_WAIT_AUGMENT 3

extern int gSystemMutexSetInitialized;
int dumpThreadPriorities(void);
void YieldThread(void);
THRHANDLE GetThreadHandle();
void SetThreadName(char *name);

TINE_EXPORT void SystemSetCycleTimer(void);
/**
 * \fn TINE_EXPORT void SystemStartCycleTimer(void)
 * \brief Starts a TINE engine 'cycler' running (SystemCycle()) in a separate thread.
 *
 * Available in multi-threaded shared-library builds, this routine starts a separate thread
 * where SystemCycle() is allowed to run in a tight loop.  Only one instance of the cycler thread
 * is allowed to run in a single process.  On windows systems, the thread will also see to it
 * that windows messages are allowed to be processed.
 *
 * \note The Windows DLLs (tine32.dll or tine64.dll) include the cycler routines but does
 * not start the cycler when the DLL is loaded unless the environment variable
 * TINE_START_CYCLER = TRUE is set.
 * Supplementary libraries such as 'tbufsrv'
 * or ACOP will however start the cycler automatically.  The call SystemSetCycleTimer()
 * can be called more than once, but has no effect if the cycler is already running.
 * When the shared library libtinemt.so is linked against,
 * the routine is automatically called when the library is loaded, but ONLY if
 * either a) the library is built with -DTINE_START_CYCLER=TRUE or if the environment
 * variable TINE_START_CYCLER = TRUE is set.  Builds using the static libraries,
 * tinemt.lib (WINDOWS) or libtinemt.a (UNIX) will
 * need to include tcycler.c in the project in order to make use of this routine.
 *
 * \note use caution setting a global environment variable TINE_START_CYCLER as that
 * might impact all TINE processes, including those which might want to manage their
 * own cycling.
 *
 * \note The developer is always at liberty to manage SystemCycle() in any manner he sees fit.
 * and to, for instance, include other tasks which need to be cycled over together in the same
 * managed loop.  However, it should be guaranteed that SystemCycle() runs in a single thread
 * per process.
 *
 * \see SystemKillCycleTimer().
 */
TINE_EXPORT void SystemStartCycleTimer(void);
/**
 * \fn TINE_EXPORT void SystemKillCycleTimer(void)
 * \brief Stops the TINE engine 'cycler' thread.
 *
 * Available in multi-threaded shared-library builds, this routine stops the separate thread
 * where SystemCycle() runs in a tight loop (started via a call to SystemSetCycleTimer().
 *
 * \see SystemStartCycleTimer().
 */
TINE_EXPORT void SystemKillCycleTimer(void);

/**
 * \fn TINE_EXPORT void SystemWaitCycleTimer(void)
 * \brief Waits for the TINE engine 'cycler' thread to stop running.
 *
 * Available in multi-threaded shared-library builds, this routine waits for the cycler thread
 * to stop running before continuing.
 *
 * \see SystemStartCycleTimer().
 */
TINE_EXPORT void SystemWaitCycleTimer(void);

extern int gIsDynamicLoad;
TINE_EXPORT int TineLoadedDynamically(void);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* THREADER_H_ */
