/*
 * File tinetype.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL),
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef TINETYPE_H_
#define TINETYPE_H_

#include "project.def"

#include <time.h>
#include <stdarg.h>
#include "tinedefs.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \internal
 * Assign a completion code and jump to a 'cleanup' label where
 * files can be fclose()ed, memory free()ed. etc. if need be ...
 *
 * A very useful macro when you're using straight "c".
 */
#define ccerr(e) {cc = (e); goto err; }

/**
 * \internal
 * \brief Upper bound on the system polling rate.
 *
 * Absolute maximum allowed wait time on a select().  Typically the system polling
 * interval is on the order of 10 to 20 milliseconds by default (a clock tick).
 */
#define MAXALLOWED_POLLINTVL 65535L

/**
 * \internal
 * \brief milliseconds in a day
 *
 * Absolute maximum millisecond differential (protects agains ntp long jumps)
 */
#define MAXALLOWED_INTVL 86400000L

/**
 * \internal
 * \brief difference in milliseconds between two struct timeval's.
 */
#define MSECS(a,b) ((a).tv_sec-(b).tv_sec > MAXALLOWED_POLLINTVL ? MAXALLOWED_POLLINTVL :\
		    ((a).tv_sec-(b).tv_sec)*1000+((a).tv_usec-(b).tv_usec)/1000)

#define tvDeltaMs(a,b) ((a).tv_sec-(b).tv_sec > MAXALLOWED_INTVL ? MAXALLOWED_INTVL :\
        ((a).tv_sec-(b).tv_sec)*1000 + ((a).tv_usec-(b).tv_usec)/1000)

/**
 * \internal
 * \brief standard path length for various locations
 */
#define STD_PATH_LENGTH 128

/**
 * \internal
 * \brief Maximal length of returned status string
 */
#define STATUS_SHORTSIZE 32
#define STATUS_SIZE 192

/**
 * \internal
 * \brief Maximal length of a FEC name.
 */
#define FEC_NAME_SIZE     16
#define FEC_DESC_SIZE     64
#define FEC_LOCATION_SIZE 48
#define FEC_VERSION_SIZE  16
#define FEC_HDW_SIZE      32
#define FEC_RESP_SIZE     32
#define FEC_OS_SIZE       16

/**
 * \internal
 * \brief maximal length of an export name.
 */
#define EXPORT_NAME_SHORTSIZE    16
#define EXPORT_NAME_SIZE    32

/**
 * \internal
 * \brief Maximal length of a local module name.
 */
#define EQM_NAME_SIZE      8
#define EQM_NAME_SHORTSIZE 6

/**
 * \internal
 * \brief Maximal length of a property name.
 */
#define PROPERTY_NAME_SIZE     64
#define PROPERTY_NAME_SHORTSIZE 32
#define PROPERTY_DESC_SIZE     64
#define PROPERTY_REDIR_SIZE    STATUS_SIZE

/**
 * \internal
 * \brief Maximal length of a device name.
 */
#define DEVICE_NAME_SIZE   64
#define DEVICE_DESC_SIZE   64
#define DEVICE_LOCATION_SIZE 64
#define DEVICE_NAME_SHORTSIZE 16
#define DEVICE_NAME_MAXSIZE 1024
#define DEVICE_REDIR_SIZE STATUS_SIZE

/**
 * \internal
 * \brief Maximal length of a context name.
 */
#define CONTEXT_NAME_SHORTSIZE  16
#define CONTEXT_NAME_SIZE  32

#define SUBSYSTEM_NAME_SIZE 16
#define SUBSYSTEM_NAME_SHORTSIZE 5

#define SET_NAME_SIZE 32

#define FORMAT_STR_SIZE 64
#define ACCESS_STR_SIZE 64
#define ADDR_SIZE 16
#define ADDR_SIZE_IPV6 40
#define UNITS_SIZE 16
#define USERNAME_SIZE 16
#define GROUPNAME_SIZE 16
#define GROUPSOURCE_SIZE 64
#define FIELD_NAME_SIZE 16
#define TAG_NAME_SIZE 16
#define TAG_NAME_SHORTSIZE 8
#define HISTORY_HOME_SIZE 128
#define XFIX_SIZE 8
#define USERTYPE_SIZE 8

#define FULL_NAME_SIZE DEVICE_NAME_SIZE + PROPERTY_NAME_SIZE + EXPORT_NAME_SIZE + CONTEXT_NAME_SIZE + 8
#define FULL_NAME_MAXSIZE DEVICE_NAME_MAXSIZE + PROPERTY_NAME_SIZE + EXPORT_NAME_SIZE + CONTEXT_NAME_SIZE

#define ALARM_TAG_SIZE 32
#define ALARM_TAG_SHORTSIZE 16
#define ALARM_TAG_LONGSIZE 64

#define CLOG_STATUS_NONE 0
#define CLOG_STATUS_INFO 1
#define CLOG_STATUS_WARN 2
#define CLOG_STATUS_ERR 3

#define CLOG_PRIORITY_NONE 0
#define CLOG_PRIORITY_USEFUL 1
#define CLOG_PRIORITY_IMPORTANT 2
#define CLOG_PRIORITY_URGENT 3

/**
 * \struct CLOG
 * \brief defines the standard central logger entry structure
 *
 */
typedef struct
{
  char context[32]; /**< the logging context (e.g. the machine context) */
  char tag[32];     /**< a logging tag identifier */
  char logger[32];  /**< the user/server/application submitting the logging entry */
  char text[256];   /**< the text to be logged */
  int  priority;    /**< the log entry priority (CLOG_PRIORITY_NONE, CLOG_PRIORITY_USEFUL, etc.)  */
  int  status;      /**< the log entry status (CLOG_STATUS_NONE, CLOG_STATUS_INFO, CLOG_STATUS_WARN, etc.) */
  double timeSent;  /**< local timestamp of the logging application at the time of submission */
  double timeLogged;/**< local timestamp of the central logger at the time of receipt */
} CLOG;

/* clslog */
/**
 * \brief Sends a logging entry to the central logging server
 *
 * An application (be it client or server) can send an entry to the central logger at any
 * time.  As this will require WRITE access, the entry will only be accepted if the caller
 * has WRITE privileges.
 *
 * \param context is the logging context.  If NULL or empty, the currently defined TINE context
 * will be used.  Logging entries can be browsed or sorted according to context.
 * \param tag is a brief logging tag. Logging entries can be browsed or sorted according to logging tag.
 * \param logger is the user submitting the logging entry.  This is typically the user name, but can be
 * a fec name, or application name, for example.  Logging entries can be browsed or sorted according the
 * logger.
 * \param priority is the logging entry priority.  This should be one of CLOG_PRIORITY_NONE, CLOG_PRIORITY_USEFUL,
 * CLOG_PRIORITY_IMPORTANT, or CLOG_PRIORITY_URGENT.
 * \param status is the logging status.  This should be one of CLOG_STATUS_NONE, CLOG_STATUS_INFO, CLOG_STATUS_WARN,
 * or CLOG_STATUS_ERR.
 * \param text is the logging entry.  Variable argument ellipses (i.e. printf-style) is allowed.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see feclog()
 */
TINE_EXPORT int clslog(char *context,char *tag,char *logger,int priority,int status,char *text,...);

/**
 * \struct NAME8
 * \brief Defines a TINE 8-character fixed-length string data object
 *
 */
typedef struct
{
  char name[8]; /**< the string buffer containing the string information */
} NAME8;

/**
 * \struct NAME16
 * \brief Defines a TINE 16-character fixed-length string data object
 *
 */
typedef struct
{
  char name[16]; /**< the string buffer containing the string information */
} NAME16;

/**
 * \struct NAME32
 * \brief Defines a TINE 32-character fixed-length string data object
 *
 */
typedef struct
{
  char name[32]; /**< the string buffer containing the string information */
} NAME32;

/**
 * \struct NAME48
 * \brief Defines a TINE 48-character fixed-length string data object
 *
 */
typedef struct
{
  char name[48]; /**< the string buffer containing the string information */
} NAME48;

/**
 * \struct NAME64
 * \brief Defines a TINE 64-character fixed-length string data object
 *
 */
typedef struct
{
  char name[64]; /**< the string buffer containing the string information */
} NAME64;

/**
 * \struct NAME8I
 * \brief Defines a TINE pairwise data object containing an 8-character fixed-length string
 * followed by a 4-byte integer
 *
 */
typedef struct
{
  char name[8]; /**< the string buffer containing the string information */
  SINT32 ival;   /**< the long integer value */
} NAME8I;

/**
 * \struct NAME16I
 * \brief Defines a TINE pairwise data object containing a 16-character fixed-length string
 * followed by a 4-byte integer
 *
 */
typedef struct
{
  char name[16]; /**< the string buffer containing the string information */
  SINT32 ival;   /**< the long integer value */
} NAME16I;

/**
 * \struct NAME32I
 * \brief Defines a TINE pairwise data object containing a 32-character fixed-length string
 * followed by a 4-byte integer
 *
 */
typedef struct
{
  char name[32]; /**< the string buffer containing the string information */
  SINT32 ival;   /**< the long integer value */
} NAME32I;

/**
 * \struct NAME48I
 * \brief Defines a TINE pairwise data object containing a 48-character fixed-length string
 * followed by a 4-byte integer
 *
 */
typedef struct
{
  char name[48]; /**< the string buffer containing the string information */
  SINT32 ival;   /**< the long integer value */
} NAME48I;

/**
 * \struct NAME64I
 * \brief Defines a TINE pairwise data object containing a 64-character fixed-length string
 * followed by a 4-byte integer
 *
 */
typedef struct
{
  char name[64]; /**< the string buffer containing the string information */
  SINT32 ival;   /**< the long integer value */
} NAME64I;

typedef struct
{
  SINT32 lval;
  SINT32 ival;
} LNGINT;

/**
 * \struct INTINT
 * \brief Defines a TINE pairwise data object containing a 4-byte integer doublet
 *
 */
typedef struct
{
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
} INTINT;

/**
 * \struct FLTINT
 * \brief Defines a TINE pairwise data object containing a float value followed by
 * a 4-byte integer value
 *
 */
typedef struct
{
  float  fval; /**< the float value */
  SINT32 ival; /**< the long integer value */
} FLTINT;

/**
 * \struct DBLINT
 * \brief Defines a TINE pairwise data object containing a double value followed by
 * a 4-byte integer value
 *
 * \note This structure has limited applicability and should only be used where structures
 * on compiled with 4-byte boundaries.
 *
 */
typedef struct
{
  double dval; /**< the double value */
  SINT32 ival; /**< the long integer value */
} DBLINT;

/**
 * \struct DBLDBL
 * \brief Defines a TINE pairwise data object containing a pairwise double doublet.
 *
 * \note This structure is required for archive data-retrieval when millisecond-resolution
 * double timestamps are requested.  In this case the data are returned as pair-wise value
 * plus timestamp entries.
 */
typedef struct
{
  double d1val; /**< the first double value */
  double d2val; /**< the second double value */
} DBLDBL;

/**
 * \struct FLTFLT
 * \brief Defines a TINE pairwise data object containing a pairwise float doublet.
 *
 * \note This is formally synonymous with typedef DPOINT
 */
typedef struct
{
  float f1val; /**< the first float value */
  float f2val; /**< the second float value */
} FLTFLT;

/**
 * \struct DPOINT
 * \brief Defines a TINE pairwise data object containing a pairwise float doublet.
 *
 */
typedef struct
{
  float x; /**< the first float value */
  float y; /**< the second float value */
} DPOINT;

/**
 * \struct NAME16FI
 * \brief Defines a TINE triplet data object containing a 16-character fixed-length string
 * value followed by a float value followed by 4-byte integer value
 *
 * \note This structure has limited applicability and should only be used where structures
 * on compiled with 4-byte boundaries.
 *
 */
typedef struct
{
  char name[16];/**< the string buffer containing the string information */
  float  fval;  /**< the first float value */
  SINT32 ival;  /**< the first long integer value */
} NAME16FI;

/**
 * \struct NAME16II
 * \brief Defines a TINE triplet data object containing a 16-character fixed-length string
 * value followed by two 4-byte integer values
 *
 * \note This structure has limited applicability and should only be used where structures
 * on compiled with 4-byte boundaries.
 *
 */
typedef struct
{
  char name[16];/**< the string buffer containing the string information */
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
} NAME16II;

/**
 * \struct NAME16DBLDBL
 * \brief Defines a TINE triplet data object containing a 16-character fixed-length string
 * value followed by a double doublet
 *
 */
typedef struct
{
  char name[16];/**< the string buffer containing the string information */
  double d1val; /**< the first double value */
  double d2val; /**< the second double value */
} NAME16DBLDBL;

/**
 * \struct NAME32DBLDBL
 * \brief Defines a TINE triplet data object containing a 32-character fixed-length string
 * value followed by a double doublet
 *
 */
typedef struct
{
  char name[32];/**< the string buffer containing the string information */
  double d1val; /**< the first double value */
  double d2val; /**< the second double value */
} NAME32DBLDBL;

/**
 * \struct NAME64DBL
 * \brief Defines a TINE doublet data object containing a 64-character fixed-length string
 * value followed by a double
 *
 */
typedef struct
{
  char name[64];/**< the string buffer containing the string information */
  double dval; /**< the first double value */
} NAME64DBL;

/**
 * \struct NAME64DBLDBL
 * \brief Defines a TINE triplet data object containing a 64-character fixed-length string
 * value followed by a double doublet
 *
 */
typedef struct
{
  char name[64];/**< the string buffer containing the string information */
  double d1val; /**< the first double value */
  double d2val; /**< the second double value */
} NAME64DBLDBL;

/**
 * \struct NAME64DBLDBLDBL
 * \brief Defines a TINE quad data object containing a 64-character fixed-length string
 * value followed by a double triplet
 *
 */
typedef struct
{
  char name[64];/**< the string buffer containing the string information */
  double d1val; /**< the first double value */
  double d2val; /**< the second double value */
  double d3val; /**< the third double value */
} NAME64DBLDBLDBL;

/**
 * \struct FLTINTINT
 * \brief Defines a TINE triplet data object containing a float
 * value followed by a long integer doublet
 *
 */
typedef struct
{
  float  fval;  /**< the first float value */
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
} FLTINTINT;

/**
 * \struct INTFLTINT
 * \brief Defines a TINE triplet data object containing a long integer value
 * followed by a float followed by a second long integer value.
 *
 * \note This structure is synonymous with a 'TDS',i.e. Time-stamp - Data - Status,
 * object used in certain calls from DOOCS.
 */
typedef struct
{
  SINT32 i1val; /**< the first long integer value */
  float  fval;  /**< the first float value */
  SINT32 i2val; /**< the second long integer value */
} INTFLTINT;

typedef struct
{
  float  xval;
  float  yval;
  SINT32 ival;
} XYS;

/**
 * \struct FLTFLTINT
 * \brief Defines a TINE triplet data object containing a float doublet followed by a
 * long integer value.
 *
 */
typedef struct
{
  float f1val; /**< the first float value */
  float f2val; /**< the second float value */
  SINT32 ival; /**< the first long integer value */
} FLTFLTINT;

/**
 * \struct INTINTINT
 * \brief Defines a TINE triplet data object containing a set of three
 * long integer values.
 *
 */
typedef struct
{
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
  SINT32 i3val; /**< the third long integer value */
} INTINTINT;

/**
 * \struct DBLDBLDBL
 * \brief Defines a TINE triplet data object containing a double triplet.
 *
 * \note This structure is required for archive data-retrieval of a number doublet
 * when millisecond-resolution
 * double timestamps are requested.  In this case the data are returned as the number doublet
 * plus timestamp entries.
 */
typedef struct
{
  double d1val; /**< the first double value */
  double d2val; /**< the second double value */
  double d3val; /**< the third double value */
} DBLDBLDBL;

/**
 * \struct DADDRESS
 * \brief Defines a TINE quadruplet data object containing four long integer values
 *
 */
typedef struct
{
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
  SINT32 i3val; /**< the third long integer value */
  SINT32 i4val; /**< the fourth long integer value */
} DADDRESS;

/**
 * \struct DWINDOW
 * \brief Defines a TINE quadruplet data object containing a long integer doublet
 * giving a time range followed by another long integer doublet giving a value range
 *
 */
typedef struct
{
  SINT32 t1; /**< the first (time_t) long integer value */
  SINT32 t2; /**< the second (time_t) long integer value */
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
} DWINDOW;

/**
 * \struct FWINDOW
 * \brief Defines a TINE quadruplet data object containing a long integer doublet
 * followed by a float doublet
 *
 */
typedef struct
{
  SINT32 i1val; /**< the first long integer value */
  SINT32 i2val; /**< the second long integer value */
  float  f1val; /**< the first float value */
  float  f2val; /**< the second float value */
} FWINDOW;

/**
 * \struct FLTINTFLTINT
 * \brief Defines a TINE quadruplet data object containing a float value followed by a
 * long integer followed by a float value followed by a long integer
 *
 */
typedef struct
{
  float  f1val; /**< the first float value */
  SINT32 i1val; /**< the first long integer value */
  float  f2val; /**< the second float value */
  SINT32 i2val; /**< the second long integer value */
} FLTINTFLTINT;

/* TDSI has been deprecated */
typedef struct
{
  SINT32 t;
  float  d;
  SINT16 s;
  SINT16 i;
} TDSI;

/**
 * \struct DSUMMARY
 * \brief Used to reflect the doocs THUMBNAIL type for summarized
 * archive data
 *
 */
typedef struct
{
  double ave; /**< data average over range */
  double dev; /**< standard deviation */
  double max; /**< data maximum */
  double min; /**< data minimum */
  UINT32 ts;  /**< centered timestamp */
  UINT32 tMax;/**< timestamp maximum */
  UINT32 tMin;/**< timestamp minimum */
  UINT32 cnt; /**< number of data points in interval */
} DSUMMARY;

/**
 * \struct DFILTER
 * \brief Defines a TINE quadruplet data object containing a long integer value followed by a
 * float triplet
 *
 */
typedef struct
{
  SINT32 ival;  /**< the long integer value */
  float  f1val; /**< the first float value */
  float  f2val; /**< the second float value */
  float  f3val; /**< the third float value */
} DFILTER;

/**
 * \struct DUSTRING
 * \brief Defines a TINE quintet data object containing a long integer value
 * followed by a float doublet followed by a second long integer value,
 * followed by an 80-character fixed-length string
 *
 */
typedef struct
{
  SINT32 ival;  /**< the first long integer value */
  float  f1val; /**< the first float value */
  float  f2val; /**< the second float value */
  SINT32 tm;    /**< the second long integer value (intpreted frequently as a time_t) */
  char   str[80]; /**< the string buffer */
} DUSTRING;

/**
 * \deprecated
 * \struct DUNAME
 * \brief Defines a TINE quintet data object containing a long integer value
 * followed by a float triplet followed
 * followed by a 16-character fixed-length name
 *
 */
typedef struct
{
  SINT32 ival;  /**< first long integer value */
  float  f1val; /**< the first float value */
  float  f2val; /**< the second float value */
  float  f3val; /**< the third float value */
  char   name[16]; /**< the name buffer */
} DUNAME;

/**
 * \struct hdrSPECTRUM
 * \brief Defines a TINE quintuplet data object containing an 80-character fixed-length string
 * followed by a long integer timestamp followed by a float doublet followed by a long integer
 * value
 *
 * \note This structure defines the header structure for data type CF_SPECTRUM.
 */
typedef struct
{
  char   comment[80];  /**< optional comment */
  SINT32 tm;           /**< long integer value interpreted as a time_t */
  float  s_start;      /**< float value giving the x-axis 'start' */
  float  s_inc;        /**< float value giving the x-axis increment */
  SINT32 status;       /**< long integer value giving the status */
} hdrSPECTRUM;

/**
 * \struct DSPECTRUM
 * \brief Defines a TINE sextuple data object containing an 80-character fixed-length string
 * followed by a long integer timestamp followed by a float doublet followed by a long integer
 * value followed by 4096-value float spectrum array
 *
 * \note format CF_SPECTRUM is used to send/receive a single spectrum whose spectrum length is given
 * by the dArrayLength field of the DTYPE object used.  Any structure following the DSPECTRUM
 * pattern is an acceptible recepticle for the data returned/sent.  That is, the d_spect_array field
 * must simple contain a capacity equal to or greater than that specified in the call.
 *
 * \see the ASPECTRUM type used in conjunction with CF_ASPECTRUM (adjustable spectrum) for alternative
 * use and use with arrays of SPECTRA.
 */
typedef struct
{
  char   comment[80];  /**< optional comment */
  SINT32 tm;           /**< long integer value interpreted as a time_t */
  float  s_start;      /**< float value giving the x-axis 'start' */
  float  s_inc;        /**< float value giving the x-axis increment */
  SINT32 status;       /**< long integer value giving the status */
  float  d_spect_array[SPECTRUM_STD_SIZE]; /**< a static reference to a spectrum buffer */
} DSPECTRUM;

/**
 * \struct USPECTRUM
 * \brief Defines a TINE sextuple data object containing an 80-character fixed-length string
 * followed by a long integer timestamp followed by a float doublet followed by a long integer
 * value followed by a reference to a float array
 *
 * \note format CF_SPECTRUM is used to send/receive a single spectrum whose spectrum length is given
 * by the dArrayLength field of the DTYPE object used. USPECTRUM (universal spectrum) can be used as
 * a cast to the returned data set when a call obtains spectrum data using 'unbound' data
 * (i.e. passing a NULL pointer in the data UNION parameter of the output DTYPE object).  'Unbound' data
 * signals the subsystem to allocate the necessary bytes to hold the requested data.
 *
 * \see the ASPECTRUM type used in conjunction with CF_ASPECTRUM (adjustable spectrum) for alternative
 * use and use with arrays of SPECTRA.
 */
typedef struct
{
  char   comment[80];  /**< optional comment */
  SINT32 tm;           /**< long integer value interpreted as a time_t */
  float  s_start;      /**< float value giving the x-axis 'start' */
  float  s_inc;        /**< float value giving the x-axis increment */
  SINT32 status;       /**< long integer value giving the status */
  float  *d_spect_array; /**< reference allocated to needed capacity */
} USPECTRUM;

/**
 * \struct ASPECTRUM
 * \brief Defines a TINE septuple data object containing an 80-character fixed-length string
 * followed by a long integer timestamp followed by a float doublet followed by a long integer
 * value followed by a spectrum data size, followed by a pointer to a spectrum buffer.
 *
 * \note format CF_ASPECTRUM is used to send/receive arrays of spectra.
 * The dArrayLength field of the DTYPE object used given the number of spectra to send or receive.
 *
 * \see the DSPECTRUM type used in conjunction with CF_DSPECTRUM.
 */
typedef struct
{
  char   comment[80];  /**< optional comment */
  SINT32 tm;           /**< long integer value interpreted as a time_t */
  float  s_start;      /**< float value giving the x-axis 'start' */
  float  s_inc;        /**< float value giving the x-axis increment */
  SINT32 status;       /**< long integer value giving the status */
  UINT32 spectBufferSize; /**< the size of the spectrum array buffer. */
  float  *spectBuffer; /**< a pointer to the spectrum array */
} ASPECTRUM;
/* from doocs: */
#define GSP_STSZ                24
#define GSP_TISZ                8

#define	SPECTRUM_LENGTH_MAX     6291456
#define	GSPECTRUM_LENGTH        SPECTRUM_LENGTH_MAX

/**
* \struct hdrGSPECTRUM
* \brief Defines a TINE mapped data object corresponding to a doocs GSPECTRUM
*
* \note This structure defines the header structure for data type CF_GSPECTRUM.
*/
typedef struct
{
  char comment[80];  /**< optional comment */
  UINT32 id;
  SINT32 tm;         /**< long integer value interpreted as a time_t */
  float  s_start;    /**< float value giving the x-axis 'start' */
  float  s_inc;      /**< float value giving the x-axis increment */
  UINT32 status;     /**< long integer value giving the status */
  int    groups;
  int    group_size;
  float  group_inc;
  int    stat_mask;
  float  statistics[GSP_STSZ];
  int    timing_param[GSP_TISZ];
} hdrGSPECTRUM;

typedef struct
{
  char comment[80];  /**< optional comment */
  UINT32 id;
  SINT32 tm;         /**< long integer value interpreted as a time_t */
  float  s_start;    /**< float value giving the x-axis 'start' */
  float  s_inc;      /**< float value giving the x-axis increment */
  UINT32 status;     /**< long integer value giving the status */
  int    groups;
  int    group_size;
  float  group_inc;
  int    stat_mask;
  float  statistics[GSP_STSZ];
  int    timing_param[GSP_TISZ];
  UINT32 spectBufferSize; /**< the size of the spectrum array buffer. */
  float  *spectBuffer; /**< a pointer to the spectrum array */
} DGSPECTRUM;

/**
 * \struct TImageSourceHeader
 * \brief Defines a sourceHeader header (used in conjuction with an image header)
 *
 */
typedef struct
{
  UINT32 baseTag; /**< base id */
  UINT32 cameraPortId; /**< specific id */
  UINT32 versionTag;    /**< version */
  UINT32 totalLength; /**< total length in bytes (all headers + frame) */
  SINT32 timestampSeconds; /**< image timestamp (UTC seconds) */
  SINT32 timestampMicroseconds; /**< image timestamp (micro seconds) */
  char cameraPortName[64];  /**< camera port name */
} TImageSourceHeader;

/**
 * \struct TImageFrameHeader
 * \brief Defines a canonical image frame header
 *
 * Standard image frame header, 100 bytes long (not divisible by 8!)
 */
typedef struct
{
  SINT32 sourceWidth;
  SINT32 sourceHeight;
  SINT32 aoiWidth;
  SINT32 aoiHeight;
  SINT32 xStart;
  SINT32 yStart;
  SINT32 bytesPerPixel;
  SINT32 effectiveBitsPerPixel;
  SINT32 horizontalBinning;
  SINT32 verticalBinning;
  SINT32 sourceFormat;
  SINT32 imageFormat;
  UINT32 frameNumber;
  UINT32 eventNumber;
  float xScale;
  float yScale;
  float imageRotation;
  float fspare1;
  float fspare2;
  float fspare3;
  SINT32 imageFlags;
  SINT32 ispare1;
  SINT32 ispare2;
  SINT32 ispare3;
  SINT32 appendedFrameSize;
} TImageFrameHeader;

typedef struct
{
  TImageSourceHeader sourceHeader;
  TImageFrameHeader frameHeader;
} hdrIMAGE;

/**
 * \struct DIMAGE
 * \brief Defines a TINE image data object containing an image source header, an image
 * frame header and the image frame itself.
 *
 * \note format CF_AIMAGE is used to send/receive arrays of images, in which case
 *  the dArrayLength field of DTYPE gives the number of images.  If CF_IMAGE is
 *  used, the dArrayLength field gives the size (in bytes) of the image frame to be sent or
 *  received.
 *  Either CF_AIMAGE or CF_IMAGE can be used in structures.
 *
 */
typedef struct
{
  TImageSourceHeader sourceHeader; /**< is the TImageSourceHeader defining the image source and camera parameters */
  TImageFrameHeader frameHeader;   /**< frameHeader is the TImageFrameHeader giving all of the image frame characteristics */
  UINT32 frameBufferSize;          /**< frameBufferSize is the size in bytes of the frame buffer
                                        With CF_IMAGE (non-adjustable) this value is an assigned fixed value giving
                                        the maximum frame size. With CF_AIMAGE (adjustable) this value will be updated
                                        with each frame, gving the size of the current image. */
  BYTE *frameBuffer;               /**< frameBuffer is a reference to the frame buffer holding the current image.
                                        With CF_IMAGE (non-adjustable) this is a fixed reference to a frame buffer.
                                        With CF_AIMAGE (adjustable) this is a volatile reference to the current image
                                        frame.  */
} DIMAGE;

TINE_EXPORT int AssignImageFrameBuffer(DIMAGE *img, BYTE *buf, int buflen);

struct bitfieldFormat
{
  UINT32 bitmask; /**< bitfield mask */
  int bitshift;   /**< bitfield shift */
  char field[16]; /**< bitfield field name */
  struct bitfieldFormat *nxt; /**< Pointer to next field in this bitfield */
};

/**
 * \struct DBITFIELD
 * \brief describes the fields within the named bitfield
 *
 */
typedef struct bitfieldStruct
{
  char srvName[2*CONTEXT_NAME_SIZE+4]; /**< server where the bitfield is registered */
  char name[TAG_NAME_SIZE];   /**< bit field name tag */
  int  fmt;                   /**< TINE format : CF_BYTE, CF_SHORT, CF_LONG */
  struct bitfieldFormat *bfs; /**< structure definition */
  struct bitfieldStruct *next;/**< Pointer to the next element in a linked list.*/
} DBITFIELD;

extern DBITFIELD *gBitFieldList;
TINE_EXPORT DBITFIELD *findBitField(char *srv,char *tag);
void rmvBitFieldList(void);

/**
 * \struct DUNION
 * \brief de-references the data set in a DTYPE object
 *
 */
typedef union
{
  void   *vptr;   /**< void pointer to data */
  char   *cptr;   /**< char pointer to data */
  char   **strptr;/**< pointer to string array */
  UINT8  *bptr;   /**< BYTE pointer to data */
  SINT16 *sptr;   /**< short (int16) pointer to data */
  UINT16 *usptr;  /**< unsigned short pointer to data */
  SINT32 *lptr;   /**< long (int32) pointer to data */
  UINT32 *ulptr;  /**< unsigned long pointer to data */
# ifdef SINT64
  SINT64 *llptr;  /**< long long (int64) pointer to data */
# endif
# ifdef UINT64
  UINT64 *ullptr; /**< unsigned long long pointer to data */
# endif
  float  *fptr;   /**< float pointer to data */
  double *dptr;   /**< double pointer to data */
} DUNION;

/**
 * \struct DTYPE
 * \brief Defines a TINE data object
 *
 */
typedef struct DTYPE
{
  UINT32 dArrayLength; /**< the size of the data set in terms of the format size specified. */
  short  dFormat;      /**< the TINE format of the data set */
  short  xferReason;   /**< the 'reason' (CX_RESPONSE, CX_STALE, etc.) the data have been delivered */
  int    dStamp;       /**< an optional, settable integer stamp which can be used to tag the data set */
  int    sysStamp;     /**< the return system applied data stamp (only valid with sychnronous calls) */
  double dTimeStamp;   /**< is the associated (threadsafe) timestamp of the returned data set, valid only for
                            synchronous calls, as the data object is usually no longer in scope for asynchronous calls. */
  char   dTag[TAG_NAME_SIZE]; /**< an optional data tag used typically to identify user-defined structures */
  DUNION data;         /**< a reference to the data set */
} DTYPE;

/**
 * \brief Gives the size of bytes according to the DTYPE fields
 *
 * \param d is a pointer to the DTYPE object whose size in bytes is desired
 *
 * \return Size of bytes of the DTYPE object given.
 */
TINE_EXPORT int GetFormatSizeInBytesFromDataType(DTYPE *d);
#define MAX_MDA_DIMS 6
/**
 * \struct MDX
 * \brief Multi-Dimension array aXis
 *
 */
typedef struct
{
  int len;        /**< the number of elements in this dimension */
  int reserved;   /**< reserved field */
  char label[16]; /**< the axis label */
  char units[16]; /**< the axis units */
  float min;      /**< the axis value of the initial element */
  float max;      /**< the axis value of the last element */
} MDX;
#define MDX_SIZE 48

/**
 * \struct DMDA
 * \brief Datatype: Multi-Dimensional Array
 *
 */
typedef struct
{
  char cmt[80]; /**< descriptive comment */
  int ndim;     /**< number of dimensions */
  int fmt;      /**< data type format of the data */
  MDX axis[MAX_MDA_DIMS]; /**< axis information array */
  DUNION data;  /**< accompanying data */
} DMDA;
#define MDA_FMT_OFFSET 84
TINE_EXPORT int MakeMDA(DMDA *mda,char *cmt,int fmt,void *data,int ndim,...);
TINE_EXPORT void *GetMDAElemRef(DMDA *mda,...);
int MDAtoDTYPE(DTYPE *d,DMDA *mda,int bufferlength);

extern int gMarshallBooleanAsByte;
void  marshallBoolean(BYTE* values, int length, int dir);
TINE_EXPORT void SetMarshallBooleanAsByte(int value);
TINE_EXPORT int GetMarshallBooleanAsByte(void);

typedef struct FilterStruct
{
  UINT32 filterIdx;
  char filterLinkStr[256];
} FilterStruct;
#define FILTERSTRUCT_SIZE 260
typedef struct FilterTarget
{
  int fcmp;
  int valid;
  double fval;
  char fstr[64];
  struct FilterTarget *nxt;
} FilterTarget;
typedef struct FilterLink
{
  int active;
  int valid;
  int lid;
  int param1;
  int param2;
  BYTE buf[128];
  char key[256];
  char fstr[64];
  struct FilterTarget *tgt;
  struct FilterLink *nxt;
} FilterLink;
extern FilterLink *fltrLnkLst;
enum FilterComparator {
  fl_eq = 0,
  fl_neq,
  fl_lt,
  fl_gt,
  fl_le,
  fl_ge
};
#define TOK_OR "||"
#define TOK_AND "&&"

TINE_EXPORT FilterLink *getFilterLink(FilterLink *fltrLnk, char *key,char *target,int comparator,int param);
TINE_EXPORT FilterLink *getFilterLinkFromParsableString(char *parsableString);
TINE_EXPORT void rmvFilterLink(FilterLink *flnk);
TINE_EXPORT void flushFilterLinks(void);
int getFilterComparator(char *str,int *len);
char *getComparatorString(int cmp);
int dumpFilterLinks(void);

/**
 * \macro DTYPEZERO
 * \brief zeros all fields of given DTYPE variable and sets
 * the format to CF_NULL
 *
 */
# define DTYPEZERO(d) memset(&(d),0,sizeof(d)), (d).dFormat = TDC_CF_NULL

 /**
  * \macro DTYPESET
  * \brief set relevant fields of given DTYPE
  *
  */
#define DTYPESET(d,f,s,t,p) d.dArrayLength=(s), d.dFormat=(f), strncpy(d.dTag,t,TAG_NAME_SIZE), d.data.vptr=(p)

extern TINE_EXPORT char gLastStatusString[STATUS_SIZE];
#define STSSIZE(p) ((p) == 5 ? STATUS_SHORTSIZE : STATUS_SIZE)

#define LNK_ERROR_LCL 0
#define LNK_ERROR_RMT 1
/* lastLnkErrSrc */
/**
 * \brief Gives the signature of the last Link Error
 *
 * If a call to ExecLink() or ExecLinkEx() fails it will return a non-zero
 * TINE error code, which can be interpreted by GetLastLinkError().  It is
 * sometime useful to know whether the return code was locally generated
 * (for instance a time out) or remotely generated (for instance 'illegal property').
 * To this end you can examine the value of lastLnkErrSrc and check it
 * against LNK_ERROR_LCL or LNK_ERROR_RMT
 *
 */
extern short lastLnkErrSrc;

/**
 * \internal
 * \brief callback prototypes
 */
typedef void (*TCBFCNP)(int,int);
typedef void (*XTCBFCNP)(int,int,void *);

/**
 * \internal
 * \brief callback and dispatch function prototypes (normal and extended)
 *
 */
typedef void (*SYSTSKP)(void);
typedef int (*EQMFCNP)(char *,char *,DTYPE *,DTYPE *,short);
typedef int (*XEQMFCNP)(char *,char *,DTYPE *,DTYPE *,short,void *);
typedef void (*EQMTSKP)(void);
typedef void (*XEQMTSKP)(void *);

/* CloseLink() */
/**
 * \brief Cancels an active data link
 *
 * Use this call to cancel an active data link (alias: CloseLink()).  Note that
 * the parameter passed must be the returned link id when the original call to
 * AttachLink() or AttachLinkEx() was made.
 *
 * \param i the link index to be canceled
 *
 * \return 0 upon success or a TINE error code
 *
 * \see AttachLink(), AttachLinkEx()
 *
 * \b Example:
 *
 * \include eg_CloseLink.c
 */
TINE_EXPORT int CloseLink(int i);

/**
 * \def Cancel(i)
 *
 * Identical to CloseLink(i)
 */
#define Cancel(i) CloseLink(i)

extern int SystemLocalDataTimeSynchronization;

/* SystemRunning */
/**
 * \brief Determines whether a server has been successfully initialized or not.
 *
 * A server can query this flag at any stage to determine if all initialization step have been
 * successfully completed or not.  You should never set this flag yourself.
 *
 * \b Default: FALSE
 *
 */
extern int SystemRunning;
/* SystemSuppressHeartbeatNofification */
/**
 * \brief Determines whether CM_DATACHANGE data links signal incoming data by calling the
 * corresonding callback routine even when the incoming data is a HEARTBEAT notification.
 *
 * A client may wish to receive data notification events only when the incoming data have
 * changed or the data link has an error, and not receive and respond to heartbeat notifications.
 * Setting this global variable to TRUE will archieve this aim.
 *
 * \param value determines whether heartbeat notification will be suppressed or not
 * \b Default: FALSE
 *
 * \see SetNotificationTolerance()
 */
TINE_EXPORT void SetSuppressHeartbeatNotification(int value);
extern int SystemSuppressHeartbeatNotification;
/* SystemPresetMemory */
/**
 * \brief Determines whether data requests preset the calling buffer to zero before issuing the call
 *
 * If a server always returns the amount of data requested, then presetting the calling buffer
 * contents can be entirely superfluous, and in the case of large data call, can result in a
 * non-negligible increase in the server CPU load. On the other hand, if a server returns less than the
 * amount of data asked for, or the caller asks for an array of fixed string length data, there
 * could be unforseen side-effects if the calling buffer has not been zeroed out.
 *
 * \b Default: TRUE
 *
 */
extern TINE_EXPORT int SystemPresetMemory;

/* SystemReset() */
/**
 * \brief Flushes all entry tables and returns all counters and semaphores to their initial state.
 *
 * Useful under some circumstances, for instance following program suspension or interrupt
 * due to CNTL-C, etc.
 *
 * \param level is reserved for future use.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \note SystemReset is macro #defined as _SystemReset(). If additional 'reset' procedures
 * need to be incorporated, one can re-\#define the SystemReset() macro.
 *
 * \see _SystemInit(), _SystemCycle(),
 *
 * \b Alias: SystemReset
 *
 * \b Example:
 *
 * \include eg_SystemReset.c
 */
TINE_EXPORT int _SystemReset(int level);
#ifndef SystemReset
#define SystemReset(i) _SystemReset(i)
#endif

TINE_EXPORT void SetSystemReinitIdleTime(int idleTimeInSeconds);

/* FecIsInitialized() */
/**
 * \brief Determines whether the FEC process has initialized services and all known equipment modules.
 *
 * In addition to equipment modules being initialized, any save/restore values must be restored before
 * this routine returns TRUE;
 *
 * \return TRUE if all equipment modules have run through their initialization routines AND all
 * save/restore values have been restored.
 *
 * \see _SystemInit(), ServerIsInitialized(),
 *
 */
TINE_EXPORT int FecIsInitialized(void);

/* ServerIsInitialized() */
/**
 * \brief Determines whether the server has initialized;
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return TRUE if the given equipment modules has run through its initialization routine.
 *
 * \see _SystemInit(), FecIsInitialized(),
 *
 */
TINE_EXPORT int ServerIsInitialized(const char *eqm);

/* SystemCycle() */
/**
 * \brief Required call for the TINE engine to function property.
 *
 * Should be called within an infinite loop.  Prior to the start of this loop, all other
 * device server registration should occur.
 * Clients with asynchronous links must see to it that _SystemCycle() is called regularly
 * (unless SystemDelay(-1) is used to block the main process).
 *
 * \param chkcmd is used to signal the presence of a local
 * command interpreter.  Developers wishing to bypass the built-in command interpreter
 * in favor of another should pass FALSE for this argument
 *
 * \note SystemCycle does will in general NOT monopolize the CPU as for most Operating
 * systems, blocking sockets are used and the TINE Engine will
 * itself block on select() for a duration not greater than the SystemPollingRate.
 *
 * \return 0 under normal circumstances. A non-zero return values indicates an immediate
 * need to re-enter (i.e. following a complete cycle, there are still incomplete data
 * transfer pending).
 *
 * \note SystemCycle is macro #defined as _SystemCycle(). If additional 'cycle' procedures
 * need to be incorporated, one can re-\#define the SystemCycle() macro.
 *
 * \see _SystemInit(), _SystemDelay(),
 *
 * \b Alias: SystemCycle
 *
 * \b Example:
 *
 * \include eg_main.c
 */
TINE_EXPORT int _SystemCycle(int chkcmd);
#ifndef SystemCycle
#define SystemCycle(i) _SystemCycle(i)
#endif

/* SystemDelay() */
/**
 * \brief Blocks any activity inside a TINE task or equipment modules for the
 * specified amount of time.
 *
 * The TINE engine will continue to run and process data which is not related to
 * 1) the equipment module being blocked if _SystemDelay() is called inside an
 * equipment module or 2) the background task if _SystemDelay() is called inside a
 * backgroung task.
 *
 * \param msec is the amount of time in milliseonds to suspend the calling routine.
 *
 * \note The TINE engine is designed to run single-threaded.  You are free to use
 * threads and you can specify multi-threaded backgound tasks if you want the TINE
 * engine itself to explicitly use threads.  In such cases, calls to 'sleep()' would
 * have their intended effect.  Under single-threaded conditions (i.e. VMS or
 * you're avoided threads because they complicate things) you should not use 'sleep()'
 * since you will block then entire TINE engine.  Use _SystemDelay() instead if you want
 * to wait at a specific spot in your code until some action has finished.
 *
 * \note SystemDelay is macro #defined as _SystemDelay(). If additional 'delay' procedures
 * need to be incorporated, one can re-\#define the SystemDelay() macro.
 *
 *
 * \see _SystemCycle(), _SystemInit(),
 *
 * \b Alias: SystemDelay
 *
 * \b Example
 *
 * \include eg_SystemDelay.c
 *
 */
TINE_EXPORT void _SystemDelay(int msec);
#ifndef SystemDelay
#define SystemDelay(i) _SystemDelay(i)
#endif

/* SystemInit() */
/**
 * \brief Required call for the TINE server engine to function property.
 *
 * Should be called prior to the start of the main TINE server engine.
 * Any deviations from default system parameters should be made before any call to
 * _SystemInit().  This might include server settings which fix the size of the local
 * connection table entries, the maximum date transport size, etc.
 * Note that _SystemInit() will begin be allocating space for its connection tables, so
 * modifying these values away from their defaults will have no effect after _SystemInit()
 * has been called.  Also, if the FEC Name is given by calling RegisterFecInformation() or
 * RegisterFecNameEx(), then this should also preceed a call to _SystemInit().
 *
 *
 * \param chkcmd is used only for UNIX servers and if TRUE will determine if the server is
 * running in the foreground and if so, will check the stdin for console commands at the
 * keyboard.
 *
 * \note _SystemInit() does not and should not be called if you are writing a 'client-only'
 * application.  _SystemInit() will attempt to register various server settings which
 * play no role, and will attempt to initiaze server sockets which could lead to confusing
 * situations on the local computer.  _SystemInit() does not initialize any client-side
 * functioniality.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted
 * by a call to GetLastLinkError().
 *
 * \note SystemInit is macro #defined as _SystemInit(). If additional 'init' procedures
 * need to be incorporated, one can re-\#define the SystemInit() macro.
 *
 * \see _SystemCycle(), _SystemDelay(),
 *
 * \b Alias: SystemInit
 *
 * \b Example:
 *
 * \include eg_main.c
 */
TINE_EXPORT int _SystemInit(int chkcmd);
#ifndef SystemInit
#define SystemInit(i)  _SystemInit(i)
#endif

/* _SystemStop() */
/**
 * \brief Free all system resources and optionally exit.
 *
 * Calling this routine will close down all client-side connections, call any
 * registered equipment module exit routines, stop all registered background tasks,
 * free up all system memory, close all system sockets and pipes, and call exit(0)
 * if the 'exitOnFree' parameter is TRUE.
 *
 * \param exitOnFree if TRUE will call exit(0) after freeing resources.  Otherwise
 * the process will remain active (for an additional startup sequence, for example).
 *
 * \note SystemStop is macro #defined as _SystemStop(). If additional 'stop' procedures
 * need to be incorporated, one can re-\#define the SystemStop() macro.
 */
TINE_EXPORT void _SystemStop(int exitOnFree);
#ifndef SystemStop
#define SystemStop(i)  _SystemStop(i)
#endif

void _SystemReInit(void);
TINE_EXPORT void SystemReInit(int idleTime);

/* SystemScheduleProperty() */
/**
 * \brief Schedules the given property for immediate delivery to all attached clients.
 *
 * When clients are attached to a particular property, they have specified a polling rate, which
 * defines the maximum latency for receiving updates at the client side.  This is usually fine.
 * However, if a server knows that important data have changed and which properties depend on these
 * data, it can signal the scheduler to call the given properties immediately and send the data to
 * all attached callers, regardless of their registered polling rates.  In this way, a server can
 * signal an event to its attached clients.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is (comma separated) list of properties which are to be scheduled.
 * If the supplied property should only be scheduled if the associated contracts call the property
 * with a specific device then the device can also be supplied in the property string as in
 * "property<device>".
 * Note: the device specification applies only to 'remote' contracts (not to local history or alarm
 * system contracts).
 *
 * \return 0 upon success or a TINE error code.
 *
 * \b Alias: SystemScheduleProperty()
 *
 * \b Example:
 *
 * \include eg_SystemScheduleProperty.c
 */
TINE_EXPORT int _SystemScheduleProperty(char *eqm, char *prp);
#ifndef SystemScheduleProperty
#define SystemScheduleProperty(a,b) _SystemScheduleProperty(a,b)
#endif

/* SystemSchedulePropertyEx() */
/**
 * \brief Schedules the given property for immediate delivery to all attached clients according
 * scope specified.
 *
 * When clients are attached to a particular property, they have specified a polling rate, which
 * defines the maximum latency for receiving updates at the client side.  This is usually fine.
 * However, if a server knows that important data have changed and which properties depend on these
 * data, it can signal the scheduler to call the given properties immediately and send the data to
 * all attached callers, regardless of their registered polling rates.  In this way, a server can
 * signal an event to its attached clients. In this extended call, the caller can specify the 'scope'
 * of the call : CA_NETWORK (all listening clients), CA_HIST (the local history subsystem) or
 * CA_ALARM (the local alarm server).
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is (comma separated) list of properties which are to be scheduled.
 * If the supplied property should only be scheduled if the associated contracts call the property
 * with a specific device then the device can also be supplied in the property string as in
 * "property<device>".
 * \param scope is any of CA_NETWORK, CA_HIST, or CA_ALARM ORed together.
 * Note: a call to _SystemScheduleProperty() is equivalent to supplying all bits.
 *
 * \return 0 upon success or a TINE error code.
 *
 * \b Alias: SystemSchedulePropertyEx()
 *
 * \b Example:
 *
 * \include eg_SystemScheduleProperty.c
 */
TINE_EXPORT int _SystemSchedulePropertyEx(char *eqm, char *prp,int scope);
#ifndef SystemSchedulePropertyEx
#define SystemSchedulePropertyEx(a,b,c) _SystemSchedulePropertyEx(a,b,c)
#endif

/**
 * \brief Sets the 'lazy' flag for scheduling properties
 *
 * When a property is scheduled, the delivery can either occur immedidately (default) and
 * thereby invoke the TINE scheduler within the call to SystemScheduleProperty() or occur
 * within the next normal delivery cycle (lazy), which could add a latency of up to a clock
 * tick to the delivery.  This call sets the 'lazy' flag deteriming this behavior.
 *
 * \param value will be interpreted in a boolean manner.  If TRUE (non-zero) then calls to
 * SystemScheduleProperty() will mark the associated properties for scheduling, which will
 * duly occur during the next call to the scheduler (usually within a clock tick) by the TINE
 * kernel.  If FALSE (default) then scheduled properties will immediately invoke the TINE
 * scheduler.
 *
 * \see _SystemScheduleProperty(), _SystemSchedulePropertyEx(), GetSystemSchedulePropertyLazy()
 */
TINE_EXPORT void SetSystemSchedulePropertyLazy(int value);

/**
 * \brief Gets the 'lazy' flag for scheduling properties
 *
 * When a property is scheduled, the delivery can either occur immedidately (default) and
 * thereby invoke the TINE scheduler within the call to SystemScheduleProperty() or occur
 * within the next normal delivery cycle (lazy), which could add a latency of up to a clock
 * tick to the delivery. This call reads the 'lazy flag'.
 *
 * \return the current setting of the schedule property 'lazy' flag.
 *
 * \see _SystemScheduleProperty(), _SystemSchedulePropertyEx(), SetSystemSchedulePropertyLazy()
 */
TINE_EXPORT int GetSystemSchedulePropertyLazy(void);

/* _SystemFireEvent() */
/**
 * \brief Schedules an event for the given property for immediate delivery to all attached clients.
 *
 * Clients can attach to a particular property and reqest event notification only by applying the
 * CM_EVENT polling mode.  When a server schedules the property by calling either this routine of
 * _SystemScheduleProperty() then the property handler is called and the attached clients receive the
 * property's data (if any) and an event notification.  This call is fully equivalent to
 * _SystemScheduleProperty().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is (comma separated) list of properties which are to be scheduled.
 *
 * \return 0 upon success or a TINE error code.
 *
 * \see  _SystemScheduleProperty()
 *
 * \b Alias: SystemFireEvent()
 *
 * \b Example:
 *
 * \include eg_SystemScheduleProperty.c
 */
TINE_EXPORT int _SystemFireEvent(char *eqm, char *prp);
#ifndef SystemFireEvent
#define SystemFireEvent(a,b) _SystemFireEvent(a,b)
#endif

extern int gSchedulerInterval;
/* SetSchedulerInterval() */
/**
 * \brief Sets the system scheduler interval.
 *
 * The TINE kernel will call the scheduler regularly in order to service the current
 * contract list.  Under some circumstances it may be wished to control this interval.
 * For instance, if it is known a priori that a server will need to respond to heavy
 * synchronous polling while at the same time maintaining a lengthy contract list, it
 * is desireable to limit the number of passes through the scheduler per second (as
 * this will traverse the contract list with each pass).  Thus setting the scheduler
 * interval to some reasonable number (e.g. 50 or 100 milliseconds) might significantly
 * reduce the CPU load on the server.  Note however, that this will impose a latency
 * on all synchronous requests!  If the value for the schedular interval is set to 0
 * (the default value) or less, then the scheduler will be called at often as
 * necessary, usually at the system polling interval (as determined by the most eager
 * contract) and/or following a client request.
 *
 * \param value is the desired scheduler interval in milliseconds. (default = 0).
 *
 * \see  GetSchedulerInterval()
 *
 */
TINE_EXPORT void SetSchedulerInterval(int value);

/* GetSchedulerInterval() */
/**
 * \brief Gets the system scheduler interval.
 *
 * The TINE kernel will call the scheduler regularly in order to service the current
 * contract list.  Under some circumstances it may be wished to control this interval.
 * For instance, if it is known a priori that a server will need to respond to heavy
 * synchronous polling while at the same time maintaining a lengthy contract list, it
 * is desireable to limit the number of passes through the scheduler per second (as
 * this will traverse the contract list with each pass).  Thus setting the scheduler
 * interval to some reasonable number (e.g. 50 or 100 milliseconds) might significantly
 * reduce the CPU load on the server.  Note however, that this will impose a latency
 * on all synchronous requests!  If the value for the schedular interval is set to 0
 * (the default value) or less, then the scheduler will be called at often as
 * necessary, usually at the system polling interval (as determined by the most eager
 * contract) and/or following a client request.
 *
 * \return the current setting for this value.
 *
 * \see  SetSchedulerInterval()
 *
 */
TINE_EXPORT int GetSchedulerInterval(void);

/* SystemGetStartupCommand() */
/**
 * \brief Returns the command line used to start this process
 *
 * If available the command line used to start the process is returned
 * including all command line argurments.
 *
 * \return the command line used to start this process
 *
 */
TINE_EXPORT char *SystemGetStartupCommand(void);

/* SystemGetStartupDirectory() */
/**
 * \brief Returns the working directory in use when this process started
 *
 * \return the working directory in use when this process started
 *
 */
TINE_EXPORT char *SystemGetStartupDirectory(void);

/* SystemGetProcessId() */
/**
 * \brief Returns the process id of the running application if
 * available.
 *
 * \return the process id of the running application
 *
 */
TINE_EXPORT UINT32 SystemGetProcessId(void);

TINE_EXPORT void _SystemFreeResources(int exitOnFree);
#ifndef SystemFreeResources
#define SystemFreeResources(e) _SystemFreeResources(e)
#endif

extern int gAutoAdjustWorkAreaSize;
/**
 * \brief Sets 'auto-adjust work area size' (based on number of registered properties)
 *
 * \param value is the desired setting (default = TRUE).
 *
 */
TINE_EXPORT void SetAutoAdjustWorkAreaSize(int value);

/* SetSystemUseDataProtection() */
/**
 * \brief Sets the data protection flag to the value given
 *
 * Multi-threaded servers using background tasks can ensure the thread safety
 * of any data sets used in both an equipment module's background thread and dispatch
 * function handler by explicit coding.  As an alternative, calling this routine at
 * initialization time with a non-zero value will automatically generate a data protection
 * mutex which will prevent concurrent exectution of a module's background task and
 * dispatch handler.
 *
 * \param value is the desired setting (default = FALSE).
 *
 */
TINE_EXPORT void SetSystemUseDataProtection(int value);

/* GetSystemUseDataProtection() */
/**
 * \brief Gets the data protection flag to the value given
 *
 * Multi-threaded servers using background tasks can ensure the thread safety
 * of any data sets used in both an equipment module's background thread and dispatch
 * function handler by explicit coding.  As an alternative, calling this routine at
 * initialization time with a non-zero value will automatically generate a data protection
 * mutex which will prevent concurrent exectution of a module's background task and
 * dispatch handler.
 *
 * \return the current setting;
 *
 */
TINE_EXPORT int GetSystemUseDataProtection(void);

/* SetWorkAreaSize() */
/**
 * \brief Sets the server-side work area size which is coupled to the maximum transport size for
 * supported by a single-threaded server.
 *
 * Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
 * of the caller's requested data size, the other buffer being in the case of single threaded servers,
 * a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
 * size needs to be adjusted accordingly prior to server initialization.  Note that multi-threaded
 * servers will dynamically allocate both buffers on a call by call basis.
 *
 * \param size is the size in bytes to be reserved for the server's work area and hence the maximum
 *  allowed transport size for any property supported by the server.
 *  (default: 64 KBytes on most operating systems).
 *
 * \return The current setting for the server's work area.
 *
 */
TINE_EXPORT UINT32 SetWorkAreaSize(UINT32 size);
/* GetWorkAreaSize() */
/**
 * \brief Gets the server-side work area size which is coupled to the maximum transport size for
 * supported by a single-threaded server.
 *
 * Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
 * of the caller's requested data size, the other buffer being in the case of single threaded servers,
 * a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
 * size needs to be adjusted accordingly prior to server initialization.  Note that multi-threaded
 * servers will dynamically allocate both buffers on a call by call basis.
 *
 * \return The current setting for the server's work area.
 *
 */
TINE_EXPORT UINT32 GetWorkAreaSize(void);
TINE_EXPORT UINT32 setWorkAreaSize(UINT32 size);

/* SetFeclogPath() */
/**
 * \brief Sets the fec log repository.
 *
 * \param path is the desired path for the fec.log file
 *
 * \return 0 upon success or 'string_too_long'
 *
 */
TINE_EXPORT int SetFecLogPath(char *path);

/* GetFeclogPath() */
/**
 * \brief Gets the current setting for the fec log repository.
 *
 * \return A string pointing to the current setting for the fec log repository
 *
 */
TINE_EXPORT char *GetFecLogPath(void);

/* GetFecHome() */
/**
 * \brief Gets the current setting for the fec home database repository.
 *
 * \return A string pointing to the current setting for the fec home database repository
 *
 */
TINE_EXPORT char *GetFecHome(void);

/* SetFecHome() */
/**
 * \brief Sets the current fec home database repository.
 *
 * \return A string pointing to the current setting for the fec home database repository
 *
 * \note If the server has already been initialized, this function will return an error.
 *
 * \return 0 upon success or a TINE error code.
 */
TINE_EXPORT int SetFecHome(char *fecHomePath);

/* GetSystemUserName() */
/**
 * \brief Gets the current value for the application user.
 *
 * \return A string pointing to the current value of the application user, typically
 * the logged in user or, if the application is a FEC, the FEC name.
 *
 */
TINE_EXPORT char *GetSystemUserName(void);

/* SetSystemUserName() */
/**
 * \brief Sets the current value for the application user.
 *
 * \param usr The desired appliation's user name.
 *
 * \return 0 if successful, otherwise a TINE return code.
 *
 */
TINE_EXPORT int SetSystemUserName(char *usr);

/**
 * \internal
 *
 */
TINE_EXPORT char *GetRPCLastError(short cc);

/* GetLastLinkError() */
/**
 * \brief The error string associated with the input error number
 *
 * \param cc is the completion code for which the error string is desired.  If cc is in fact
 * the last Link Error then the returned error string is used, otherwise the system error
 * string for the completion code in question is returned
 * \param errstr is the error string buffer which is to contain the string description of the
 * the cc parameter.  A pointer to the same string buffer is returned (A null pointer returns "[null error string]").
 *
 * \return The error string associated with the input error number.
 *
 * \see GetCompletionStatus()
 *
 * \b Example:
 *
 * \include eg_GetRPCLastError.c
 */
TINE_EXPORT char *GetLastLinkError(short cc, char  *errstr);

/* SetEqmCompletion() */
/**
 * \brief Sets the error string to accompany the current server call
 *
 * For user-defined errors codes (equal to or above 512) the accompanying error string
 * should be defined by calling this routine prior to returning from the equipment module
 * call.  When system error codes such as 'illegal_format' are used, a call to SetEqmCompletion()
 * is not necessary.
 *
 * \param eqmName the local equipment module name whose completion code is to be set
 * \param errStr is the 96-character string which is to accompany the outgoing error code.
 *
 * \b Example:
 *
 * \include eg_SetRPCCompletion.c
 */
TINE_EXPORT void SetEqmCompletion(char *eqmName,char *errStr);

/**
 * \internal
 */
TINE_EXPORT void SetRPCCompletion(char *errstr);

/**
 * \internal
 */
TINE_EXPORT void SetRPCRedirection(char *rdrSrv,char *rdrPrp);

/* SetRedirectionParameters() */
/**
 * \brief Sets the redirection string to accompany the current server call
 *
 * This routine can be used when Property Query Functions are in play.  In such cases the
 * equipment module is responsible for deciding if a call should be redirected to another
 * server or not.  If so, the server can call this routine to set the redirection information.
 *
 * \param eqmName the local equipment module name whose redirection parameters should be set
 * \param rdrCon is the device context to which the call should be redirected
 * \param rdrSrv is the device server to which the call should be redirected
 * \param rdrPrp is the device property to which the call should be redirected. (Note that
 * property length is restricted to 32 characters for redirected calls).
 * \param rdrDev is the device name to which the call should be redirected (Note that
 * device length is restricted to 32 characters for redirected calls).
 *
 */
TINE_EXPORT void SetRedirectionParameters(char *eqmName,char *rdrCnt,char *rdrSrv,char *rdrDev,char *rdrPrp);

extern int (*lookupRedirectionNameStub)(char *eqm,char *prpName,char *devName);

/* SetlookupRedirectionNameStub() */
/**
 * \brief Sets a stub to an existing lookupRedirectionNameStub function for cases
 * where a Property Query Function is in force and the server needs to redirect
 * calls.
 *
 * A lookupRedirectionNameStub function should return server_redirection if the
 * property and device combination is to be redirected.  It should also
 * call the SetRPCRedirection() routine to establish the redirection string.
 *
 * \param fcn is the address of the lookupRedirectionNameStub function.
 */
TINE_EXPORT void SetlookupRedirectionNameStub(int (*fcn)(char *eqm,char *prpName,char *devName));

/* ExecLink() */
/**
 * \brief Executes a synchronous link
 *
 * Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or
 * a timeout has ensued. ExecLink() differs from ExecLinkEx() in that the default timeout
 * of 1000 msec is assumed
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only
 * be successful when the client and server are running together in the same process).
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any set of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.)
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \see ExecLocalLink(), ExecLinkEx(), DTYPE
 *
 *
 * \b Example:
 *
 * \include eg_ExecLink.c
 */
TINE_EXPORT int ExecLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access);

/* ExecLinkEx() */
/**
 * \brief Executes a synchronous link (Extended call).
 *
 * Synchronous data exchange. ExecLinkEx() does not complete until the data transfer has completed or
 * a timeout has ensued. ExecLinkEx() differs from ExecLink() in that the user may specify a timeout
 * period for the call to complete.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only
 * be successful when the client and server are running together in the same process).
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori
 * that a call takes a long time to complete than you can set this value accordingly.
 * ExecLink() makes a call to ExecLinkEx() with the timeout parameter set to 1000 msec.
 * Note that by default, a 'retry' will automatically be attempted in case the call does not complete in the
 * time specified.  This 'retry' can add significantly to the total call completion time.  For instance if the
 * target is simply not running, the call will take approximately two times the timeout specified (plus an
 * additional 100 msec) before returning.  If this behavior is NOT desired, then the 'access' parameter should
 * be ORed with the CA_NORETRY flag.  Also note that lossy networks can easily lead to packet loss in the case
 * of a (default) UDP transaction.  In a similar sense, a busy server might also cause a TCP transaction to fail
 * to complete within the time specified.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \see ExecLocalLink(), ExecLink(), DTYPE
 *
 * \b Example:
 *
 * \include eg_ExecLinkEx.c
 */
TINE_EXPORT int ExecLinkEx(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,UINT16 timeout);

/* AttachLink() */
/**
 * \brief Initiates an asynchronous link
 *
 * Asynchronous data exchange. AttachLink() returns immediately with a positive link index if the
 * device name can be resolved and there are sufficient resources on the client side.  Otherwise,
 * a call to AttachLink() returns a negative completion code which can be interpreted with GetLastLinkError().
 *
 * AttachLink() differs from AttachLinkEx() in that the callback will always return the link identifier
 * as the callback id parameter in the callback routine.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only
 * be successful when the client and server are running together in the same process).
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * Note that the input data set is used to specify the contract!  This means that the data values (in addition
 * to the data array length and data format) will be cached at the server side in the case of persistent links
 * with input data.
 * \param access is the data access mode.  This can be any of the TINE access codes ORed together (CA_READ,
 * CA_WRITE, etc.)
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers,
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see AttachLinkEx(), DTYPE
 *
 * \b Example:
 *
 * \include eg_AttachLink.c
 *
 */
TINE_EXPORT int AttachLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
               int pollingRate,void (*callback)(int,int),int mode);

/* AttachLinkEx() */
/**
 * \brief Initiates an asynchronous link
 *
 * Asynchronous data exchange. AttachLinkEx() returns immediately with a positive link index if the
 * device name can be resolved and there are sufficient resources on the client side.  Otherwise,
 * a call to AttachLinkEx() returns a negative completion code which can be interpreted with GetLastLinkError().
 *
 * AttachLinkEx() differs from AttachLink() in that the caller can supply a callback identifier to be
 * returned in the callback function.  In lieu of this, the callback will return the link identifier.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only
 * be successful when the client and server are running together in the same process).
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * Note that the input data set is used to specify the contract!  This means that the data values (in addition
 * to the data array length and data format) will be cached at the server side in the case of persistent links
 * with input data.
 * \param access is the data access mode.  This can be any of the TINE access codes ORed together
 * (CA_READ, CA_WRITE, etc.).
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers,
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback function.
 * Pass UNASSIGNED_CALLBACKID in order to receive the link ID in the supplied callback function.
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see AttachLink(), DTYPE
 *
 * \b Example:
 *
 * \include eg_AttachLinkEx.c
 */
TINE_EXPORT int AttachLinkEx(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
               int pollingRate,void (*callback)(int,int),int mode,UINT32 callbackID);

/* AttachLinkEx2() */
/**
 * \brief Initiates an asynchronous link
 *
 * Asynchronous data exchange. AttachLinkEx2() returns immediately with a positive link index if the
 * device name can be resolved and there are sufficient resources on the client side.  Otherwise,
 * a call to AttachLinkEx() returns a negative completion code which can be interpreted with GetLastLinkError().
 *
 * AttachLinkEx2() differs from AttachLinkEx() and AttachLink() in that the caller can supply a callback identifier
 * and a reference pointer to be returned in the callback function.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only
 * be successful when the client and server are running together in the same process).
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * Note that the input data set is used to specify the contract!  This means that the data values (in addition
 * to the data array length and data format) will be cached at the server side in the case of persistent links
 * with input data.
 * \param access is the data access mode.  This can be any of the TINE access codes ORed together
 * (CA_READ, CA_WRITE, etc.).
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers,
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, the second argument (cc) gives the completion code, and the third
 * argument is the user supplied reference pointer.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback function.
 * Pass UNASSIGNED_CALLBACKID in order to receive the link ID in the supplied callback function.
 * \param reference is a user-supplied pointer which will be returned in the callback routine.  This can be a very
 * useful means for de-referencing the callback.
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see AttachLink(), AttachLinkEx(),DTYPE
 *
 * \b Example:
 *
 * \include eg_AttachLinkEx2.c
 */
TINE_EXPORT int AttachLinkEx2(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*callback)(int,int,void *),int mode,UINT32 callbackID,void *reference);

/* ModifyLinkAttributes() */
/**
 * \brief Allows the caller to assign new link attributes to an active link
 *
 * A TINE client by making this call can re-assign a callback function, access mode or polling rate currently in place for
 * an active link.
 *
 * \param i is the link identifier for the link in question (the returned identifier when the call to
 * AttachLink() or AttachLinkEx() was originally made
 * \param access is the data access mode.  This can be any combination of TINE access code ORed
 * together (e.g. CA_READ|CA_WRITE).
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers,
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback funtion
 *
 * \return 0 if successful, otherwise a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see AttachLink(), AttachLinkEx(), DTYPE
 *
 */
TINE_EXPORT int ModifyLinkAttributes(int i,short access,int pollingRate,
               void (*callback)(int,int),int mode,UINT32 callbackID);

/* SetNotificationTolerance() */
/**
 * \brief Allows the caller to apply a tolerance pertaining to link notification.
 *
 * When a TINE client subscribes to data using CM_DATACHANGE mode, the data is sent from server to client
 * when the data change without regard to any applied tolerance (zero tolerance).  The client application
 * may nonetheless wish to be notified only when the data have changed outside a given tolerance.
 * This can be achieved by calling this routine.  The same applies to CM_TIMER mode, where the data are
 * sent server-to-client according to the schedule.  In the latter case, all data comparisons are made
 * entirely at the client side.  Note that if this feature is desired in CM_TIMER mode, the CM_NODUPLICATES
 * switch must be applied to the mode parameter in the call to AttachLink() or AttachLinkEx().  This
 * switch will be appended automatically if SetSuppressHeartbeatNotification(TRUE) is called.
 *
 * \param linkId is the link identifier for the link in question (the returned identifier when the call to
 * AttachLink() or AttachLinkEx() was originally made
 * \param toleranceAbsolute if non-zero is a tolerance in the data units of the in-coming data.  It is
 * checked against irrespective of the current data values.
 * \param tolerancePercent if non-zero is a tolerance expressed as a percent, and refers to the previous
 * data set.
 *
 * If both an absolute tolerance and a percent tolerance are used, the absolute value of both contributions
 * are added together to define the total tolerance which should be checked against.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see AttachLink(), AttachLinkEx()
 *
 */
TINE_EXPORT int SetNotificationTolerance(int linkId, float toleranceAbsolute, float tolerancePercent);

/* SetNotificationToleranceEx() */
/**
 * \brief Allows the caller to apply a tolerance pertaining to link notification (extended call).
 *
 * When a TINE client subscribes to data using CM_DATACHANGE mode, the data is sent from server to client
 * when the data change without regard to any applied tolerance (zero tolerance).  The client application
 * may nonetheless wish to be notified only when the data have changed outside a given tolerance.
 * This can be achieved by calling this routine.  The same applies to CM_TIMER mode, where the data are
 * sent server-to-client according to the schedule.  In the latter case, all data comparisons are made
 * entirely at the client side.  Note that if this feature is desired in CM_TIMER mode, the CM_NODUPLICATES
 * switch must be applied to the mode parameter in the call to AttachLink() or AttachLinkEx().  This
 * switch will be appended automatically if SetSuppressHeartbeatNotification(TRUE) is called.
 *
 * \param linkId is the link identifier for the link in question (the returned identifier when the call to
 * AttachLink() or AttachLinkEx() was originally made
 * \param toleranceAbsolute if non-zero is a tolerance in the data units of the in-coming data.  It is
 * checked against irrespective of the current data values.
 * \param tolerancePercent if non-zero is a tolerance expressed as a percent, and refers to the previous
 * data set.
 * \param offset specifies the array offset to be used in data comparison (DEFAULT: 0).  If the given offset
 * is less then 0 or greater than the length of the acquired data array (i.e. > 1 if a single value is acquired)
 * it will be automatically set to 0.
 * \param length specifes the length of the data array (beginning at the given offset) to be used in data
 * comparison (DEFAULT: length of the acquire data array, or 1 if a single value is acquired). If the given
 * value exceeds the array bounds, then it will be adjusted.
 * \param flags is one of DTF_SUPPRESS or DTF_NOTIFY and specifies the desired form of the the
 * 'out of tolerance' notification (DEFAULT: DTF_SUPPRESS).  If DTF_SUPPRESS is specified, then callbacks will
 * be suppressed unless there is an error or the most recent data set is out of tolerance.
 * If DTF_NOTIFY is specified, then callbacks will not be suppressed. However, if the most recent data set
 * is out of tolerance, then an otherwise successful callback completion code will be flagged as
 * CE_SENDDATA + out_of_tolerance.  The caller can then make use of this form of 'out of tolerance' signal to
 * suit his needs.
 *
 * If both an absolute tolerance and a percent tolerance are used, the absolute value of both contributions
 * are added together to define the total tolerance which should be checked against.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see AttachLink(), AttachLinkEx()
 *
 */
TINE_EXPORT int SetNotificationToleranceEx(int linkId, float toleranceAbsolute, float tolerancePercent, int offset, int length, int flags);

/* recvNetGlobal() */
/**
 * \brief Initiates a net globals data link
 *
 * A TINE client requests to receive network global parameters via this routine.  Note that the
 * client must reside on a network which itself is receiving broadcast data or onto which multicasts
 * are being routed in order for this call to take effect.
 *
 * recvNetGlobal() differs from recvNetGlobalEx() in that the callback will return the link identifier
 * in the callback function.
 *
 * \param keyword is the name of the network global which is desired.
 * If the globals keyword exists on a specific context, then the
 * keyword given should be prefixed with the context.  For example "/PETRA/Energy".
 * If the keyword is entered without a prefix then the 'SITE' globals are assumed (e.g. "PEENERGY").
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see recvNetGlobalEx(), sendNetGlobal(), DTYPE
 *
 * \b Example:
 *
 * \include eg_recvNetGlobal.c
 */
TINE_EXPORT int recvNetGlobal(const char *keyword,DTYPE *dout,void (*callback)(int,int));

/* CloseNetGlobal() */
/**
 * \brief Closes an active globals data link according to the globals Keyword
 *
 * Use this call to close an active data link.
 * If the call is successful, and the link is the last globals
 * link on a context's multicast group, the application will detach from the group.
 *
 * \param keyword the globals keyword used to establish the link.
 *
 * \see recvNetGlobal() or recvNetGlobalEx() and CloseGlobalLink()
 *
 */
TINE_EXPORT int CloseNetGlobal(const char *keyword);
#define closeNetGlobal(a1) CloseNetGlobal(a1)

/* CloseGlobalLink() */
/**
 * \brief Closes an active globals data link
 *
 * Use this call to close an active data link.  Note that
 * the parameter passed must be the returned link id when the original call to
 * recvNetGlobal() was made.  If the call is successful, and the link is the last globals
 * link on a context's multicast group, the application will detach from the group.
 *
 * \param linkId the globals link index to be canceled
 *
 * \see recvNetGlobal() or recvNetGlobalEx() and CloseNetGlobal()
 *
 */
TINE_EXPORT int CloseGlobalLink(int linkId);
#define closeGlobalLink(a1) CloseGlobalLink(a1)

/* GetGlobalsTimeStamp() */
/**
 * \brief Returns the timestamp of the globals keyword with the given link ID
 *
 * \param linkId the globals link index for which the timestamp is desired.
 *
 * \see recvNetGlobal() or recvNetGlobalEx() and CloseNetGlobal()
 *
 */
TINE_EXPORT double GetGlobalsTimeStamp(int linkId);

/* GetGlobalsTimeStampFromCallbackId() */
/**
 * \brief Returns the timestamp of the globals keyword with the given link callback ID
 *
 * \param id the callback ID passed to the intial call to recvNetGlobalEx().
 *
 * \see recvNetGlobalEx() and CloseNetGlobal()
 *
 */
TINE_EXPORT double GetGlobalsTimeStampFromCallbackId(int id);

/* recvNetGlobalEx() */
/**
 * \brief Initiates a net globals data link (extended call).
 *
 * A TINE client requests to receive network global parameters via this routine.  Note that the
 * client must reside on a network which itself is receiving broadcast data or onto which multicasts
 * are being routed in order for this call to take effect.
 *
 * recvNetGlobalEx() differs from recvNetGlobal() in that the caller can supply a callback identify to be
 * returned in the callback function.  In lieu of this, the callback will return the link identifier.
 *
 * \param keyword is the name of the network global which is desired.
 * If the globals keyword exists on a specific context, then the
 * keyword given should be prefixed with the context.  For example "/PETRA/Energy".
 * If the keyword is entered without a prefix then the 'SITE' globals are assumed (e.g. "PEENERGY").
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback funtion
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see recvNetGlobal(), sendNetGlobal(), DTYPE
 *
 * \b Example:
 *
 * \include eg_recvNetGlobalEx.c
 */
TINE_EXPORT int recvNetGlobalEx(const char *keyword,DTYPE *dout,void (*callback)(int,int),UINT32 callbackID);

/* recvNetGlobalEx2() */
/**
 * \brief Initiates a net globals data link (doubly extended call).
 *
 * A TINE client requests to receive network global parameters via this routine.  Note that the
 * client must reside on a network which itself is receiving broadcast data or onto which multicasts
 * are being routed in order for this call to take effect.
 *
 * recvNetGlobalEx2() differs from recvNetGlobal() and recvNetGlobalEx() in that the caller
 * can supply a callback identify and a reference pointer to be
 * returned in the callback function.  In lieu of this, the callback will return the link identifier
 * and a null pointer.
 *
 * \param keyword is the name of the network global which is desired.
 * If the globals keyword exists on a specific context, then the
 * keyword given should be prefixed with the context.  For example "/PETRA/Energy".
 * If the keyword is entered without a prefix then the 'SITE' globals are assumed (e.g. "PEENERGY").
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback funtion.
 * \param reference is supplied by the caller as a reference to be returned in the callback function.
 * In this way, the callback can then de-reference the originating source of the callback.
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be
 * interpreted by a call to GetLastLinkError().
 *
 * \see recvNetGlobal(), sendNetGlobal(), DTYPE
 *
 * \b Example:
 *
 * \include eg_recvNetGlobalEx2.c
 */
TINE_EXPORT int recvNetGlobalEx2(const char *keyword,DTYPE *dout,void (*callback)(int,int,void *),UINT32 callbackID,void *reference);

/**
 * \brief Structure used to hold engineering units, ranges, etc. for properties
 */
typedef struct PrpEguStruct
{
  char units[16];  /**< engineering units */
  float min;       /**< nominal maximum value */
  float max;       /**< nominal minimum value */
  BYTE graph;      /**< graph type (one of GT_LIN, GT_LOG, GT_SLOG, GT_TD, GT_BIT, GT_TEXT, GT_ABS, GT_HEX) */
  BYTE reserved[3];/**<  */
} PrpEgu;

/* RegisterFecInformation() */
/**
 * \brief Assigns a FEC Name and descriptive information to the server process.
 *
 * Servers must be assigned to a Front End Computer (FEC), which for many operating systems
 * corresponds to a process running on the computer rather than the computer itself (meaning
 * there can in some cases be more than one servers with different 'FEC' names running on
 * the same computer.  There can be several device servers attached to the same FEC (i.e.
 * sharing the same address space), so the FEC name is a distinct quantity from the
 * device server name (see RegisterExport()), and this name must also be system-wide
 * unique.  The call RegisterFecInformation() supercedes RegisterFecNameEx() and RegisterFecName()
 * and differs from them in that it accepts 'Context' and 'SubSystem' as parameters, both of
 * which will be applied to all device servers attaching to the same FEC name.
 *
 * The FEC name can also be (and frequently is) registered via the presence of an
 * 'fecid.csv' startup database file.
 *
 * \param name is the FEC name indentifying the Front End Computer and to which all registered
 *  equipment modules are bound.
 * \param sub is the subsystem to which attached device servers belong.  Note that the subsystem
 *  (unlike context) is itself not part of the name space, although the name space can be
 *  queried for device servers belonging to a subsystem.
 *  System.  This information is now deduced automatically from the library build.
 * \param cntxt is a 32-character string giving the context of all registered equipment modules.
 *  (earlier versions of TINE truncated at 16 characters).
 * \param dsc is a 64-character desciption of the FECs server duties.  A 'subsystem' can be
 *  associated with ALL equipment modules found on the FEC by prefixing the description with
 *  a subsystem tag of the form "[<tag>]".  This information will be parsed by the Equipment
 *  name server and used for sub-system specific queries.  For example, the associate the FEC
 *  "HEPHF" with the subsystem "RF", the description field might be "[RF]Hera Proton HF Control".
 * \param loc is a 32-character string giving the phyical location of the FEC.
 * \param hwd is a 32-character brief description of the IO hardware found on the FEC.
 * \param resp is a 32-character string listing the developer(s) responsible for the FEC. Note:
 *  The Equipment Name Server will allow the removal of FECs and associated equipment modules
 *  to the user(s) specified here.
 * \param poff is the 'Port Offset' to be applied to the FEC.  This parameter plays an important
 * role where more that one FEC is to run on a machine running an operating system using virtual
 * memory.  In such cases 'Front End Computer' is a misnomer, since "FEC" really refers to a
 * process running on the computer to which one or more equipment modules are bound.  For
 * operating systems where all processes run in the same address space (such as VxWorks,
 * Win16, NIOS) there is in fact only one such processes managing all requistered equipment
 * modules.  For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.),
 * you can have many such processes running independently of one another.  Since each such
 * process must listen on a unique set of server ports, you must see to it that all such
 * FEC processes are registered with a unique 'Port Offset'.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \note FEC Names can also be registered by including the startup file 'fecid.csv'.  In the
 * file, most of the input parameters appear as (optional) column entries.  It is frequently
 * preferable to work with startup configuation files, since it is frequently desireable to
 * avoid hard-coding names and description parameters.
 *
 * \see RegisterFecName(), RegisterFecNameEx().
 *
 * \b Example:
 *
 * \include eg_RegisterFecInformation.c
 */
TINE_EXPORT int RegisterFecInformation(char *name,char *sub,char *cntxt,char *dsc,char *loc,char *hdw,char *resp,UINT16 poff);

/* RegisterFecNameEx() */
/**
 * \brief Assigns a FEC Name to the server process.  Extended call.
 *
 * Servers must be assigned to a Front End Computer (FEC), which for many operating systems
 * corresponds to a process running on the computer rather than the computer itself (meaning
 * there can in some cases be more than one servers with different 'FEC' names running on
 * the same computer.  There can be several device servers attached to the same FEC (i.e.
 * sharing the same address space), so the FEC name is a distinct quantity from the
 * device server name (see RegisterExport()), and this name must also be system-wide
 * unique.  The extended call RegisterFecNameEx() differs from RegisterFecName() in that
 * it accepts the server 'Context' as an additional parameter.  The context will be applied
 * to all device servers attaching to the same FEC name.
 *
 * The FEC name can also be (and frequently is) registered via the presence of an
 * 'fecid.csv' startup database file.
 *
 * \param name is the FEC name indentifying the Front End Computer and to which all registered
 *  equipment modules are bound.
 * \param dsc is a 64-character desciption of the FECs server duties.  A 'subsystem' can be
 *  associated with ALL equipment modules found on the FEC by prefixing the description with
 *  a subsystem tag of the form "[<tag>]".  This information will be parsed by the Equipment
 *  name server and used for sub-system specific queries.  For example, the associate the FEC
 *  "HEPHF" with the subsystem "RF", the description field might be "[RF]Hera Proton HF Control".
 * \param os is included for backward compatibility.  Historicall it defined the FEC's Operating
 *  System.  This information is now deduced automatically from the library build.
 * \param loc is a 32-character string giving the phyical location of the FEC.
 * \param hwd is a 32-character brief description of the IO hardware found on the FEC.
 * \param resp is a 32-character string listing the developer(s) responsible for the FEC. Note:
 *  The Equipment Name Server will allow the removal of FECs and associated equipment modules
 *  to the user(s) specified here.
 * \param poff is the 'Port Offset' to be applied to the FEC.  This parameter plays an important
 * role where more that one FEC is to run on a machine running an operating system using virtual
 * memory.  In such cases 'Front End Computer' is a misnomer, since "FEC" really refers to a
 * process running on the computer to which one or more equipment modules are bound.  For
 * operating systems where all processes run in the same address space (such as VxWorks,
 * Win16, NIOS) there is in fact only one such processes managing all requistered equipment
 * modules.  For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.),
 * you can have many such processes running independently of one another.  Since each such
 * process must listen on a unique set of server ports, you must see to it that all such
 * FEC processes are registered with a unique 'Port Offset'.
 * \param context is a 32-character string giving the context of all registered equipment modules.
 * (earlier versions of TINE truncated at 16 characters).
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \note FEC Names can also be registered by including the startup file 'fecid.csv'.  In the
 * file, most of the input parameters appear as (optional) column entries.  It is frequently
 * preferable to work with startup configuation files, since it is frequently desireable to
 * avoid hard-coding names and description parameters.
 *
 * \see RegisterFecName(), RegisterFecInformation().
 *
 */
TINE_EXPORT int RegisterFecNameEx(char *name,char *dsc,char *os,char *loc,char *hdw,char *resp,UINT16 poff,char *context);

/* RegisterFecName() */
/**
 * \brief Assigns a FEC Name to the server process
 *
 * Servers must be assigned to a Front End Computer (FEC), which for many operating systems
 * corresponds to a process running on the computer rather than the computer itself (meaning
 * there can in some cases be more than one servers with different 'FEC' names running on
 * the same computer.  There can be several device servers attached to the same FEC (i.e.
 * sharing the same address space), so the FEC name is a distinct quantity from the
 * device server name (see RegisterExport()), and this name must also be system-wide
 * unique.
 *
 * The FEC name can also be (and frequently is) registered via the presence of an
 * 'fecid.csv' startup database file.
 *
 * \param name is the FEC name indentifying the Front End Computer and to which all registered
 *  equipment modules are bound.
 * \param dsc is a 64-character desciption of the FECs server duties.  A 'subsystem' can be
 *  associated with ALL equipment modules found on the FEC by prefixing the description with
 *  a subsystem tag of the form "[<tag>]".  This information will be parsed by the Equipment
 *  name server and used for sub-system specific queries.  For example, the associate the FEC
 *  "HEPHF" with the subsystem "RF", the description field might be "[RF]Hera Proton HF Control".
 * \param os is included for backward compatibility.  Historicall it defined the FEC's Operating
 *  System.  This information is now deduced automatically from the library build.
 * \param loc is a 32-character string giving the phyical location of the FEC.
 * \param hwd is a 32-character brief description of the IO hardware found on the FEC.
 * \param resp is a 32-character string listing the developer(s) responsible for the FEC. Note:
 *  The Equipment Name Server will allow the removal of FECs and associated equipment modules
 *  to the user(s) specified here.
 * \param poff is the 'Port Offset' to be applied to the FEC.  This parameter plays an important
 * role where more that one FEC is to run on a machine running an operating system using virtual
 * memory.  In such cases 'Front End Computer' is a misnomer, since "FEC" really refers to a
 * process running on the computer to which one or more equipment modules are bound.  For
 * operating systems where all processes run in the same address space (such as VxWorks,
 * Win16, NIOS) there is in fact only one such processes managing all requistered equipment
 * modules.  For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.),
 * you can have many such processes running independently of one another.  Since each such
 * process must listen on a unique set of server ports, you must see to it that all such
 * FEC processes are registered with a unique 'Port Offset'.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \note FEC Names can also be registered by including the startup file 'fecid.csv'.  In the
 * file, most of the input parameters appear as (optional) column entries.  It is frequently
 * preferable to work with startup configuation files, since it is frequently desireable to
 * avoid hard-coding names and description parameters.
 *
 * \see RegisterFecNameEx(), RegisterFecInformation().
 *
 */
TINE_EXPORT int RegisterFecName(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff);

/* GetPortOffset() */
/**
 * \brief Obtains the FEC port offset appropriate for the give FEC name
 *
 * In cases where the pure registration API routines are used to register FEC and server information
 * (as opposed to a common fecid.csv configuration file) it is often a more cumbersome task to ensure a
 * unique FEC port on the host machine.  This routine can be used to this end.  It will search the
 * existing FEC manifest on the current host in order to find the matching entry for the given FEC name.
 * If the entry is not found, then the 'next free port' is returned.  In this way the FEC can always
 * register with its own proper FEC port, without the need to hard code this number.
 *
 * If this routine is called after the server is running and the FEC name is not found in the manifest
 * the -name_unknown is returned.
 *
 * \param fecName is the FEC name in question (up to 16 characters).
 *
 * \return the port offset applicable for the given FEC name or the a negative number giving the
 * negative of the TINE return code indicating an error.
 *
 * \b example
 *
 * \include eg_GetPortOffset.c
 *
 * \see RegisterFecInformation()
 *
 */
TINE_EXPORT int GetPortOffset(const char *fecName);

/* RegisterPropertyAlias() */
/**
 * \brief Assigns an alias to the specified property.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param property is the property name in question (up to 64 characters).
 * \param alias is an alias which maps to the property in question (up to 64 characters).
 *
 * \return 0 if successful, otherwise the a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyInformation(), GetPropertyId()
 *
 */
TINE_EXPORT int RegisterPropertyAlias(char *eqm,char *property,char *alias);

/* RegisterPropertyAccessDeadband() */
/**
 * \brief Assigns a minimum access deadband to the designated property.
 *
 * By assigning a minimum access deadband to a property, a server can require
 * that successive WRITE calls cannot be processed at a smaller time interval
 * than the given deadband.  Attempts to do so will receive an 'operation_busy'
 * return code.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param property is the property name in question (up to 64 characters).
 * \param access is the data access for which the deadband is to be applied.
 * (usually one of CA_READ, CA_WRITE, or CA_READ|CA_WRITE).
 * \param deadbandInMilliSeconds is the desired deadband in milliseonds.  A value of
 * 0 (or less than 0) turns off the deadband checking.
 *
 * \return 0 if successful, otherwise the a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyInformation(), GetPropertyId()
 *
 */
TINE_EXPORT int RegisterPropertyAccessDeadband(char *eqm,char *property,int access,int deadbandInMilliSeconds);

/* RegisterPropertyEx() */
/**
 * \brief Assigns pertinent information for the specified property.
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.  However,
 * if input data sets require different data set specifications than the output data
 * sets then the call to RegisterPropertyInformation() should be used.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prop is property name in question (up to 64 characters).
 * \param siz is expected data size (i.e. maximum allowed data size) for this property
 * \param fmt is exected data format for this property
 * \param acc is the data 'access' (CA_READ or CA_WRITE) accepted by the server.
 *  Note: Applying the CA_NETWORK flag to the access will enforce asynchronous NETWORK
 *  (i.g. multicast acquisition) by the client.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param propId is the property identifier to be associated with the property name
 * returned in calls to GetPropertyId().
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyInformation(), GetPropertyId()
 *
 * \b Example:
 *
 * \include eg_RegistePropertyAlias.c
 */
TINE_EXPORT int RegisterPropertyEx(char *eqm,char *prop,int siz,short fmt,short acc,char *dsc,int prpId);

/* RegisterProperty() */
/**
 * \brief Assigns pertinent information for the specified property.
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.  However,
 * if input data sets require different data set specifications than the output data
 * sets then the call to RegisterPropertyInformation() should be used.  If the property
 * identifier is important, then either RegisterPropertyInformation() or
 * RegisterPropertyEx should be used, otherwise, the associated property identifier will
 * be supplied by the server engine and be given via the return code.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prop is property name in question (up to 64 characters).
 * \param siz is expected data size (i.e. maximum allowed data size) for this property
 * \param fmt is exected data format for this property
 * \param acc is the data 'access' (CA_READ or CA_WRITE) accepted by the server.
 *  Note: Applying the CA_NETWORK flag to the access will enforce asynchronous NETWORK
 *  (i.g. multicast acquisition) by the client.
 * \param dsc is the 64-character description of what the property reads or writes.
 * returned in calls to GetPropertyId().
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterPropertyEx(), RegisterPropertyInformation(), GetPropertyId()
 *
 */
TINE_EXPORT int RegisterProperty(char *eqm,char *prop,int siz,short fmt,short acc,char *dsc);

/* RegisterPropertyInformation() */
/**
 * \brief Assigns pertinent information for the specified property.
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prop is property name in question (up to 64 characters).
 * \param dout is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the output data set to be returned by the server
 * to the caller.
 * \param din is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the input data set to be supplied by the caller.
 * \param acc is the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server.
 * This parameter can be used to coerce client-side links into using preferred link
 * attributes. For example, by applying the CA_NETWORK flag (CA_READ|CA_NETWORK) to the access,
 * asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the client side.
 * Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the
 * registered property (i.e. the same security rules applied to CA_WRITE will be in force
 * for READ acquisitions, either permanently - no CA_READ flag, or temporarily in case of
 * an applied access lock - if both CA_READ and CA_XREAD are applied).
 * Applying the CA_SAVERESTORE flag will apply the internal save-and-restore logic to the
 * registered property (restoring the last property settings upon server startup, and saving
 * any changed property settings).
 * Applying the CA_INDEPENDENT flag will signal that any WRITE calls to this property
 * should not block equipment module access to other properties (if the property is to
 * be called in its own thread).
 * \param atype is the data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the
 * returned data.
 * \param rowlength if > 0 gives the row length of the returned array data.  This is
 * most relevant for double arrays, where the number of rows is then given by the output
 * data size divided by the row length.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param propId is the property identifier to be associated with the property name
 * returned in calls to GetPropertyId().
 * \param rdr is a redirection string, if the property is to be redirected
 * to another server.  Most properties are not redirected, and this parameter is usually
 * NULL.  If used, if will be parsed according to
 * /\<device context\>/\<device server\>/\<device name\>[device property], where the 'device name' and
 * 'device property' are optional.
 * If 'device context' is not present, the same context will be assumed as the currently
 * registered context.
 * If 'device name' is not present, all devices will be
 * redirected to the specified device server.  If 'device property' is not present
 * then the identical property name as registered will be redirected to the device server
 * specified.  If on the other hand, the device property is present in the string to be
 * parsed, then the property registered will be redirected to the device server specified
 * as well as to the device property specified.
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyEx(), GetPropertyId()
 *
 * \b Example:
 *
 * \include eg_RegisterPropertyInformation.c
 */
TINE_EXPORT int RegisterPropertyInformation(char *eqm,char *prop,DTYPE *dout,DTYPE *din,
                                       short acc,short atype,UINT16 rowlength,char *dsc,int propId,char *rdr);
/* RegisterPropertyInformationEx() */
/**
 * \brief Assigns pertinent information for the specified property (extended version).
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is property name in question (up to 64 characters).
 * \param dout is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the output data set to be returned by the server
 * to the caller.
 * \param din is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the input data set to be supplied by the caller.
 * \param acc is the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server.
 * This parameter can be used to coerce client-side links into using preferred link
 * attributes. For example, by applying the CA_NETWORK flag (CA_READ|CA_NETWORK) to the access,
 * asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the client side.
 * Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the
 * registered property (i.e. the same security rules applied to CA_WRITE will be in force
 * for READ acquisitions, either permanently - no CA_READ flag, or temporarily in case of
 * an applied access lock - if both CA_READ and CA_XREAD are applied).
 * Applying the CA_SAVERESTORE flag will apply the internal save-and-restore logic to the
 * registered property (restoring the last property settings upon server startup, and saving
 * any changed property settings).
 * Applying the CA_INDEPENDENT flag will signal that any WRITE calls to this property
 * should not block equipment module access to other properties (if the property is to
 * be called in its own thread).
 * \param atype is the data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the
 * returned data.
 * \param rowlen if > 0 gives the row length of the returned array data.  This is
 * most relevant for double arrays, where the number of rows is then given by the output
 * data size divided by the row length.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param egu is a reference to a PrpEgu object containing the units and max and min
 * settings for the nominal value of the property.
 * \param xegu is a reference to a PrpEgu object containing the units and max and min
 * settings for the x-axis value of the property in the event that the property returns a
 * double array.
 * \param pid is the property identifier to be associated with the property name
 * returned in calls to GetPropertyId().
 * \param rdr is a redirection string, if the property is to be redirected
 * to another server.  Most properties are not redirected, and this parameter is usually
 * NULL.  If used, if will be parsed according to
 * /\<device context\>/\<device server\>/\<device name\>[device property], where the 'device name' and
 * 'device property' are optional.
 * If 'device context' is not present, the same context will be assumed as the currently
 * registered context.
 * If 'device name' is not present, all devices will be
 * redirected to the specified device server.  If 'device property' is not present
 * then the identical property name as registered will be redirected to the device server
 * specified.  If on the other hand, the device property is present in the string to be
 * parsed, then the property registered will be redirected to the device server specified
 * as well as to the device property specified.
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyInformation(), GetPropertyId()
 *
 * \b Example:
 *
 * \include eg_RegisterPropertyInformationEx.c
 */
TINE_EXPORT int RegisterPropertyInformationEx(char *eqm,char *prp,DTYPE *dout,DTYPE *din,
                  short acc,short atype,UINT16 rowlen,char *dsc,PrpEgu *egu,PrpEgu *xegu,int pid,char *rdr);

/**
* \brief Removes a property name from an equipment module's property list
*
* This call will remove a previously assigned property from the current
* property list.  The call does \em not free memory associated with the
* property or any of its overloads.  This memory will only be freed if
* the equipment module itself is removed.  Thus, calling
* RegisterPropertyInformation with the same property name following a
* a call to RemoveProperty will find and reuse the previously allocated
* memory.
*
* \param eqm is the local equipment module name (maximum 6 characters in length)
* For example: "BPMEQM".
* \param property is property name to be removed
*
* \return 0 if successful, otherwise a TINE completion code
*
* \see RegisterDeviceName()
*/
TINE_EXPORT int RemoveProperty(char *eqm, char *property);

/* RedirectProperty() */
/**
 * \brief Applies the redirection string to the given property
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * A server can export properties or devices which live on other servers.  To do so a
 * redirection string needs to be applied to the property in question.  As different registered
 * devices can also be redirected to different servers.  This routine can be called several
 * times for the same property, once for each remote device.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prop is property name in question (up to 64 characters).
 * \param rdr is the redirection string of the form "/<context>/<server>/<device>/[property]"
 * where <server> is the only non-optional entry in the redirection string.  If <device> is
 * omitted, all calls for the given property will be redirected to the given <server>.  If
 * device is included, only those calls to the specified device and property will be redirected.
 * If <property> is specified then the given property name will be redirected to the remote
 * <server> and the remote <property>.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation(), RedirectDeviceName()
 *
 * \b Example:
 *
 * \include eg_RedirectProperty.c
 */
TINE_EXPORT int RedirectProperty(char *eqm,char *prop,char *rdr);

/* SetDeviceDescription() */
/**
 * \brief Assigns a device description to the specified device
 *
 * Many servers handle instances of a given device type, such as a set of
 * vacuum pumps, beam position monitors, magnet power supply controllers, etc. It is
 * occasionally useful to assign individual description texts pro device (e.g. a hardware address)
 * An alternative is to provide a startup database file 'devices.csv' containing
 * device descriptions.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param description is the device description to be associated with the device specified.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int SetDeviceDescription(char *eqm,char *devname,char *description);
TINE_EXPORT int RegisterDeviceDescription(char *eqm,char *devname,char *description);

/* SetDeviceLocation() */
/**
 * \brief Assigns a device location text to the specified device
 *
 * Many servers handle instances of a given device type, such as a set of
 * vacuum pumps, beam position monitors, magnet power supply controllers, etc. It is
 * occasionally useful to assign individual physical location information pro device, for
 * instance when a device is an 'ethernet' device and the specific bldg, room, rack, etc.
 * for the device needs to be made available.
 * An alternative is to provide a startup database file 'devices.csv' containing
 * device locations.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param location is the device location to be associated with the device specified.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int SetDeviceLocation(char *eqm,char *devname,char *location);

/* SetDeviceMask() */
/**
 * \brief Assigns a device mask to the specified device
 *
 * Many servers handle instances of a given device type, such as a set of
 * vacuum pumps, beam position monitors, magnet power supply controllers, etc. It is
 * occasionally useful to assign a type 'mask' to a device.
 * Wild card calls with an input mask or meta-property calls with '.DEVMASK.'
 * will then return only lists of devices with the corresponding mask bits set.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param mask is the device mask to be associated with the device specified.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int SetDeviceMask(char *eqm,char *devname,int mask);

/* SetDeviceOfflineStatus() */
/**
 * \brief Assigns an offline status to the specified device
 *
 * This purely software flag can be used to steer the returned device list in
 * wildcard or certain kinds of multichannel array calls.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param offline is the offline status to be associated with the device specified.
 * (non-zero => off-line).
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int SetDeviceOfflineStatus(char *eqm,char *devname,int offline);

/* SetDeviceZPosition() */
/**
 * \brief Assigns a Z (logitudinal) direction to the specified device
 *
 * Many servers handle instances of a given device type, such as a set of
 * vacuum pumps, beam position monitors, magnet power supply controllers, etc. It is
 * occasionally useful to assign a Z or longitudinal position to a device
 * when the relative location along the direction of the beam is pertinent
 * for the registered device.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param zpos is the Z position to be associated with the device specified.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int SetDeviceZPosition(char *eqm,char *devname,float zpos);

/* SetDeviceRegion() */
/**
 * \brief Assigns a region code to the device in question
 *
 * Region codes can help determine the location of a device with a machine. This
 * code is applied to the alarm message whenever a specific devices sets an alarm.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param region is the alarm region tag to apply to the device
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int SetDeviceRegion(char *eqm,char *devname,char *region);

/* RegisterDeviceName() */
/**
 * \brief Assigns a device name to the specified device number
 *
 * Many servers handle instances of a given device type, such as a set of
 * vacuum pumps, beam position monitors, magnet power supply controllers, etc. It is
 * frequently very efficient to code in terms of device numbers which
 * might be an index into an array, but
 * to specify modules in calls by their human-readable names.  For such purposes a device
 * name can be assigned to a number at initialization via a call to RegisterDeviceName()
 * (alias: RegisterModuleName()).  An alternative is to provide a startup database file 'devices.csv' containing
 * a cross-reference for numbers and names.  Internally, a hash table is maintained for
 * fast lookups inside equipment module routines.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name to be assigned
 * \param devnr is the device number associated with the device name specified.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *
 * \b Example:
 *
 * \include eg_RegisterDeviceName.c
 */
TINE_EXPORT int RegisterDeviceName(char *eqm,char *devname,int devnr);
#define RegisterModuleName(eqm,dev,devnr) RegisterDeviceName(eqm,dev,devnr)

/**
* \brief Removes a device name from an equipment module's device list
*
* This call will remove a previously assigned device from the current
* device list.  That is, it will remove any allocated memory specific to
* the removed device.
* It does \em not remove the device \em slot and the device list \em capacity
* is unaltered by this call.
*
* \param eqm is the local equipment module name (maximum 6 characters in length)
* For example: "BPMEQM".
* \param devname is device name to be removed
*
* \return 0 if successful, otherwise a TINE completion code
*
* \note If RegisterMultiChannelGroupDevice() has been used to \em assign
* the given device to a multi-channel group device, then it is the
* caller's responsibility to ensure the further integrity of the multi-channel
* group device and its constituent members
* (e.g. by calling RegisterMultiChannelGroupDevice() again with the \em new
* information for the remaining members).
* Likewise any multi-channel array logic contained in a property or equipment
* module handler must be adjusted by the caller.
*
* \see RegisterDeviceName()
*/
TINE_EXPORT int RemoveDevice(char *eqm, char *devname);
/* RegisterMultiChannelGroupDevice() */
/**
 * \brief Assigns a device name to the specified multi-channel group device
 *
 * In a device-oriented model, one can still make use of multi-channel array
 * efficiency by registering a 'group' device which knows its device members
 * and group size. Individual devices can then be designated as multi-channel
 * array 'participants' such that any persistent client-side links to
 * group members results in a delivery of the entire group to the client
 * and an associated 'collapse' in the number of associated single-element
 * contracts to a 'group' contract.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param grpname is device group name
 * \param devname is device name to be assigned to a group
 * \param grpindex is the array index of the device member within the group
 * \param grpsize is the array length of the group
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_RegisterMultiChannelGroupDevice.c
 */
TINE_EXPORT int RegisterMultiChannelGroupDevice(char *eqm,char *grpname,char *devname,int grpindex,int grpsize);

/**
 * \brief Gets the NAME64 list of member devices associated with the given group
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param grpname is device group name
 * \param members [out] is a pointer to a NAME64 buffers to receive the list
 * \param len [in, out] is a pointer to buffer length (input) and contains the length
 * of members list returned.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int GetGroupDeviceMembers(char *eqm, char *grpname, NAME64 *members, int *len);

/**
 * \brief Gets a NAME64 list of registered device groups
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param groups [out] is a pointer to a NAME64 buffers to receive the list
 * \param len [in, out] is a pointer to buffer length (input) and contains the length
 * of device groups list returned.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int GetDeviceGroups(char *eqm, NAME64 *groups, int *len);
extern int enforceMcaAcquisition;
/* SetEnforceMcaAcquisition() */
/**
 * \brief Forces multi-channel array handshaking if set to TRUE
 *
 * A server can absolutely enforce multi-channel array handshaking independent of
 * the release level of the client.  This has relevance as to whether persistent
 * single-element links are allowed or not.  If allowed then a 'modern' client might
 * be able to latch on to a pre-existing single-element link established by an older
 * client.
 *
 * \param value TRUE or FALSE
 *
 * \note An 'old' client will receive the error code 'propety_is_mca' which it will not
 * be able to interpret in any meaningful way.
 *
 */
TINE_EXPORT void SetEnforceMcaAcquisition(int value);

/* GetEnforceMcaAcquisition() */
/**
 * \brief returns the setting for multi-channel array handshaking enforcement.
 *
 * A server can absolutely enforce multi-channel array handshaking independent of
 * the release level of the client.  This has relevance as to whether persistent
 * single-element links are allowed or not.  If allowed then a 'modern' client might
 * be able to latch on to a pre-existing single-element link established by an older
 * client.
 *
 * \return the current setting for this value
 *
 * \see SetEnforceMcaAcquisition
 */
TINE_EXPORT int GetEnforceMcaAcquisition(void);

/* ResetMultiChannelProperty() */
/**
 * \brief sends (schedules) a 'reset_mca_property' signal to any listening client
 *
 * A server with registered multi-channel array properties can inform any attached
 * clients that the mutli-channel array configuration has changed (e.g. array elements
 * have been added, removed, or otherwise edited) by using this call.
 * Any listening clients with multi-channel array links will then return to the original
 * client startup conditions and 're-learn' the new array indexing now in place.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is (comma separated) list of properties which are to be reset at the client
 * side.
 *
 * \note This routine will call _SystemSchedulePropertyEx() to assure immediate notification
 * at the client side.
 *
 * \return 0 upon success, otherwise a TINE return code.
 *
 * \see SetEnforceMcaAcquisition
 */
TINE_EXPORT int ResetMultiChannelProperty(char *eqm, char *prp);

/* AssignPropertyList() */
/**
 * \brief Assigns the given property list to the given device to be used in device-oriented queries
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * Most device servers provide a list of all instances, a device list, and a list of all properties,
 * where every instance (device) supports all exported properties.  Under some circumstances,
 * it is desireable to support only a (device-specific) subset of properties per device.  This
 * can be achieved by supplying a property query function, or by configuration file, or by using
 * this call to assign a property list to the specified device.  The latter is the preferred
 * method, as then all registered properties are contained within the TINE property registry,
 * which will contain sometimes crucial additional information about the registered property.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name in question (up to 32 characters, preferably 16 or less).
 * \param listname is a 'tag' name to be assigned to the list. Where many devices share the same
 * property list, this tag is used to avoid unnecessarily allocating new memory for the same list.
 * \param listsize is the length of the property list.
 * \param list is the property list as an array of NAME64 entities.  If the list provided is not
 * at least as long as indicated by parameter listsize, the server's behavior is undefined.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RegisterPropertyQueryFunction(), RegisterXPropertyQueryFunction().
 *
 */
TINE_EXPORT int AssignPropertyList(char *eqm,char *devname,char *listname,int listsize,NAME64 *prplist);

/* AssignDeviceListToProperty() */
/**
 * \brief Assigns the given device list to the given registered property
 *
 * Properties (e.g. multi-channel array properties) which wish to present a
 * property-specific device list can assign the list to the property in this way.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is property name for which the device list is to be assigned.
 * \param devlst is the device list which is to be associated with the given property
 * \param lstlen is the length of the device list.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation()
 *
 */
TINE_EXPORT int AssignDeviceListToProperty(char *eqm,char *prp,NAME64 *devlst,int lstlen);

/* RedirectDeviceName() */
/**
 * \brief Applies the redirection string to the given device for all properties
 *
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * A server can export devices which live on other servers.  To do so a
 * redirection string needs to be applied to the device in question.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is device name in question.
 * \param rdr is the redirection string whose general form is "/<context>/<server>/<device>/[property]".
 * The only relevant tag in the redirection string is the <server> tag as a call to RedirectDeviceName
 * will apply to all properties for the specified device.  The <context> tag is as yet always
 * assumed to be the same context as the redirecting server.  Thus the redirection string can
 * be specified as simply "<server>"
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RedirectProperty(), RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation()
 *
 */
TINE_EXPORT int RedirectDeviceName(char *eqm,char *devname,char *rdr);

/* GetDeviceDescription() */
/**
 * \brief Gives the registered device description (if known) for the device number
 *
 * A server can determine the device description which has been associated with a device number
 * by making this call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devnr is the device number to be looked up.
 *
 * \return The device description corresponding to the input device number if successful, otherwise a
 * error string indicating the problem.
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT char *GetDeviceDescription(char *eqm,int devnr);

/* GetDeviceLocation() */
/**
 * \brief Gives the registered device location (if known) for the device number
 *
 * A server can determine the device location which has been associated with a device number
 * by making this call. This information should provide the physical location of the device
 * hardware (e.g. Bldg 1, Rm 101, Rack 15, slot 2).
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devnr is the device number to be looked up.
 *
 * \return The device location corresponding to the input device number if successful, otherwise a
 * error string indicating the problem.  The call with return the exact registered location
 * string (or the FEC location string).  The the location is redirected to a 'locator' server then
 * the location string is liable to contain the redirection information and NOT the location itself.
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT char *GetDeviceLocation(char *eqm,int devnr);

/* GetDeviceMask() */
/**
 * \brief Gives the registered device mask for the device number
 *
 * A server can determine the device mask which has been associated with a device number
 * by making this call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devnr is the device number to be looked up.
 *
 * \return The registered device mask (default = 0 signifies NO mask).
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int GetDeviceMask(char *eqm,int devnr);

/* GetDeviceOfflineStatus() */
/**
 * \brief Gives the current off line status for the device in question
 *
 * This is a software flag which can be used to steer the output in
 * wildcard calls or certain multi-channel array calls.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devnr is the device number to be looked up.
 *
 * \return The current offline status (default = 0 signifies on line).
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int GetDeviceOfflineStatus(char *eqm,int devnr);

/* GetDeviceZPosition() */
/**
 * \brief Gives the registered device Z (or longitudinal) position for the device number
 *
 * A server can determine the device Z (or longitudinal) position which has been
 * associated with a device number by making this call. This will give the position of the
 * measurement along the path of the beam.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devnr is the device number to be looked up.
 *
 * \return The registered Z position (default = 0 signifies not relevant or not registered).
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT float GetDeviceZPosition(char *eqm,int devnr);

/* GetDeviceName() */
/**
 * \brief Gives the registered device name for the specified equipment module and device number
 *
 * A server can determine the name which has been associated with a device number by making this
 * call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devnr is the device number to be looked up.
 *
 * \return The device name corresponding to the input device number if successful, otherwise a
 * string reiterating the input device number.
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 * \b Alias: GetModuleName
 *
 * \b Example:
 *
 * \include eg_GetDeviceName.c
 */
TINE_EXPORT char *GetDeviceName(char *eqm,int devnr);
#define GetModuleName(eqm,devnr) GetDeviceName(eqm,devnr)

/* GetExportedDeviceList() */
/**
 * \brief Returns the list of exported device names explicitly associated with the
 * equipment module passed as the input parameter.
 *
 * A server can retieve the list of explicitly exported device names by making this
 * call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return A array of NAME32 objects containing the exported device names or a NULL
 * pointer if there is an error, such as the equipment module given is not registered or
 * there are no exported device names associated with the equipment module.
 *
 * \see RegisterDeviceName()
 *
 */
TINE_EXPORT NAME64 *GetExportedDeviceList(char *eqm);

/* GetRegisteredPropertyList() */
/**
 * \brief Returns the list of registered properties explicitly associated with the
 * equipment module passed as the first argument input parameter.
 *
 * A server can retieve the list of registered properties by making this
 * call.  This is sometimes useful when the property list is maintained in an
 * 'exports.csv' file and the server application needs to retrieve this list.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpNames (output) is the address location which is to hold the property list
 * upon completion
 * \param nprps (intput/output) is a pointer to an integer which as input specifies the size of the
 * buffer pointed to by the second parameter. Upon return it contains the actual number of
 * registered properties.
 *
 * \return '0' upon success, othereise a tine return code.
 *
 */
TINE_EXPORT int GetRegisteredPropertyList(char *eqm,NAME64 *prpNames,int *nprps);

/* getLocalName() */
/**
 * \brief Returns the local equipment module name associated with the export name given.
 *
 * \param exportName is the exported device server name for which the local equipment module
 * name is desired.
 *
 * \return a string pointing to the local equipment module name.
 *
 */
TINE_EXPORT char *GetLocalName(char *exportName);
#define getLocalName(a1) GetLocalName(a1)

TINE_EXPORT int GetNumberEquipmentModules(void);

/* GetNumberRegisteredDevices() */
/**
 * \brief Gives the number of registered devices explicitly associated with the equiment module name
 * given.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The number of registered devices if successful, or the
 * negative of a TINE return code (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int GetNumberRegisteredDevices(char *eqm);
/**
 * \brief Gives the registered exported equipment module name
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The the registered exported equipment module name
 *
 * \see GetRegisteredContext()
 *
 */
TINE_EXPORT char *GetRegisteredExportName(char *eqm);
/**
 * \brief Gives the registered context
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The the registered context
 *
 * \see GetRegisteredExportName()
 *
 */
TINE_EXPORT char *GetRegisteredContext(char *eqm);

/* GetSizeDeviceCapacity() */
/**
 * \brief Gives the maximum size of the device table associated with the equiment module name
 * given.
 *
 * This given the originally registered amount of device space for this equipment module.
 * Attempts to register devices with device numbers greater than this value will fail.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The size of the device table associated with the given equipment module or
 * negative of a TINE return code (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 */
TINE_EXPORT int GetSizeDeviceCapacity(char *eqm);

/* GetValidDeviceNumber() */
/**
 * \brief Gives the valid registered device number for the specified device name and property name.
 *
 * Callers frequently address specific devices by their human-readable names rather than
 * a device number (which might be an entry point into an array, for example).  If the
 * caller registers device names via calls to RegisterDeviceName() or has registered a "<property>.NAM"
 * meta property (corresponding to a property-specific set of device names for <property>), then the
 * cross reference exists in the form of a hash table.  A server's equipment module can thus rapidly
 * determine the corresponding device number from an input device name by calling GetValidDeviceNumber().
 *
 * Passing a device name of the form "#7" and ceiling < 0 will return '7'
 * only if '7' does in fact correspond to a valid device.  This is different from a call to
 * either GetDeviceNumber() or GetDeviceNumberEx(), which dutifully return the parsed and converted
 * device number.  To retain this behavior, you should pass ceiling = 0.  Passing a value of
 * ceiling > 0 will use the passed value to determine the validity of the device string.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is the device name to be looked up
 * \param prpname if non-NULL should be the targeted property name for which the device names apply
 * \param ceiling is the valid upper limit for a valid device number.  This plays a role only if
 * a devname with a leading '#' is passed.
 * Pass a value < 0 to validate according to the registered device information.  Pass a value > 0
 * to validate according to literal value passed.  ceiling = 0 will simply return the converted
 * string value (same behavior as GetDeviceName()).
 *
 * \return The device number (i.e. multi-channel array index) corresponding to the input device name
 * if successful, otherwise the negative of a TINE return code
 * (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber(), GetDeviceNumberEx()
 *
 */
TINE_EXPORT int GetValidDeviceNumber(char *eqm,char *devname,char *prpname,int ceiling);

/* GetDeviceNumberEx() */
/**
 * \brief Gives the registered device number for the specified device name and property name
 *
 * Callers frequently address specific devices by their human-readable names rather than
 * a device number (which might be an entry point into an array, for example).  If the
 * caller registers device names via calls to RegisterDeviceName() or has registered a "<property>.NAM"
 * meta property (corresponding to a property-specific set of device names for <property>), then the
 * cross reference exists in the form of a hash table.  A server's equipment module can thus rapidly
 * determine the corresponding device number from an input device name by calling GetDeviceNumberEx().
 *
 * Passing a device name of the form "#7" will return '7'
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is the device name to be looked up
 * \param prpname if non-NULL should be the targeted property name for which the device names apply
 *
 * \return The device number (i.e. multi-channel array index) corresponding to the input device name
 * if successful, otherwise the negative of a TINE return code
 * (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 * \b Example:
 *
 * \include eg_GetDeviceNumberEx.c
 */
TINE_EXPORT int GetDeviceNumberEx(char *eqm,char *devname,char *prpname);
#define GetModuleNumberEx(eqm,dev,prp) GetDeviceNumberEx(eqm,dev,prp)

/* GetDeviceNumber() */
/**
 * \brief Gives the registered device number for the specified device name
 *
 * Callers frequently address specific devices by their human-readable names rather than
 * a device number (which might be an entry point into an array, for example).  If the
 * caller registers device names via calls to RegisterDeviceName() then the cross reference
 * exists in the form of a hash table.  A server's equipment module can thus rapidly determine
 * the corresponding device number from an input device name by calling GetDeviceNumber().
 *
 * Passing a device name of the form "#7" will return '7'
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param devname is the device name to be looked up
 *
 * \return The device number corresponding to the input device name if successful, otherwise the
 * negative of a TINE return code (indicating that the name input is not registered as a device on the server).
 *
 * \see RegisterDeviceName(), GetDeviceNumberEx()
 *
 * \b Alias: GetModuleNumber
 *
 * \b Example:
 *
 * \include eg_GetDeviceNumber.c
 */
TINE_EXPORT int GetDeviceNumber(char *eqm,char *devname);
#define GetModuleNumber(eqm,dev) GetDeviceNumber(eqm,dev)

/* GetPropertyId() */
/**
 * \brief Gives the associated property identifier for the given property name
 *
 * Callers will alway address properties by their human-readable names.
 * Equipment modules can either initiate a series of strcmp()s in order to determine
 * which property is being called or make direct use of the associated property identifier
 * in a case switch or address table.  Note that properties registered with RegisterProperty()
 * simply return an index in the server's ordered list.  You can associate any indentifier you
 * like by calling either RegisterPropertyEx() or RegisterPropertyInformation() to register
 * the property.  The cross reference from property name to property identifier
 * exists in the form of a hash table.  A server's equipment module can thus rapidly determine
 * the corresponding identifier from an input property by calling GetPropertyId().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName is the property name to be looked up
 *
 * \return The property identifier corresponding to the input property name if successful,
 * otherwise -1 (indicating that the name input is not registered as a property on the server).
 *
 * \see RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation()
 *
 * \b Example:
 *
 * \include eg_GetPropertyId.c
 */
TINE_EXPORT int GetPropertyId(char *eqm,char *prpName);

int RegisterExport(char *expName,char *eqmName,int numdevices);

/* JoinEquipmentGroup() */
/**
 * \brief Instructs the equipment module to join the specified equipment group.
 *
 * It is sometimes desireable to access a collection of like devices via a
 * device group, rather than a device server.  This is of course true when the
 * devices in the group span more than one physical device server.
 * You can use this API call to instruct the equipment module to join the specified
 * group.  You should use this call prior to calling SystemCycle().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param groupname is the name of the group the equipment module is to join.
 * \param groupindex is an ordered index, which can specify which location within the
 * group this equipment module's devices are to have.  If '0' is passed, the group
 * server will assign an order.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_JoinEquipmentGroup.c
 *
 * \see JoinEquipmentGroupEx()
 */
TINE_EXPORT int JoinEquipmentGroup(char *eqm,char *groupname,int groupindex);

/**
 * \brief Instructs the equipment module to join the specified equipment group.
 *
 * It is sometimes desireable to access a collection of like devices via a
 * device group, rather than a device server.  This is of course true when the
 * devices in the group span more than one physical device server.
 * You can use this API call to instruct the equipment module to join the specified
 * group.  You should use this call prior to calling SystemCycle().
 * With the extended call you can also pass an optional prefix or postfix to
 * be appended to the device names used at the group server.  This can be a very
 * useful option where multiple instances of a device server each supply a device
 * list of identical names (e.g. where the device names do not correspond to
 * specific locations, etc.).
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param groupname is the name of the group the equipment module is to join.
 * \param groupindex is an ordered index, which can specify which location within the
 * group this equipment module's devices are to have.  If '0' is passed, the group
 * server will assign an order.
 * \param devPrefix is a (up to 8-character) prefix to be added to all device names
 * acquired from the physical server.
 * \param devPostfix is a (up to 8-character) prostfix to be appended to all device names
 * acquired from the physical server.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_JoinEquipmentGroup.c
 */
TINE_EXPORT int JoinEquipmentGroupEx(char *eqmName,char *groupname,int groupindex,char *devPrefix,char *devPostfix);

/* RegisterEquipmentModule() */
/**
 * \brief Registers an equipment module with the TINE server engine
 *
 * The point of contract between the TINE server engine and server-specific code is the
 * equpment module, which can be regarded as an interupt service routine.
 *
 * \param expName is the exported device server name to be associated with this equipment module.
 * If this is NULL, then it is assumed that an 'exports.csv' local database file exists, which
 * contains the export name.  Otherwise it will contain a system-wide unique name (e.g. "HERA-P-BPM")
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param ndevices is this number of devices managed via this equipment module.  If this is 0, then
 * it is assumed that an 'exports.csv' local database file exists, which contains the number of devices.
 * \param fcn is a reference to the equipment module, which must have the prototype shown.  Namely, it
 * is an interrupt service routine which will receive the device name, device property, the output data
 * set, the input data set, and the data access flags as parameters.
 * \param ini is a reference to an initialization routine which should be called prior to enabling the
 * equipment module.  It does not take arguments.
 * \param tsk is a reference to a background task which will be called at regular intervals, and is
 * frequently used for hardware IO, etc.
 * Where multi-threaded servers are specified, or VxWorks tasks are available these tasks exists as
 * separate threads, or tasks.  In other cases, they are managed by the TINE server engine scheduler.
 * \param rate is the rate in milliseconds used to schedule the background task supplied by the 'tsk'
 * parameter.
 * \param exi is a reference to an exit routine which should be called prior to shutting down the
 * equipment module.  It does not take arguments, and applies of course to graceful exits of the server.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_RegisterEquipmentModule.c
 */
TINE_EXPORT int RegisterEquipmentModule(
    char *expName,
    char *eqm,
    int ndevices,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short),
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
);

/* RegisterEquipmentModuleEx() */
/**
 * \brief Registers an equipment module with the TINE server engine (extended call)
 *
 * The point of contract between the TINE server engine and server-specific code is the
 * equpment module, which can be regarded as an interupt service routine. This extended call allows a
 * reference to be passed, which in turn will be used in all associated dispatch routines.
 *
 * \param expName is the exported device server name to be associated with this equipment module.
 * If this is NULL, then it is assumed that an 'exports.csv' local database file exists, which
 * contains the export name.  Otherwise it will contain a system-wide unique name (e.g. "HERA-P-BPM")
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param ndevices is this number of devices managed via this equipment module.  If this is 0, then
 * it is assumed that an 'exports.csv' local database file exists, which contains the number of devices.
 * \param fcn is a reference to the equipment module, which must have the prototype shown.  Namely, it
 * is an interrupt service routine which will receive the device name, device property, the output data
 * set, the input data set, and the data access flags as parameters, as well as the original reference pointer.
 * \param ini is a reference to an initialization routine which should be called prior to enabling the
 * equipment module.  When it is called, it passes the original reference pointer.
 * \param tsk is a reference to a background task which will be called at regular intervals, and is
 * frequently used for hardware IO, etc.  When it is called, it passes the original reference pointer.
 * Where multi-threaded servers are specified, or VxWorks tasks are available these tasks exists as
 * separate threads, or tasks.  In other cases, they are managed by the TINE server engine scheduler.
 * \param rate is the rate in milliseconds used to schedule the background task supplied by the 'tsk'
 * parameter.
 * \param exi is a reference to an exit routine which should be called prior to shutting down the
 * equipment module.  When it is called, it passes the original reference pointer,
 * and applies of course to graceful exits of the server.
 * \param reference is a reference pointer which will be returned in all dispatch routines, and thus
 * can be used as a means for de-referencing the equipment module container.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_RegisterEquipmentModuleEx.c
 */
TINE_EXPORT int RegisterEquipmentModuleEx(
    char *expName,
    char *eqmName,
    int numdevices,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short,void *),
    void (*ini)(void *),
    void (*tsk)(void *),
    int rate,
    void (*exi)(void *),
    void *reference
);

/* RegisterServer() */
/**
 * \brief Registers an internal equipment module with the TINE server engine
 * An internal equipment module is provided by the TINE server engine. So for creating a TINE server
 * it is sufficient to register properties and respective property handlers,
 * see also API call RegisterPropertyAndHandler
 *
 * \param expName is the exported device server name to be associated with this equipment module.
 * \param ndevices is this number of devices managed via this equipment module.
 * \param deviceNameList can be an string array of the length of ndevices, containing device names. If NULL
 * then a list of ndevices generic device names (00000000, 00000001, 00000002 ...) is generated.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_RegisterServer.c
 */
TINE_EXPORT int RegisterServer(char *expName, int numdevices, char *deviceNameList[]);

/* RegisterPropertyAndHandler() */
/**
 * \brief Registers a device property and a property handler with the TINE server engine.
 * If no equipment module name is given, an internal equipment module is provided by the TINE
 * server engine, see API call RegisterServer.
 * If not an internal equipment module but a specifically for this server created equipment module
 * should be used, this can be indicated by the parameter eqmName, see below.
 *
 * \param eqm defines the associated equipment module. If NULL then a TINE kernel internal equiment module
 * is assumed, else it should be the registered 6-character local equipment module name (e.g. "BPMEQM") of
 * the server to which the property belongs.
 * \param hndlr is a reference to the property handler function, which must have the prototype
 *
 * int (*hndlr)(char *,char *,DTYPE *,DTYPE *,short)
 *
 * Namely, it
 * is an interrupt service routine which will receive the device name, device property, the output data
 * set, the input data set, and the data access flags as parameters, as well as the original reference pointer.
 *
 * the remaining parameters are exactly the same as in API call RegisterPropertyInformation
 *
 * \param prp is property name in question (up to 64 characters).
 * \param dout is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the output data set to be returned by the server
 * to the caller.
 * \param din is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the input data set to be supplied by the caller.
 * \param acc is the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server.
 * This parameter can be used to coerce client-side links into using preferred link
 * attributes. For example, by applying the CA_NETWORK flag (CA_READ|CA_NETWORK) to the access,
 * asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the client side.
 * Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the
 * registered property (i.e. the same security rules applied to CA_WRITE will be in force
 * for READ acquisitions, either permanently - no CA_READ flag, or temporarily in case of
 * an applied access lock - if both CA_READ and CA_XREAD are applied).
 * Applying the CA_SAVERESTORE flag will apply the internal save-and-restore logic to the
 * registered property (restoring the last property settings upon server startup, and saving
 * any changed property settings).
 * \param atype is the data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the
 * returned data.
 * \param rowlen if > 0 gives the row length of the returned array data.  This is
 * most relevant for double arrays, where the number of rows is then given by the output
 * data size divided by the row length.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param egu is a reference to a PrpEgu object containing the units and max and min
 * settings for the nominal value of the property.
 * \param xegu is a reference to a PrpEgu object containing the units and max and min
 * settings for the x-axis value of the property in the event that the property returns a
 * double array.
 * \param pid is the property identifier to be associated with the property name
 * returned in calls to GetPropertyId().
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_RegisterPropertyAndHandler.c
 */
TINE_EXPORT int RegisterPropertyAndHandler(
    char *eqm,
    EQMFCNP hndlr,
    char *prp,DTYPE *dout,DTYPE *din,short acc,short atype,
    UINT16 rowlen,char *dsc,
    PrpEgu*egu,PrpEgu *xegu,
    int pid
);

/* RegisterPropertyAndHandlerEx() */
/**
 * \brief Registers a device property and a property handler with the TINE server engine (extended call).
 * If no equipment module name is given, an internal equipment module is provided by the TINE
 * server engine, see API call RegisterServer.
 * If not an internal equipment module but a specifically for this server created equipment module
 * should be used, this can be indicated by the parameter eqmName, see below.
 *
 * \param eqm defines the associated equipment module. If NULL then a TINE kernel internal equiment module
 * is assumed, else it should be the registered 6-character local equipment module name (e.g. "BPMEQM") of
 * the server to which the property belongs.
 * \param hndlr is a reference to the property handler function, which must have the prototype
 *
 * int (*hndlr)(char *,char *,DTYPE *,DTYPE *,short,void *)
 *
 * Namely, it
 * is an interrupt service routine which will receive the device name, device property, the output data
 * set, the input data set, and the data access flags as parameters, as well as the original reference pointer.
 * \param prp is property name in question (up to 64 characters).
 * \param dout is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the output data set to be returned by the server
 * to the caller.
 * \param din is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the input data set to be supplied by the caller.
 * \param acc is the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server.
 * This parameter can be used to coerce client-side links into using preferred link
 * attributes. For example, by applying the CA_NETWORK flag (CA_READ|CA_NETWORK) to the access,
 * asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the client side.
 * Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the
 * registered property (i.e. the same security rules applied to CA_WRITE will be in force
 * for READ acquisitions, either permanently - no CA_READ flag, or temporarily in case of
 * an applied access lock - if both CA_READ and CA_XREAD are applied).
 * Applying the CA_SAVERESTORE flag will apply the internal save-and-restore logic to the
 * registered property (restoring the last property settings upon server startup, and saving
 * any changed property settings).
 * \param atype is the data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the
 * returned data.
 * \param rowlen if > 0 gives the row length of the returned array data.  This is
 * most relevant for double arrays, where the number of rows is then given by the output
 * data size divided by the row length.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param egu is a reference to a PrpEgu object containing the units and max and min
 * settings for the nominal value of the property.
 * \param xegu is a reference to a PrpEgu object containing the units and max and min
 * settings for the x-axis value of the property in the event that the property returns a
 * double array.
 * \param pid is the property identifier to be associated with the property name
 * returned in calls to GetPropertyId().
 * \param ref is a reference pointer which will be returned in the given property handler
 * dispatch routines, and thus can be used as a means for de-referencing the container.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_RegisterPropertyAndHandlerEx.c
 */
TINE_EXPORT int RegisterPropertyAndHandlerEx(
    char *eqm,
    XEQMFCNP hndlr,
    char *prp,DTYPE *dout,DTYPE *din,short acc,short atype,
    UINT16 rowlen,char *dsc,
    PrpEgu*egu,PrpEgu *xegu,
    int pid, void *ref
);

/* RemoveEquipmentModule() */
/**
 * \brief Unregisters an equipment module with the TINE server engine and frees all associated memory
 *
 * The point of contract between the TINE server engine and server-specific code is the
 * equpment module, which can be regarded as an interupt service routine. This extended call allows a
 * reference to be passed, which in turn will be used in all associated dispatch routines. Under (rare)
 * circumstances it may be desired to unregister an equipment module.  This routine can be used to
 * accomplish that task.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterEquipmentModule(), RegisterEquipmentModuleEx()
 *
 * \b Example:
 *
 * \include eg_RemoveEquipmentModule.c
 */
TINE_EXPORT int RemoveEquipmentModule(const char *eqm);

/**
 * \brief Resets a server baseline timestamp to the current time.
 *
 * A client can monitor the stock property "SRVBASELINE" and react to jumps in the readback value.
 * Normally this value will not change over the course of the server process's life cycle.
 * The value will naturally change when a server is restarted.  However the server process can
 * call this routine when for instance the startup conditions change dynamically (e.g. a device
 * was added or removed on-the-fly).  A running client which monitors the SRVBASELINE will then
 * have a chance to re-acquire startup information 'on-the-fly' as well.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int ResetServerBaseline(const char *eqm);

/**
 * \brief Flags a link so that the given error value is used as the data value in case of
 * any link error.
 *
 * \param linkId is a valid link table id of the link for which the local timestamp should be applied.
 * \param errValue is a reference to the error value data set which should be applied in case of
 * a link error.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int PutLnkErrorValue(int linkId, void *errValue);
#define putLnkErrorValue(a1,a2) PutLnkErrorValue(a1,a2)

/**
 * \brief Flags a link so that the local timestamp is used for all incoming data sets.
 *
 * \param linkId is a valid link table id of the link for which the local timestamp should be applied.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int PutLnkTimeStamp(int linkId);

/**
 * \struct PropertyQueryStruct
 * \brief Defines a simple property query structure (legacy, minimal information)
 *
 * \note As there is no way to distinguish input data set from output data sets with this
 * simple form of the query structure, the assumption that the input size and type are the
 * same as the output size and type (if input data are required) is usually made.  The
 * extended property query structure 'PropertyQueryStructEx' is more complete in this regard.
 *
 */
typedef struct
{
  char  prpName[PROPERTY_NAME_SHORTSIZE]; /**< the property name */
  char  prpDescription[32];   /**< the property description */
  UINT16 prpSize;             /**< the maximum allowed size of an output data set associated with the property */
  BYTE prpFormat;             /**< the default output data format accepted for this property */
  BYTE prpAccess;             /**< the allowed data access for this property (either CA_READ, CA_WRITE or both) */
} PropertyQueryStruct;
#define PROPERTYQUERYSTRUCT_SIZE (32+32+2+1+1)

#define PQSR2_TAG "PQS"

/**
 * \struct PropertyQueryStructEx
 * \brief Defines an extended property query structure (legacy)
 *
 */
typedef struct
{
  char  prpName[PROPERTY_NAME_SHORTSIZE]; /**< the property name */
  char  prpDescription[64]; /**< the property description */
  char  prpRedirection[STATUS_SHORTSIZE]; /**< a redirection string if the property is redirected to another server */
  char  prpTag[TAG_NAME_SHORTSIZE];   /**< the data tag for output data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
  char  prpTagIn[TAG_NAME_SHORTSIZE]; /**< the data tag for input data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
  char  prpUnits[16];   /**< a string containing the natural units of the data returned by the property */
  float prpMinValue;    /**< the lower limit of valid data associated with the property */
  float prpMaxValue;    /**< the upper limit of valid data associated with the property */
  UINT32 prpSize;       /**< the maximum allowed size of an output data set associated with the property */
  UINT32 prpSizeIn;     /**< the maximum allowed size of an input data set associated with the property */
  UINT32 prpNumOverloads; /**< the number of overloads associated with this property name  */
  UINT16 prpHistoryDepthShort; /**< the depth of the short term history kept for this property */
  UINT16 prpHistoryDepthLong; /**< the depth of the long term history kept for this property */
  BYTE prpFormat;       /**< the default output data format accepted for this property */
  BYTE prpFormatIn;     /**< is the default input data format accepted for this property */
  BYTE prpAccess;       /**< the allowed data access for this property (either CA_READ, CA_WRITE or both) */
  BYTE prpGraphType;    /**< is the preferred graph type for output data associated with this property, e.g. GT_LIN, GT_LOG, etc */
} PropertyQueryStructEx;
#define PROPERTYQUERYSTRUCTEX_SIZE (PROPERTY_NAME_SHORTSIZE+64+32+TAG_NAME_SHORTSIZE*2+16+4*5+2*2+4)

#define PQSR3_TAG "PQSX"

/**
 * \struct XPropertyQueryStruct
 * \brief Defines an extended property query structure (legacy)
 *
 */
typedef struct
{
  char  prpName[PROPERTY_NAME_SHORTSIZE]; /**< the property name */
  char  prpDescription[64]; /**< the property description */
  char  prpRedirection[STATUS_SHORTSIZE]; /**< a redirection string if the property is redirected to another server */
  char  prpTag[TAG_NAME_SHORTSIZE];   /**< the data tag for output data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
  char  prpTagIn[TAG_NAME_SHORTSIZE]; /**< the data tag for input data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
  char  prpUnits[16];   /**< a string containing the natural units of the data returned by the property */
  float prpMinValue;    /**< the lower limit of valid data associated with the property */
  float prpMaxValue;    /**< the upper limit of valid data associated with the property */
  UINT32 prpSize;       /**< the maximum allowed size of an output data set associated with the property */
  UINT32 prpSizeIn;     /**< the maximum allowed size of an input data set associated with the property */
  UINT32 prpNumOverloads; /**< the number of overloads associated with this property name  */
  UINT16 prpHistoryDepthShort; /**< the depth of the short term history kept for this property */
  UINT16 prpHistoryDepthLong; /**< the depth of the long term history kept for this property */
  BYTE prpFormat;       /**< the default output data format accepted for this property */
  BYTE prpFormatIn;     /**< is the default input data format accepted for this property */
  BYTE prpAccess;       /**< the allowed data access for this property (either CA_READ, CA_WRITE or both) */
  BYTE prpGraphType;    /**< is the preferred graph type for output data associated with this property, e.g. GT_LIN, GT_LOG, etc */
  char  rngUnits[16];   /**< a string containing the natural units of the x-axis (range) in case of a spectrum type array */
  float rngMinValue;    /**< the lower limit of x-axis range associated with the property  */
  float rngMaxValue;    /**< the upper limit of x-axis range associated with the property  */
  UINT16 numRows;       /**< the number of rows in an array property */
  UINT16 rowSize;       /**< the size of a row in an array property */
  UINT16 prpArrayType;  /**< the property array type (AT_NONE, AT_SCALAR, AT_CHANNEL, AT_SPECTRUM, etc.) */
  UINT16 reserved[3];   /**< reserved field) */
} XPropertyQueryStruct;
#define XPROPERTYQUERYSTRUCT_SIZE (PROPERTYQUERYSTRUCTEX_SIZE + 16+2*4+6*2)

#define XPQSR3_TAG "XPQS"

/**
 * \struct PrpQueryStruct
 * \brief Defines the definitive (non-legacy) property query structure
 *
 */
typedef struct
{
  char  prpName[PROPERTY_NAME_SIZE]; /**< the property name */
  char  prpDescription[PROPERTY_DESC_SIZE]; /**< the property description */
  char  prpRedirection[PROPERTY_REDIR_SIZE]; /**< a redirection string if the property is redirected to another server */
  char  prpTag[TAG_NAME_SIZE];   /**< the data tag for output data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
  char  prpTagIn[TAG_NAME_SIZE]; /**< the data tag for input data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
  char  prpUnits[UNITS_SIZE];   /**< a string containing the natural units of the data returned by the property */
  float prpMinValue;    /**< the lower limit of valid data associated with the property */
  float prpMaxValue;    /**< the upper limit of valid data associated with the property */
  UINT32 prpSize;       /**< the maximum allowed size of an output data set associated with the property */
  UINT32 prpSizeIn;     /**< the maximum allowed size of an input data set associated with the property */
  UINT32 prpNumOverloads; /**< the number of overloads associated with this property name  */
  UINT16 prpHistoryDepthShort; /**< the depth of the short term history kept for this property */
  UINT16 prpHistoryDepthLong; /**< the depth of the long term history kept for this property */
  BYTE prpFormat;       /**< the default output data format accepted for this property */
  BYTE prpFormatIn;     /**< is the default input data format accepted for this property */
  BYTE prpAccess;       /**< the allowed data access for this property (either CA_READ, CA_WRITE or both) */
  BYTE prpGraphType;    /**< is the preferred graph type for output data associated with this property, e.g. GT_LIN, GT_LOG, etc */
  char  rngUnits[UNITS_SIZE];   /**< a string containing the natural units of the x-axis (range) in case of a spectrum type array */
  float rngMinValue;    /**< the lower limit of x-axis range associated with the property  */
  float rngMaxValue;    /**< the upper limit of x-axis range associated with the property  */
  UINT16 numRows;       /**< the number of rows in an array property */
  UINT16 rowSize;       /**< the size of a row in an array property */
  UINT16 prpArrayType;  /**< the property array type (AT_NONE, AT_SCALAR, AT_CHANNEL, AT_SPECTRUM, etc.) */
  UINT16 reserved[3];   /**< reserved field) */
} PrpQueryStruct;
#define PRPQUERYSTRUCT_SIZE (PROPERTY_NAME_SIZE+PROPERTY_DESC_SIZE+PROPERTY_REDIR_SIZE+2*TAG_NAME_SIZE+UNITS_SIZE+5*4+2*2+4+UNITS_SIZE+2*4+6*2)

#define PQSR4_TAG "PRPQSr4"

/* helper macros for some routines: */
#define prpqs2pqs(p) {\
  strncpy(pqs[p].prpName,prpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(pqs[p].prpDescription,prpqs[p].prpDescription,32);\
  pqs[p].prpSize = (UINT16)prpqs[p].prpSize;\
  pqs[p].prpFormat = prpqs[p].prpFormat;\
  pqs[p].prpAccess = prpqs[p].prpAccess;\
}
#define prpqs2pqsx(p) {\
  strncpy(pqsx[p].prpName,prpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(pqsx[p].prpDescription,prpqs[p].prpDescription,64);\
  strncpy(pqsx[p].prpRedirection,&prpqs[p].prpRedirection[PRP_REDIR_SRV_OFFSET],16);\
  strncpy(&pqsx[p].prpRedirection[16],&prpqs[p].prpRedirection[PRP_REDIR_PRP_OFFSET],16);\
  strncpy(pqsx[p].prpTag,prpqs[p].prpTag,TAG_NAME_SHORTSIZE);\
  strncpy(pqsx[p].prpTagIn,prpqs[p].prpTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(pqsx[p].prpUnits,prpqs[p].prpUnits,16);\
  pqsx[p].prpMinValue = prpqs[p].prpMinValue;\
  pqsx[p].prpMaxValue = prpqs[p].prpMaxValue;\
  pqsx[p].prpSize = prpqs[p].prpSize;\
  pqsx[p].prpSizeIn = prpqs[p].prpSizeIn;\
  pqsx[p].prpNumOverloads = prpqs[p].prpNumOverloads;\
  pqsx[p].prpHistoryDepthShort = prpqs[p].prpHistoryDepthShort;\
  pqsx[p].prpHistoryDepthLong = prpqs[p].prpHistoryDepthLong;\
  pqsx[p].prpFormat = prpqs[p].prpFormat;\
  pqsx[p].prpFormatIn = prpqs[p].prpFormatIn;\
  pqsx[p].prpAccess = prpqs[p].prpAccess;\
  pqsx[p].prpGraphType = prpqs[p].prpGraphType;\
}
#define prpqs2xpqs(p) {\
  strncpy(xpqs[p].prpName,prpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(xpqs[p].prpDescription,prpqs[p].prpDescription,64);\
  strncpy(xpqs[p].prpRedirection,&prpqs[p].prpRedirection[PRP_REDIR_SRV_OFFSET],16);\
  strncpy(&xpqs[p].prpRedirection[16],&prpqs[p].prpRedirection[PRP_REDIR_PRP_OFFSET],16);\
  strncpy(xpqs[p].prpTag,prpqs[p].prpTag,TAG_NAME_SHORTSIZE);\
  strncpy(xpqs[p].prpTagIn,prpqs[p].prpTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(xpqs[p].prpUnits,prpqs[p].prpUnits,16);\
  xpqs[p].prpMinValue = prpqs[p].prpMinValue;\
  xpqs[p].prpMaxValue = prpqs[p].prpMaxValue;\
  xpqs[p].prpSize = prpqs[p].prpSize;\
  xpqs[p].prpSizeIn = prpqs[p].prpSizeIn;\
  xpqs[p].prpNumOverloads = prpqs[p].prpNumOverloads;\
  xpqs[p].prpHistoryDepthShort = prpqs[p].prpHistoryDepthShort;\
  xpqs[p].prpHistoryDepthLong = prpqs[p].prpHistoryDepthLong;\
  xpqs[p].prpFormat = prpqs[p].prpFormat;\
  xpqs[p].prpFormatIn = prpqs[p].prpFormatIn;\
  xpqs[p].prpAccess = prpqs[p].prpAccess;\
  xpqs[p].prpGraphType = prpqs[p].prpGraphType;\
  strncpy(xpqs[p].rngUnits,prpqs[p].rngUnits,16);\
  xpqs[p].rngMinValue = prpqs[p].rngMinValue;\
  xpqs[p].rngMaxValue = prpqs[p].rngMaxValue;\
  xpqs[p].numRows = prpqs[p].numRows;\
  xpqs[p].rowSize = prpqs[p].rowSize;\
  xpqs[p].prpArrayType = prpqs[p].prpArrayType;\
  memcpy(xpqs[p].reserved,prpqs[p].reserved,6);\
}
#define xpqs2prpqs(p) {\
  memset(&prpqs[p],0,sizeof(PrpQueryStruct));\
  strncpy(prpqs[p].prpName,xpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(prpqs[p].prpDescription,xpqs[p].prpDescription,64);\
  strncpy(&prpqs[p].prpRedirection[PRP_REDIR_SRV_OFFSET],xpqs[p].prpRedirection,16);\
  strncpy(&prpqs[p].prpRedirection[PRP_REDIR_PRP_OFFSET],&xpqs[p].prpRedirection[16],16);\
  strncpy(prpqs[p].prpTag,xpqs[p].prpTag,TAG_NAME_SHORTSIZE);\
  strncpy(prpqs[p].prpTagIn,xpqs[p].prpTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(prpqs[p].prpUnits,xpqs[p].prpUnits,16);\
  prpqs[p].prpMinValue = xpqs[p].prpMinValue;\
  prpqs[p].prpMaxValue = xpqs[p].prpMaxValue;\
  prpqs[p].prpSize = xpqs[p].prpSize;\
  prpqs[p].prpSizeIn = xpqs[p].prpSizeIn;\
  prpqs[p].prpNumOverloads = xpqs[p].prpNumOverloads;\
  prpqs[p].prpHistoryDepthShort = xpqs[p].prpHistoryDepthShort;\
  prpqs[p].prpHistoryDepthLong = xpqs[p].prpHistoryDepthLong;\
  prpqs[p].prpFormat = xpqs[p].prpFormat;\
  prpqs[p].prpFormatIn = xpqs[p].prpFormatIn;\
  prpqs[p].prpAccess = xpqs[p].prpAccess;\
  prpqs[p].prpGraphType = xpqs[p].prpGraphType;\
  strncpy(prpqs[p].rngUnits,xpqs[p].rngUnits,16);\
  prpqs[p].rngMinValue = xpqs[p].rngMinValue;\
  prpqs[p].rngMaxValue = xpqs[p].rngMaxValue;\
  prpqs[p].numRows = xpqs[p].numRows;\
  prpqs[p].rowSize = xpqs[p].rowSize;\
  prpqs[p].prpArrayType = xpqs[p].prpArrayType;\
  memcpy(prpqs[p].reserved,xpqs[p].reserved,6);\
}

/**
 * \internal
 * \brief Structure used to hold property redirection information.
 *
 * \note The 'new' string pattern <server><property><device><context> is
 * consistent with the 'old' pattern <server><property>.
 */
typedef struct PrpRedirBlkStruct
{
  char prpRedirection[STATUS_SIZE];
  int matchAnyDevice;
  int reserved;
  struct PrpRedirBlkStruct *next;
} PrpRedirBlk;
#define RDR_MATCHANYDEVICE "#ANY"

typedef void (*PRPSIG)(int,int,int,int,void *);
typedef void (*CONSIG)(const char *,const char *,const char *,int,int,void *);
extern CONSIG gConSigfcn;
extern int gConSigMask;
extern void *gConSigRef;

#define PRP_XREAD_NONE 0x00
#define PRP_XREAD_LOCK 0x01
#define PRP_XREAD_ALL  0x02

/**
 * \internal
 * \brief access control bits
 */
#define CHECK_USR 0x0001
#define CHECK_NET 0x0002
#define CHECK_GRP 0x0004
#define EQM_CHECK_USR 0x0001
#define EQM_CHECK_NET 0x0002
#define EQM_CHECK_GRP 0x0004
#define PRP_CHECK_USR 0x0010
#define PRP_CHECK_NET 0x0020
#define PRP_CHECK_GRP 0x0040
#define DEV_CHECK_USR 0x0100
#define DEV_CHECK_NET 0x0200
#define DEV_CHECK_GRP 0x0400

#ifndef USE_IPV6
typedef struct sockaddr_in SCKADR;
typedef struct in_addr INADR;
#else
typedef struct sockaddr_storage SCKADR;
typedef struct in6_addr INADR;
#endif

typedef struct
{
  SCKADR addr;
  int mask;
} SCKACL;

/**
 * \internal
 * \brief Structure used to hold access control list info
 */
typedef struct
{
  NAME16 *usrs; /**< user write access list */
  NAME16 *grps; /**< user group write access list */
  SCKACL *nets; /**< net addr write access list */
  int nUsrs;    /**< number of allowed users */
  int nGrps;    /**< number of allowed groups */
  int nNets;    /**< number of allowed ip nets */
} AclInfoType;

/**
 * \internal
 * \brief Structure used to hold exported property information.
 */
typedef struct ExportPropertyList
{
  char prpName[PROPERTY_NAME_SIZE];/**< property name */
  char prpAlias[PROPERTY_NAME_SIZE];/**< property alias name */
  int isReserved;          /**< if TRUE then entry is just a place holder */
  SINT32 prpId;            /**< property id */
  UINT32 prpSize;          /**< maximum supported array size */
  UINT32 prpSizeIn;        /**< maximum supported input array size */
  short prpFormat;         /**< format output data */
  short prpFormatIn;       /**< format input data */
  short prpAccessMode;     /**< bit 0: read access, bit 1: write access */
  short prpArrayType;      /**< property array type if array */
  EQMFCNP prpHndlr;        /**< function pointer to property handler - or NULL */
  void *hdlrRef;           /**< reference pointer to use in extended dispatch functions */
  UINT16 numRows;          /**< num rows if array */
  UINT16 rowSize;          /**< row size if array */
  NAME16 prpFormatTag;     /**< output data structure tag if non zero-length */
  NAME16 prpFormatTagIn;   /**< input data structure tag if non zero-length */
  PrpRedirBlk *prd;        /**< re-director if not NULL */
  NAME64 *devNames;        /**< device list if not NULL */
  int numDevices;          /**< number of devices in device list */
  int contractRenewalLengh;/**< if > 0 then establishes the contract renewal length for contracts using this property */
  AclInfoType aclLst;      /**< property specific acls */
  UINT16 hasExclusiveRead; /**< flag == PRP_XREAD_NONE, _LOCK, _ALL */
  UINT16 minAccessDeadband;/**< if > 0 then minimum allowed time between calls */
  short runInSeparateThread;/**< flag == TRUE -> eqm call in separate thread */
  short enforceDataReq;    /**< flag == TRUE -> contract coercion with invalid_datarequest */
  char prpDescription[PROPERTY_DESC_SIZE];/**< short description of property */
  char *prpUrl;            /**< url to additional information (if not NULL) */
  PrpEgu egu;              /**< EGU information for display */
  PrpEgu xegu;             /**< x-axis EGU information for spectrum display */
  BYTE *prpBuf;            /**< deeply bound data buffer if not NULL */
  BYTE *srBuf;             /**< save/restore data buffer if not NULL */
  UINT32 srBufSiz;         /**< save/restore data buffer size */
  PRPSIG sigfcn;           /**< property signal function if registered */
  void *sigref;            /**< assigned reference to use in a signal function */
  int sigmask;             /**< assigned signal mask to use in a signal function */
  UINT32 numCalls;         /**< number of times newly accessed since startup */
  time_t mcaValidFloor;    /**< minimum client startime where mca indexing is valid*/
  struct ExportPropertyList *next;
} ExportPropertyListStruct;

/* RegisterPropertySignalFunction() */
/**
 * \brief Registers a property signal function
 *
 * If a property is accessed by remote callers, it will be represented within a server's contract
 * list.  The appropriate equipment module will be called, which is the extent of the transaction in
 * most cases.  The server can also optionally receive signals during the course of the transaction
 * by attaching a signal function to the property in question.  Signals will include
 *
 * - \b CS_ACCESS (is being accessed by a new caller),
 * - \b CS_RETRY (is being retried),
 * - \b CS_LATE (is being returned late),
 * - \b CS_PENDING (is being called while last transmission still pending),
 * - \b CS_CALLED (has returned from call to the equipment module dispatch routine),
 * - \b CS_PROCESSED (has returned from dispatch call and has finished processing all returned information),
 * - \b CS_SENT (has been sent to caller),
 *
 * \note the access bits CA_FIRST and CA_LAST can be used with the equipment module to check the
 * scope of the caller's transaction.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is the registered property for the signal function is to be applied.
 * \param fcn is the property signal function of prototype:
 * void sigfcn(int signal,int contractId,int propertyId,int currentStatus,void *reference);
 * \param mask is a signal mask indicating which signals are of interest
 * (use PS_ALL to receive all signals).
 * \param ref is a caller supplied reference which will be returned when the signal function.
 * As this is a void pointer, it can refer to any structure or function the caller wants to have
 * returned when the signal function is called.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * example:
 *
 * \include eg_propertysignal.c
 *
 * \see RegisterPropertyInformation(), SetContractSignalFunction()
 *
 */
TINE_EXPORT int RegisterPropertySignalFunction(const char *eqm,const char *prp,PRPSIG fcn,int mask,void *ref);

/* SetContractSignalFunction() */
/**
 * \brief Registers a contract signal function
 *
 * If a property is accessed by remote callers, it will be represented within a server's contract
 * list.  The appropriate equipment module will be called, which is the extent of the transaction in
 * most cases.  The server can also optionally receive signals during the course of the transaction
 * by attaching a signal function to the property in question (see RegisterPropertySignalFunction()),
 * or by setting up a general contract signal function.  A contract signal function will be called on
 * behalf of any registered property from any registered equipment module. Signals will include
 *
 * - \b CS_ACCESS (is being accessed by a new caller),
 * - \b CS_RETRY (is being retried),
 * - \b CS_LATE (is being returned late),
 * - \b CS_PENDING (is being called while last transmission still pending),
 * - \b CS_CALLED (has returned from call to the equipment module dispatch routine),
 * - \b CS_PROCESSED (has returned from dispatch call and has finished processing all returned information),
 * - \b CS_SENT (has been sent to caller),
 *
 * \note the access bits CA_FIRST and CA_LAST can be used with the equipment module to check the
 * scope of the caller's transaction.
 *
 * \param fcn is the contract signal function of prototype:
 * void sigfcn(const char *eqm,const char *dev,const char *prp,int currentStatus,int signal,void *reference);
 * \param mask is a signal mask indicating which signals are of interest
 * (use PS_ALL to receive all signals).
 * \param ref is a caller supplied reference which will be returned when the signal function.
 * As this is a void pointer, it can refer to any structure or function the caller wants to have
 * returned when the signal function is called.
 *
 * \see RegisterPropertySignalFunction()
 */
TINE_EXPORT void SetContractSignalFunction(CONSIG fcn,int mask,void *ref);

/* AssignDeviceAccessList() */
/**
 * \brief Assigns an access list for the device given.
 *
 * In addtion to the overall access lists for the server process, devices can be individually
 * assigned a user access list.
 *
 * \note As the general access lists constitute the first barrier which must
 * be overcome, any user name assigned to a property's access list but not assigned to the server
 * (if the server has an access list) will not be granted WRITE access.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param dev is the registered device for the access list is to be applied.
 * \param users is a list of user names for whom WRITE access is to be granted.
 * \param nusers is the size of the users list supplied.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyInformation(), AssignDeviceNetsList()
 *
 */
TINE_EXPORT int AssignDeviceAccessList(const char *eqm,const char *dev,NAME16 *users,int nusers);

/* AssignDeviceNetsList() */
/**
 * \brief Assigns an ip nets access list for the device given.
 *
 * In addtion to the overall access lists for the server process, devices can be individually
 * assigned a network access list.
 *
 * \note As the general access lists constitute the first barrier which must
 * be overcome, any user name assigned to a property's access list but not assigned to the server
 * (if the server has an access list) will not be granted WRITE access.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param dev is the registered device for the access list is to be applied.
 * \param ipnets is a list of network names for which WRITE access is to be granted. For
 * example a list entry "192.168.100.18" would grant the host at "192.168.100.18" WRITE
 * access.  The list entry "192.168.100.255" would grant all hosts on the .100 subnet
 * write access, and so on.
 * \param nipnets is the size of the ipnets list supplied.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyInformation(), AssignDeviceAccessList()
 *
 */
TINE_EXPORT int AssignDeviceNetsList(const char *eqm,const char *dev,NAME64 *ipnets,int nipnets);

/* AssignPropertyAccessList() */
/**
 * \brief Assigns an access list for the property given.
 *
 * In addtion to the overall access lists for the server process, properties can be individually
 * assigned a user access list.
 *
 * \note As the general access lists constitute the first barrier which must
 * be overcome, any user name assigned to a property's access list but not assigned to the server
 * (if the server has an access list) will not be granted WRITE access.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is the registered property for the access list is to be applied.
 * \param users is a list of user names for whom WRITE access is to be granted.
 * \param nusers is the size of the users list supplied.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyInformation(), AssignPropertyNetsList()
 *
 */
TINE_EXPORT int AssignPropertyAccessList(const char *eqm,const char *prp,NAME16 *users,int nusers);

/* AssignPropertyNetsList() */
/**
 * \brief Assigns an ip nets access list for the property given.
 *
 * In addtion to the overall access lists for the server process, properties can be individually
 * assigned a network access list.
 *
 * \note As the general access lists constitute the first barrier which must
 * be overcome, any user name assigned to a property's access list but not assigned to the server
 * (if the server has an access list) will not be granted WRITE access.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prp is the registered property for the access list is to be applied.
 * \param ipnets is a list of network names for which WRITE access is to be granted. For
 * example a list entry "192.168.100.18" would grant the host at "192.168.100.18" WRITE
 * access.  The list entry "192.168.100.255" would grant all hosts on the .100 subnet
 * write access, and so on.
 * \param nipnets is the size of the ipnets list supplied.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyInformation(), AssignPropertyAccessList()
 *
 */
TINE_EXPORT int AssignPropertyNetsList(const char *eqm,const char *prp,NAME64 *ipnets,int nipnets);

/* GetPropertyBuffer() */
/**
 * \brief Returns the address of the buffer previously assigned to the property given.
 *
 * Generally an equipment module's dispatch handler is called with a reference to an
 * allocated buffer into which the property's output is written.  In some circumstances
 * (e.g. large video frames) it is prudent to instruct the dispatch handler to make use
 * of a pre-assigned buffer area (which is likely already filled with data) and to forgo
 * any double buffering and memory copying that might otherwise occur.  If a buffer has
 * been assigned, its address can be obtained with this call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName is the registered property for which the buffer space is desired.
 *
 * \return The address of the assigned buffer space (NULL if not assigned).
 *
 * \see SetPropertyBuffer()
 *
 */
TINE_EXPORT BYTE *GetPropertyBuffer(char *eqmName,char *prpName);

/* SetPropertyBuffer() */
/**
 * \brief Assigns a fixed buffer to handle output data for the given property
 *
 * Generally an equipment module's dispatch handler is called with a reference to an
 * allocated buffer into which the property's output is written.  In some circumstances
 * (e.g. large video frames) it is prudent to instruct the dispatch handler to make use
 * of a pre-assigned buffer area (which is likely already filled with data) and to forgo
 * any double buffering and memory copying that might otherwise occur.  This is accomplished
 * with this call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName is the registered property for the buffer is to be assigned.
 * \param buffer is the address of the buffer to assign to the property.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetPropertyBuffer()
 *
 */
TINE_EXPORT int SetPropertyBuffer(char *eqmName,char *prpName,BYTE *buffer);

/* RegisterPropertyQueryFunction() */
/**
 * \brief Registers a property query function
 *
 * If a server's properties are all registered via a call to RegisterPropertyInformation(),
 * RegisterPropertyInformationEx(), RegisterProperty(), etc.,
 * then there is no need to deal with property query functions.  Some servers may wish to
 * have a different property list for any given device, in which case a one-time query of the list of
 * properties will not apply in general.  The suggested procedure is to register all properties
 * that a device server might have to handle and then assign the relevant 'lists' of properties to
 * those devices which require a subset (e.g. AssignDeviceListToProperty()).
 * In this manner, the TINE kernel can maintain all known properties in a registry.
 * Where this is impractical, a server can also register its own property query structure,
 * which is required to field all property queries explicitly.
 * The query function must have the propotype shown in the parameter 'fcn'.  That is, it should be a routine
 * which accepts a device name as a parameter, and returns via a second paramter a reference to a list of
 * extended property query structures (PrpQueryStruct).
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param fcn is the property query function which will handle property information queries
 * \param numprops is the number of properties which are registered.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see AssignDeviceListToProperty()
 *
 * \b Example:
 *
 * \include eg_RegisterPropertyQueryFunction.c
 */
TINE_EXPORT short RegisterPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,PrpQueryStruct **prpqs),int numprops);

/* RegisterXPropertyQueryFunction() */
/**
 * \brief Registers an extended property query function
 *
 * \deprecated Please use RegisterPropertyQueryFunction() instead
 *
 * If a server's properties are all registered via a call to RegisterPropertyInformation(), RegisterPropertyEx(),
 * or RegisterProperty(), then there is no need to deal with property query functions.  Some servers may wish to
 * have a different property list for any given device, in which case a one-time query of the list of
 * properties will not apply in general.  This is infrequently the case, but when it arises, a server will need
 * to field all property queries explicitly by registering its own property query structure.
 * The query function must have the propotype shown in the parameter 'fcn'.  That is, it should be a routine
 * which accepts a device name as a parameter, and returns via a second paramter a reference to a list of
 * extended property query structures (PropertyQueryStructEx).
 *
 * \note This extended call deals with lists of the more modern extended property query structure as opposed to
 * the legacy 'simple' property query structure.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param fcn is the extended property query function which will handle property information queries
 * \param numprops is the number of properties which are registered.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyQueryFunction()
 *
 */
TINE_EXPORT short RegisterXPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,XPropertyQueryStruct **xpqs),int numprops);

int RegisterLocalAlias(char *eqmName,char *alias);

/* SetCallPropertyInSeparateThread() */
/**
 * \brief Determines whether the specified property is called in a separate handler thread or not.
 *
 * Some equipment module properties might take a non-negligible time to complete.  In general the equipment module
 * should try to minimize the time spent in the dispatch handler (i.e. simply copying results into the request buffer),
 * but this is often not possible if long calculations need to be performed or hardware access is required, etc.
 * If it is known before hand that this is the case it is advisable to specify that the property in question is called
 * on a special dispatch handler thread (avialable only in the multi-threaded library build).  Other properties to the
 * equipment modules will not be called while such a call is underway, as it can not be known before hand if the equiment
 * module is thread safe. However the caller will receive an 'operation busy'
 * notification rather than a 'time out' if the callers timeout limit was exceeded.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param property is the property which is to be called on a separate dispatch handler thread.
 * \param value determines whether the property is called in a separate thread (TRUE) or not (FALSE).
 * If specifically CA_WRITE|CA_INDEPENDENT (0x8002) is passed, then the property
 * will also be flagged so as to not block other concurrent equipment module calls.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyInformation()
 *
 */
TINE_EXPORT int SetCallPropertyInSeparateThread(char *eqm,char *property,int value);

/* AssignNumDevices() */
/**
 * \brief Fixes the number of modules (devices) attached to the specified equipment module
 *
 * Servers generally do not need to make this call, as the number of modules or devices is either specified
 * in the local database file 'exports.csv', or has been passed as an argument in a call to
 * RegisterEquipmentModule().  However, in cases where the number so assigns represents an upper limit, and the
 * actual number of modules is only known after the server is running, then a call to AssignNumDevices() can
 * be made to reduce the querieable number of modules to this number.
 *
 * \note An upper limit of the number of devices (capacity) must have already been specified either through the exports.csv
 * file or through a call to RegisterEquipmentModule().  The number specied in a call to AssignNumDevices()
 * must be smaller than this number.  The capacity can however be increased (if memory is available) via a call to
 * SetSizeDeviceCapacity().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param num is the number of modules (devices) to be associated with the equipment module
 *
 * \return 0 is successful or a TINE completion code
 *
 * \see SetSizeDeviceCapacity()
 */
TINE_EXPORT int AssignNumDevices(char *eqm,int num);
#define AssignNumModules(eqm,num) AssignNumDevices(eqm,num)

/* SetSizeDeviceCapacity() */
/**
 * \brief Sets (increases) the maximum size of the device table and associated tables.
 *
 * Servers spcecify the maximum device capacity with the equipment module registration, either through
 * exports.csv, fec.xml or a call to RegisterEquipmentModule().  In some dynamic situations, a server
 * might note the available capacity is too small and can augment it via this API call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param size is the maximum number (capacity) of allowed devices.  If the 'size' passed is smaller
 * than the current device list capacity, nothing happens, and the call returns 'already_assigned'.
 * If the call is made before the initial capacity has been determined (through equipment module
 * registration), the call returnes 'not_initialized'.
 *
 * \return 0 is successful or a TINE completion code
 *
 * \see AssignNumDevices()
 */
TINE_EXPORT int SetSizeDeviceCapacity(char *eqm,int size);

/* IsServerRunning() */
/**
 * \brief Returns TRUE (non zero) if the server kernel is fully initialized and ready
 *
 * \return 0 if not running
 */
TINE_EXPORT int IsServerRunning(void);

/* FindNameServerOnNetwork() */
/**
 * \brief Issues a multicast (or broadcast) to which the TINE equipment name server responds
 *
 * In lieu of accessing the file 'cshosts.csv' (whose location is determined by the
 * environment variable TINE_HOME) a server can make this call to ascertain the
 * address(es) of the Equipment Name Server (ENS).
 * This is an option for platforms which do
 * not have a hard disk and do not know a-priori the address of the equipment name server.
 * This call will in fact be made automatically as a last resort to find the ENS address if
 * (1) there is no cshost.csv information, (2) the ENS address was not explicitly assigned, and
 * (3) there is no DNS entry for 'tineens' (or 'tineens1', 'tineens2', 'tineens3') on the local
 * domaine.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \note A route for the services multicast group from the ENS to the host using this call
 * must exist in order for this call to be successful.
 *
 * \see SetNameServerAddress()
 */
TINE_EXPORT int FindNameServerOnNetwork(void);

/* SetNameServerAddress() */
/**
 * \brief Sets the address of the Equipment Name Server via API call
 *
 * In lieu of accessing the file 'cshosts.csv' (whose location is determined by the
 * environment variable TINE_HOME) a server can make this call to establish the
 * address(es) of the Equipment Name Server (ENS).  This is convenient for platforms which do
 * not have a hard disk.
 *
 * \param ens a string containing a list of the IP addresses of the ENSes to be used,
 * separated by commas (','), and optionally providing a port offset (separated by a colon ':').
 *
 * \b Example (sets two ENS addresses, the second with port offset = 2):
 *
 * SetNameServerAddress("131.169.120.141,131.169.120.146:2");
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int SetNameServerAddress(char *ens);

/* SetAllowNetworkAddressResolution() */
/**
 * \fn TINE_EXPORT void SetAllowNetworkAddressResolution(int value)
 * \brief Determines whether NETWORK address resolution is allowed
 *
 * In an ENS-less (minimal) infrastructure, an initializing client can
 * attempt to find server addresses by sending a NETWORK (multicast) probe
 * to ascertain a server's address.  In lieu of an ENS configuration, this requires a
 * running server (configured to respond to NETWORK requests).  Otherwise the
 * servers address cannot be resolved.
 *
 * \param value is the desired setting (default = FALSE)
 *
 * \note this value can also be determined by environment
 * TINE_NETWORKADDRESS_RESOLUTION = TRUE
 *
 * \see GetAllowNetworkAddressResolution()
 */
TINE_EXPORT void SetAllowNetworkAddressResolution(int value);

/* GetAllowNetworkAddressResolution() */
/**
 * \fn TINE_EXPORT int GetAllowNetworkAddressResolution(void)
 * \brief returns whether NETWORK address resolution is allowed
 *
 * In an ENS-less (minimal) infrastructure, an initializing client can
 * attempt to find server addresses by sending a NETWORK (multicast) probe
 * to ascertain a server's address.  In lieu of an ENS configuration, this requires a
 * running server (configured to respond to NETWORK requests).  Otherwise the
 * servers address cannot be resolved.
 *
 * \note this value can also be determined by environment
 * TINE_NETWORKADDRESS_RESOLUTION = TRUE
 *
 * \return the current setting (default = FALSE)
 * \see SetAllowNetworkAddressResolution()
 */
TINE_EXPORT int GetAllowNetworkAddressResolution(void);

/* GetNumContracts() */
/**
 * \brief Returns the current number of contracts associated with the equipment module given
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The total number of contracts associated with the given equipment module.  If a
 * NULL parameter is passed, the function returns the total number of contracts for all
 * registered equipment modules.
 *
 * \b Example:
 *
 * \include  eg_GetNumContracts.c
 */
TINE_EXPORT int GetNumContracts(char *eqm);

/* GetNumConsumers() */
/**
 * \brief Returns the current number of consumers associated with the equipment module given
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The total number of consumers (i.e. clients) associated with the given equipment module.  If a
 * NULL parameter is passed, the function returns the total number of consumers for all
 * registered equipment modules.  This call is synonymous with GetNumCallers().
 *
 * \b Example:
 *
 * \include  eg_GetNumConsumers.c
 */
TINE_EXPORT int GetNumConsumers(char *eqm);

/* GetNumCalls() */
/**
 * \brief Returns the current number of calls to the given equipment module since startup
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The total number of calls to the given equipment module since startup.
 *
 */
TINE_EXPORT int GetNumCalls(char *eqm);

/* GetNumCallers() */
/**
 * \brief Returns the current number of callers associated with the equipment module given
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return The total number of callers (i.e. clients) associated with the given equipment module.  If a
 * NULL parameter is passed, the function returns the total number of consumers for all
 * registered equipment modules.  This call is synonymous with GetNumConsumers().
 *
 */
TINE_EXPORT int GetNumCallers(char *eqm);

/* GetCurrentLinkStatus() */
/**
 * \brief Returns the completion code associated with the input link index
 *
 * \param i is the link index of the link for which the current completion code is desired.
 *
 * \return The completion code associated with the input link index
 *
 * \b Example:
 *
 * \include eg_GetCurrentLinkStatus.c
 */
TINE_EXPORT int GetCurrentLinkStatus(int i);

/* GetCurrentDataStatus() */
/**
 * \brief Returns the data status code associated with the input link index
 *
 * If the remote equipment module is called, this is the return code delivered.  If the link
 * returns or notifies the caller and the equipment module has not been called (e.g. in
 * case of time out)
 * this value will be 'not posted'.
 *
 * \param i is the link index of the link for which the current completion code is desired.
 *
 * \return The status code associated with the input link index
 *
 * \see GetCurrentLinkStatus()
 *
 */
TINE_EXPORT int GetCurrentDataStatus(int i);

/* GetCurrentLinkStatusFromCallbackId() */
/**
 * \brief Returns the completion code associated with the
 * callback identifier supplied
 *
 * \param id is the callback id of the link for which the current completion code is desired.
 *
 * \return The completion code associated with the callback identifier supplied
 *
 * \b Example:
 *
 * \include eg_GetCurrentLinkStatusFromCallbackId.c
 */
TINE_EXPORT int GetCurrentLinkStatusFromCallbackId(int id);

/* GetCurrentDataStatusFromCallbackId() */
/**
 * \brief Returns the data status code associated with the
 * callback identifier supplied
 *
 * If the remote equipment module is called, this is the return code delivered.  If the link
 * returns or notifies the caller and the equipment module has not been called (e.g. in
 * case of time out)
 * this value will be 'not posted'.
 *
 * \param id is the callback id of the link for which the current completion code is desired.
 *
 * \return The data status code associated with the callback identifier supplied
 *
 * \see GetCurrentLinkStatusFromCallbackId().
 *
 */
TINE_EXPORT int GetCurrentDataStatusFromCallbackId(int id);

/* GetCaller() */
/**
 * \brief Gives the user name of the current caller
 *
 * If a call to an equipment module is made on behalf of more than one user, then GetCaller() only
 * refers to the first in the list.  Typically, GetCaller() is used to see who is making a WRITE access
 * call, which is usually synchronous and has only one user attached.  When more information is required
 * or the contract is liable to be monitored by many clients, then GetCallerInfo() should be used instead.
 *
 * \param eqmName (in) is the equipment function name (local name) for which the exported name
 * is desired.
 *
 * \return The user name of the current caller
 *
 * \see GetCallerInformation()
 *
 * \b Example:
 *
 * \include eg_GetCaller.c
 */
TINE_EXPORT char *GetCaller(char *eqmName);

/* GetCallerInfo() */
/**
 * \brief Returns the user name(s) and network address(es) of all callers interested in the current contract
 *
 * \param eqmName (in) is the equipment function name (local name) for which the exported name
 * is desired.
 * \param un (out) if non-NULL is a reference to an array of NAME16 objects to be filled with the caller list
 * \param ipx (out) if non-NULL is a reference to a byte array to be filled with the IPX addresses of the caller(s).
 * \param ip (out) if non-NULL is a reference to an unsigned long array to be filled with the IP addresses of the caller(s).
 * \param prot (out) if non-NULL is a reference to a short array to be filled with the requested protocol level of the caller(s).
 * \param num (in/out) is a reference to an int value which initially contains the buffer size of the array
 * references used in the parameter list. Upon return of the call, it contains the actual number of callers.
 *
 * \note The non-NULL arguments supplied must point to buffer with enough room to hold at least the input 'num'
 * entries, otherwise a system crash is likely.
 * \note This call has been deprecated in favor of GetCallerInformation
 *
 * \return 0 if successful, or a TINE completion code
 *
 * \see GetCallerInformation()
 *
 * \b Example:
 *
 * \include eg_GetCallerInfo.c
 */
TINE_EXPORT int GetCallerInfo(char *eqmName,NAME16 *un,BYTE *ipx,UINT32 *ip,short *prot,int *num);

/* ResubmitDataLink() */
/**
 * \brief Resubmits a (CM_SINGLE) data link
 *
 * This routine is useful for asynchronous single links which should be
 * 'retried' following a link timeout notification in the callback.
 *
 * \param id is the link index for the link which is to be resubmitted
 *
 * \return 0 upon success or invalid_link
 *
 */
TINE_EXPORT int ResubmitDataLink(int id);

/* GetCompletionDataSize() */
/**
 * \brief Returns the most recent data size of the link index supplied
 *
 * \param i is the link index for which the completed data size is desired
 *
 * \return The most recent data size of the link index supplied
 *
 * \see GetCompletionDataSizeFromCallbackId()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataSize.c
 */
TINE_EXPORT int GetCompletionDataSize(int i);
#define getCompletionDataSize(i) GetCompletionDataSize(i)

/* GetCompletionDataSizeFromCallbackId() */
/**
 * \brief Returns the most recent data size of the link
 * associated with the callback id supplied
 *
 * \param id is the callback id of the link for which the completed data size is desired
 *
 * \return The most recent data size of the link associated with the callback id supplied
 *
 * \see GetCompletionDataSize()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataSizeFromCallbackId.c
 */
TINE_EXPORT int GetCompletionDataSizeFromCallbackId(int id);
#define getCompletionDataSizeFromCallbackId(i) GetCompletionDataSizeFromCallbackId(i)

/* GetCompletionDataType() */
/**
 * \brief Returns the most recent data type of the link index
 * supplied
 *
 * \param i is the link index for which the data type is desired
 *
 * \return The most recent data type of the link index supplied
 *
 * \see GetCompletionDataTypeFromCallbackId()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataType.c
 */
TINE_EXPORT int GetCompletionDataType(int i);
#define getCompletionDataType(i) GetCompletionDataType(i)

/* GetCompletionDataTypeFromCallbackId() */
/**
 * \brief Returns the most recent data type of the link
 * associated with the callback id supplied
 *
 * \param id is the callback id of the link for which the completed data type is desired
 *
 * \return The most recent data type of the link associated with the callback id supplied
 *
 * \see GetCompletionDataType()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataTypeFromCallbackId.c
 */
TINE_EXPORT int GetCompletionDataTypeFromCallbackId(int id);
#define getCompletionDataTypeFromCallbackId(i) GetCompletionDataTypeFromCallbackId(i)

/* GetCompletionSource() */
/**
 * \brief Returns the error source associated with the input link index
 *
 * \param i is the link index of the link for which the current status string is desired.
 *
 * \return The error source associated with the input link index as one of LNK_ERROR_LCL
 * or LNK_ERROR_RMT, or -1 if the link index is invalid.
 *
 * \see GetCompletionSourceFromCallbackId()
 *
 */
TINE_EXPORT short GetCompletionSource(int i);
#define getCompletionSource(i) GetCompletionSource(i)

/* GetCompletionStatus() */
/**
 * \brief Returns the error string associated with the input link index
 *
 * \param i is the link index of the link for which the current status string is desired.
 *
 * \return The error string associated with the input link index, or NULL if the
 * the link index is invalid.
 *
 * \see GetCompletionStatusFromCallbackId()
 *
 * \b Example:
 *
 * \include eg_GetCompletionStatus.c
 */
TINE_EXPORT char *GetCompletionStatus(int i);
#define getCompletionStatus(i) GetCompletionStatus(i)
TINE_EXPORT int FillCompletionStatusString(int i,char *errstr);

/* GetCompletionSourceFromCallbackId() */
/**
 * \brief Returns the error source associated with the callback identifier supplied.
 *
 * \param id is the callback id of the link for which the current status string is desired.
 *
 * \return The error source associated with the callback identifier supplied as one of
 * LNK_ERROR_LCL or LNK_ERROR_RMT, or -1 if the callback identifier is invalid.
 *
 */
TINE_EXPORT short GetCompletionSourceFromCallbackId(int id);
#define getCompletionSourceFromCallbackId(i) GetCompletionSourceFromCallbackId(i)

/* GetCompletionStatusFromCallbackId() */
/**
 * \brief Returns the error string associated with the
 * callback identifier supplied
 *
 * \param id is the callback id of the link for which the current status string is desired.
 *
 * \return The error string associated with the callback identifier supplied, or NULL
 * if the callback id is invalid.
 *
 * \see GetCompletionStatus()
 *
 * \b Example:
 *
 * \include eg_GetCompletionStatusFromCallbackId.c
 */
TINE_EXPORT char *GetCompletionStatusFromCallbackId(int id);
#define getCompletionStatusFromCallbackId(i) GetCompletionStatusFromCallbackId(i)
TINE_EXPORT int FillCompletionStatusStringFromCallbackId(int id,char *errstr);

/* GetLinkName() */
/**
 * \brief Returns the full link name associated with the
 * input link index
 *
 * \param idis the link index of the link for which the link name is desired.
 *
 * \return The full link name as "/<Context>/<Server>/<Device>[<Property>]", or
 * "link not found" if the callback id is invalid.
 *
 */
TINE_EXPORT char *GetLinkName(int i);
#define getLinkName(i) GetLinkName(i)

/* GetLinkNameFromCallbackId() */
/**
 * \brief Returns the full link name associated with the
 * callback identifier supplied
 *
 * \param id is the callback id of the link for which the link name is desired.
 *
 * \return The full link name as "/<Context>/<Server>/<Device>[<Property>]", or
 * "link not found" if the callback id is invalid.
 *
 */
TINE_EXPORT char *GetLinkNameFromCallbackId(int id);
#define getLinkNameFromCallbackId(i) GetLinkNameFromCallbackId(i)

/* GetExportedName() */
/**
 * \brief Returns the exported name associated with the equipment function name given.
 *
 * A server as coded, frequently does not know its system wide identity (i.e. its 'export name'),
 * as this is information is contained in a registration file such as 'exports.csv'.  Nonetheless
 * it might be desireable to know what name was actually exported, for use in say conditional
 * trapping or logging.  This call returns the name finally exported after the server has been
 * initialized.
 *
 * \param eqmName is the equipment function name (local name) for which the exported name
 * is desired.
 *
 * \return If successful, the export name associated with the equipment function name supplied.
 * Otherwise a NULL pointer is returned.
 *
 * \see GetCompletionStatus()
 */
TINE_EXPORT char *GetExportedName(char *eqmName);

/* GetExportedContext() */
/**
 * \brief Returns the exported context associated with the equipment function name given.
 *
 * A server as coded, frequently does not know the context under which it is registered,
 * as this is information is often contained in a registration file such as 'fecid.csv' or 'exports.csv'.
 * Nonetheless
 * it might be desireable to know what name was actually exported, for use in say conditional
 * trapping or logging.  This call returns the name finally exported after the server has been
 * initialized.
 *
 * \param eqmName is the equipment function name (local name) for which the exported context
 * is desired.
 *
 * \return If successful, the export context associated with the equipment function name supplied.
 */
TINE_EXPORT char *GetExportedContext(char *eqmName);

/**
 * \brief Assigns the exported context associated with the equipment function name given.
 *
 * Calling this function after the equipment module is registered will have no effect.  Any
 * argument list error will be logged but otherwise will have no effect.
 *
 * \param eqmName is the equipment function name (local name) for which the exported context
 * should be set.
 * \context is the desired context
 *
 * \b Example:
 *
 * \include eg_RegisterEquipmentModule.c
 */
TINE_EXPORT void SetExportedContext(char *eqmName,char *context);

/* GetExportedSubSystem() */
/**
 * \brief Returns the exported subsystem associated with the equipment function name given.
 *
 * A server as coded, frequently does not know the subsystem under which it is registered,
 * as this is information is often contained in a registration file such as 'fecid.csv' or 'exports.csv'.
 * Nonetheless
 * it might be desireable to know what name was actually exported, for use in say conditional
 * trapping or logging.  This call returns the name finally exported after the server has been
 * initialized.
 *
 * \param eqmName is the equipment function name (local name) for which the exported subsystem
 * is desired.
 *
 * \return If successful, the export subsystem associated with the equipment function name supplied.
 */
TINE_EXPORT char *GetExportedSubSystem(char *eqmName);

/**
 * \brief Assigns the exported subsystem associated with the equipment function name given.
 *
 * Calling this function after the equipment module is registered will have no effect.  Any
 * argument list error will be logged but otherwise will have no effect.
 *
 * \param eqmName is the equipment function name (local name) for which the exported subsystem
 * should be set.
 * \subsystem is the desired sub system
 *
 */
TINE_EXPORT void SetExportedSubSystem(char *eqmName,char *subsystem);

/**
 * \brief Sets the default device region to the value specified.
 *
 * This region code will be used in cases where it has not been overridden by a specific device.
 *
 * \param eqmName is the equipment function name (local name) for which the exported subsystem
 * is desired.
 * \param region is the desired default device region tag
 *
 */
TINE_EXPORT void SetDefaultDeviceRegion(char *eqmName,char *region);

/**
 * \brief Returns the default device region
 *
 * \param eqmName is the equipment function name (local name) for which the exported subsystem
 * is desired.
 *
 * \return If successful, the default device region code (used if not overridden by a specific device).
 */
TINE_EXPORT int GetDefaultDeviceRegion(char *eqmName);

/* GetExportedFecName() */
/**
 * \brief Returns the FEC name exported at the time of server initialization.
 *
 * \return If successful, the fec name registered at the equipment name server.
 */
TINE_EXPORT char *GetExportedFecName(void);

/* GetLastGlobalDataSize() */
/**
 * \brief Returns the most recent data size of the global link index supplied
 *
 * \param i is the global link index for which the data size is desired.
 * If i is -1 the last recorded incoming global data size (not a thread-safe thing to do).
 *
 * \return The most recent data size of the global link index supplied
 *
 * \see GetLastGlobalDataSizeFromCallbackId()
 *
 */
TINE_EXPORT int GetLastGlobalDataSize(int id);

/* GetGlobalDataType() */
/**
 * \brief Returns the data type of global link index supplied
 *
 * This call is useful if a globals link was started with the target
 * data format = CF_DEFAULT (and a sufficient buffer size was provided).
 * In such cases the global data size can be obtained by calling
 * GetLastGlobalDataSize().
 *
 * \param i is the global link index for which the data type is desired.
 *
 *
 * \return The data type of the global in question.
 *
 * \see GetLastGlobalDataSize()
 *
 */
TINE_EXPORT int GetGlobalDataType(int id);

/* GetLastGlobalDataSizeFromCallbackId() */
/**
 * \brief Returns the most recent data size of the global link callback id supplied
 *
 * \param id is the global link callback id for which the data size is desired.
 *
 * \return The most recent data size of the global link callback id supplied
 *
 * \note This is a best effort to determine the propery entry table for the callback id
 * supplied as an argument.  It is the duty of the caller to manage a unique  set of
 * callback ids for his application.
 *
 * \see GetLastGlobalDataSize()
 *
 */
TINE_EXPORT int GetLastGlobalDataSizeFromCallbackId(int id);

/* GetGlobalDataTypeFromCallbackId() */
/**
 * \brief Returns the data type of global link callback id supplied
 *
 * This call is useful if a globals link was started with the target
 * data format = CF_DEFAULT (and a sufficient buffer size was provided).
 * In such cases the global data size can be obtained by calling
 * GetLastGlobalDataSize().
 *
 * \param id is the global link callback id passed in the initial call to
 * recvNetGlobalEx() or recvNetGlobalEx2().
 *
 *
 * \return The data type of the global in question.
 *
 * \see GetLastGlobalDataSize()
 *
 */
TINE_EXPORT int GetGlobalDataTypeFromCallbackId(int id);

/* GetRegisteredPropertyListStruct() */
/**
 * \brief Returns the ExportPropertyListStruct structure for the registered property
 * specified by the first two arguments.
 *
 * A server can retieve the ExportPropertyListStruct structure of the input registered
 * property by making this.  This will give the full information pertaining to the
 * property given.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName (input) is the name of the property for which the ExportPropertyListStruct
 * structure is desired.
 * \param prpLstStruct (output) contains the output structure when the call completes.
 *
 * \return '0' upon success, othereise a tine return code.
 *
 * \see GetRegisteredPropertyList()
 *
 */
TINE_EXPORT int GetRegisteredPropertyListStruct(char *eqm,char *prpName,ExportPropertyListStruct *prpLstStruct);

/* RegisteredPropertyIsWritable() */
/**
 * \brief Returns TRUE if the given property is registered with the CA_WRITE access
 * bit (in at least one overload).  Otherwise the functionr returns FALSE.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpName (input) is the name of the property for which the WRITE access is to be
 * determined.
 *
 * \return TRUE if the given property is registered with the CA_WRITE access bit, otherwise FALSE.
 *
 * \see GetRegisteredPropertyList()
 *
 */
TINE_EXPORT int RegisteredPropertyIsWritable(char *eqm,char *prpName);

/* SystemVersion() */
/**
 * \brief Returns the system version as a 4-byte array containg the major and minor version numbers as
 *  the first two bytes and the current revision in the next two bytes.
 *
 * \return The tine version in use.
 *
 * \b Example:
 *
 * \include eg_SystemVersion.c
 */
TINE_EXPORT BYTE *SystemVersion(void);

/* GetSystemVersionString() */
/**
 * \brief Returns the system version appended with the library build id as a character string.
 *
 * \return The tine version + build id;
 *
 */
TINE_EXPORT char *GetSystemVersionString(void);

extern char gFeclogPath[STD_PATH_LENGTH];
/* feclog() */
/**
 * \brief Puts entries into a server's FEC log file.
 *
 * Servers by default maintain a log file (fec.log) which contains system information at startup
 * and shutdown time, as well as WRITE access command calls, as well as other notifications.
 * The server can all place its own text in the server's log file by using this routine.  All
 * entries will be pre-pended with a time stamp and the FEC NAME of the server providing the
 * log entry.  A new-line character will be appended if the supplied string does not already
 * end in a new line.
 *
 * \param text is the text string (up to 200 characters) to be logged.
 *
 * \return 0 if successful, otherwise a TINE error code.
 *
 * \see GetCompletionStatus()
 * \see gFeclogPath
 * \see vfeclog()
 * \see SetFeclogDepth()
 * \see srvlog()
 */
TINE_EXPORT int feclog(char *text,...);
/* feclogEx() */
/**
 * \brief Puts entries into a server's FEC log file and in error.log file if certain conditions match
 *
 * Servers by default maintain a log file (fec.log) which contains system information at startup
 * and shutdown time, as well as WRITE access command calls, as well as other notifications.
 * The server can all place its own text in the server's log file by using this routine.  All
 * entries will be pre-pended with a time stamp and the FEC NAME of the server providing the
 * log entry.  A new-line character will be appended if the supplied string does not already
 * end in a new line.
 * In this extended call an int constant parameter indicates if entry should also be logged to error.log
 * This is done, if and only if loglevel parameter is greater than global gErrorLogLevel. Loglevel constants
 * are defined in errors.h: TINE_LOGLEVEL_INFO ..._WARN ..._ERROR ..._FATAL (FATAL is largest value).
 *
 * \param text is the text string (up to 200 characters) to be logged.
 * \param loglevel indicates if text should be written to error.log file
 *
 * \return 0 if successful, otherwise a TINE error code.
 *
 * \see GetCompletionStatus()
 * \see gFeclogPath
 * \see vfeclog()
 * \see SetFeclogDepth()
 * \see srvlog()
 */
TINE_EXPORT int feclogEx(int loglevel, char *text,...);

/* gErrorLogLevel */
/**
 * \brief Determines which messages are written to error.log file
 *
 * by default it is set to TINE_LOGLEVEL_WARN, so WARN, ERROR and FATAL are written to error.log
 *
 * \see feclogEx()
 */
extern TINE_EXPORT int gErrorLogLevel;

/* srvlog() */
/**
 * \brief Puts entries into a named log file.
 *
 * Servers can write log entries to any log file of their choosing by making use
 * of this call.
 * All entries will be pre-pended with a time stamp and the given tag and
 * appended with a new-line character.
 *
 * \param filename is the desired file name. If this name does not specify a full path
 * then the FEC_HOME location determines the location of the log file.
 * Likewise, if the filename does not specify an extension, the extension ".log" will
 * be automatically appended.
 * \param tag is an optional tag to be applied to each log entry, following the timestamp.
 * This can be used to specify the routine or process supplying the log information, for
 * example.  If not specified, the tag "Logger" will be used.
 * \param text is the text string (up to 200 characters) to be logged.
 *
 * \return 0 if successful, otherwise a TINE error code.
 *
 * \see feclog()
 * \see GetCompletionStatus()
 * \see gFeclogPath
 * \see vfeclog()
 * \see SetFeclogDepth()
 */
TINE_EXPORT int srvlog(char *filename,char *tag,char *text,...);

/**
 * \brief Determines the approximate size of the server's log file in lines.
 *
 * The server's log file, if located on the local disk, is a file called 'fec.log'.  When it's length
 * surpasses the FecLogDepth, this file is rotated to 'fec.bak' and a new 'fec.log' file is started.  A
 * call for the server's log file will always be able to scan both the 'fec.bak' file and 'fec.log', and
 * will thus always be able to scan at least FeclogDepth entries.
 * The log file depth is controlled via this API call.
 *
 * \param value is the desired log depth in lines of fec.log.  Default: 100. Minimum: 100.
 *
 */
TINE_EXPORT void SetFeclogDepth(int value);

/**
 * \brief Returns the maximum size of the server's log file in lines.
 *
 * The server's log file, if located on the local disk, is a file called 'fec.log'.  When it's length
 * surpasses the FecLogDepth, this file is rotated to 'fec.bak' and a new 'fec.log' file is started.  A
 * call for the server's log file will always be able to scan both the 'fec.bak' file and 'fec.log', and
 * will thus always be able to scan at least FeclogDepth entries.
 *
 * The value of the log depth also determines the depth of any named log file (see srvlog()).
 *
 * \return the current setting.
 *
 */
TINE_EXPORT int GetFeclogDepth(void);

/* ExecLocalLink() */
/**
 * \brief Executes a synchronous link within the local process
 *
 * Synchronous data exchange within a process. ExecLocalLink() does not complete until the data transfer
 * has completed or a timeout has ensued. ExecLocalLink() has marginal applicability other than processing
 * and returning the result of a call within a server, which could be of use if multiple equipment functions
 * are bound together or a server's background task needs to know the outcome of a call into the server, etc.
 *
 * \param devName is the local device name, prefixed with the corresponding local equipment module name.
 (<Equipment Module Name\>/\<Device\>) of the device to contact.
 * For example: "BPMEQM/WL167".
 * \note Using ExecLink() with context "LOCALHOST" or "LOCAL" and the equipment module name also reverts to a
 * call to ExecLocalLink().
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server
 * to the client.  The structure element ‘dout->data’ should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of TINE access code ORed
 * together (e.g. CA_READ|CA_WRITE).
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see ExecLink(), ExecLinkEx(), DTYPE
 *
 */
TINE_EXPORT int ExecLocalLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access);

/* getDataTimeString() */
/**
 * \brief Returns the TINE data timestamp as a human-readable string
 *
 * A call to getDataTimeStamp() returns an 8-byte double containing the current
 * TINE system timestamp. This will be the UTC timestamp appended with
 * milliseconds after the decimal (if available).  This timestamp can be
 * converted to a human readable string with this call.  The format will depend
 * on the value of the second parameter.  If useLongStringFormat is non-zero,
 * the a date of the form:
 *
 * "Mon Jan 10 12:08:43.355 CET 2005"
 *
 * will be returned.  If useLongStringFormat is zero, then the date string will
 * be of the form (dd.mm.yy hh:mm:ss.msec TZ):
 *
 * "10.01.05 12:08:43.355 CET"
 *
 * \param ts is the TINE Data Timestamp for which the string conversion is to be
 * applied.
 * \param useLongStringFormat is a flag which if non-zero returns a longer representation
 * of the timestamp.
 *
 * \note This call is not thread-safe.  See the discussion in getDataTimeStamp() for
 * details.
 *
 * \return A pointer to a string containing the timestamp representation.
 *
 * \see MakeDataTimeStamp
 * \see GetDataTimeStamp, GetCurrentDataTimeStamp, GetCurrentDataTimeStampFromCallbackId
 * \see ctime()
 * \see localtime()
 * \see findDaylight()
 */
TINE_EXPORT char *GetDataTimeString(double ts,int useLongStringFormat);
#define getDataTimeString(a1,a2) GetDataTimeString(a1,a2)

/**
 * \brief Returns the TINE data timestamp as a human-readable string
 *
 * A call to getDataTimeStamp() returns an 8-byte double containing the current
 * TINE system timestamp. This will be the UTC timestamp appended with
 * milliseconds after the decimal (if available).  This timestamp can be
 * converted to a human readable string with this call.  The format will depend
 * on the value of the second parameter.  If useLongStringFormat is non-zero,
 * the a date of the form:
 *
 * "Mon Jan 10 12:08:43.355 CET 2005"
 *
 * will be returned.  If useLongStringFormat is zero, then the date string will
 * be of the form (dd.mm.yy hh:mm:ss.msec TZ):
 *
 * "10.01.05 12:08:43.355 CET"
 *
 * \param ts is the TINE Data Timestamp for which the string conversion is to be
 * applied.
 * \param useLongStringFormat is a flag which if non-zero returns a longer representation
 * of the timestamp.
 * \param tsstr is a buffer to contain the time string (should be at least 64 characters).
 * this same buffer will be returned by the call as a matter of convenience.
 *
 * \note This call is thread-safe.
 *
 * \return A pointer to a string containing the timestamp representation.
 *
 * \see MakeDataTimeStamp
 * \see GetDataTimeStamp, GetCurrentDataTimeStamp, GetCurrentDataTimeStampFromCallbackId
 * \see ctime()
 * \see localtime()
 * \see findDaylight()
 */
TINE_EXPORT char *GetDataTimeStringEx(double ts, int useLongStringFormat, char *tsstr);

/* GetDataTimeStampAsTimeval() */
/**
 * \brief Returns the last established data timestamp as a pointer to a struct timeval.
 *
 * A call to getDataTimeStampAsTimeval() immediately following a call to ExecLink() or within a
 * callback routine will yield the data timestamp of the call.
 *
 * \note This call is not thread-safe.  See the discussion in getDataTimeStamp() for
 * details.
 *
 * \return A pointer to a struct timeval containing the data timestamp currently in place.
 *
 * \see MakeDataTimeStamp
 * \see GetDataTimeStamp, GetCurrentDataTimeStamp, GetCurrentDataTimeStampFromCallbackId
 */
TINE_EXPORT struct timeval *GetDataTimeStampAsTimeval(void);

/* GetTimeStampFromString() */
/**
 * \brief Returns a TINE timestamp converted from the string argument given
 *
 * A call to getDataTimeStampAsTimeval() immediately following a call to ExecLink() or within a
 * callback routine will yield the data timestamp of the call.
 *
 * \param ts is the string representation of a timestamp. 'NOW' returns the current system clock
 * as a TINE timestamp. A UTC value passed as a string (with or without milliseconds) is converted
 * to a double value and returned.  Otherwise a string of the form 'mm.dd.yy hh:mm:ss.msec' is
 * assumed.
 *
 * \return A TINE timestamp (0 if the argument passed is NULL or a zero-length string).
 *
 */
TINE_EXPORT double GetTimeStampFromString(char *ts);

/* getDataTimeStamp() */
/**
 * \brief Returns the last established data timestamp
 *
 * A call to getDataTimeStamp() immediately following a call to ExecLink() or within a
 * callback routine will yield the data timestamp of the call.
 *
 * \note This call is not thread-safe.  Synchronous data acquisition on two different threads
 * could conceivably interfer with the value returned from this routine.  A thread-safe way of
 * obtaining the data timestamp associated with a call is to use the .dTimestamp field of the
 * DTYPE object for synchronous calls such as ExecLink() or ExecLinkEx() or to use
 * getCurrentDataTimeStampFromCallbackId() or getCurrentDataTimeStamp() (which take a link
 * identifier as input parameter) for asynchronous links.
 *
 * \return The data timestamp currently in place.
 *
 * \see MakeDataTimeStamp
 * \see GetDataTimeStampAsTimeval, GetCurrentDataTimeStamp, GetCurrentDataTimeStampFromCallbackId
 * \see GetSystemDataStamp
 */
TINE_EXPORT double GetDataTimeStamp(void);
TINE_EXPORT double getDataTimeStamp(void);

/* getCurrentDataTimeStamp() */
/**
 * \brief Returns the data timestamp associated with the input link index
 *
 * \param i is the link index of the link for which the current data timestamp is desired.
 *
 * \return The data timestamp associated with the input link index
 *
 * \see GetDataTimeStampAsTimeval, GetDataTimeStamp, GetCurrentDataTimeStampFromCallbackId
 */
TINE_EXPORT double GetCurrentDataTimeStamp(int i);
#define getCurrentDataTimeStamp(a1) GetCurrentDataTimeStamp(a1)

/* getCurrentDataTimeStampFromCallbackId() */
/**
 * \brief Returns the data timestamp associated with the
 * callback identifier supplied
 *
 * \param i is the callback identifier of the link for which the current data timestamp is desired.
 *
 * \return The data timestamp associated with the input callback identifier
 *
 * \see GetDataTimeStampAsTimeval, GetDataTimeStamp, GetCurrentDataTimeStamp
 */
TINE_EXPORT double GetCurrentDataTimeStampFromCallbackId(int id);
#define getCurrentDataTimeStampFromCallbackId(a1) GetCurrentDataTimeStampFromCallbackId(a1)

/* setDataTimeStamp() */
/**
 * \brief Sets the intrinsic data timestamp to the value given
 *
 * This call is typically used inside an equipment function to set the
 * data timestamp associated with the request being made.  For instance,
 * when the data are taken from the hardware IO loop, a call to makeDataTimeStamp()
 * is made to get the current system timestamp. Then when a request for data is made
 * the caller can know the precise time of data acquisition.
 * If this routine is not called, then the current system time will be used to stamp the
 * data.
 *
 * \param ts is the timestamp to be associated with the next data set.
 *
 * \see PutDataTimeStamp, MakeDataTimeStamp
 */
TINE_EXPORT void SetDataTimeStamp(double ts);
#ifdef WIN32
TINE_EXPORT void setDataTimeStamp(double ts);
#else
#define setDataTimeStamp(a1) SetDataTimeStamp(a1)
#endif
/* PutDataTimeStamp() */
/**
 * \brief Returns a data timestamp according to the input parameters given
 *
 * This call is has the same purpose as makeDataTimeStamp(), but can assign a TINE
 * data timestamp according to the input seconds, milliseconds and system time offset.
 * This call is used only under special circumstances.
 *
 * \param toffset is the systematic time offset to be applied.
 * \param tsec is the data timestamp in seconds
 * \param tmsec is the millisecond part of the data timestamp
 *
 * \return A tine data timestamp.
 *
 * \see SetDataTimeStamp, MakeDataTimeStamp, GetDataTimeStamp
 */
TINE_EXPORT double PutDataTimeStamp(double toffset, time_t tsec, int tmsec);
#define putDataTimeStamp(a1,a2,a3) PutDataTimeStamp(a1,a2,a3)

/**
 * \brief Returns a data timestamp according to the input parameters given
 *
 * This call is has the same purpose as makeDataTimeStamp(), but can assign a TINE
 * data timestamp according to the input seconds, microseconds and system time offset.
 * This call is used only under special circumstances.
 *
 * \param toffset is the systematic time offset to be applied.
 * \param tsec is the data timestamp in seconds
 * \param tusec is the microsecond part of the data timestamp
 *
 * \return A tine data timestamp.
 *
 * \see SetDataTimeStamp, MakeDataTimeStamp, GetDataTimeStamp
 */
TINE_EXPORT double PutDataTimeStampU(double toffset, time_t tsec, int tusec);

/* MakeDataTimeStamp() */
/**
 * \brief Returns a data timestamp according to the current system time.
 *
 * \return A tine data timestamp.
 *
 * \see SetDataTimeStamp, PutDataTimeStamp, GetDataTimeStamp, MakeSystemDataStamp()
 */
TINE_EXPORT double MakeDataTimeStamp(void);
TINE_EXPORT double makeDataTimeStamp(void);

/* GetDataTimeStampOffset() */
/**
 * \brief Returns a the current data timestamp offset as assigned by
 * the TINE global time service
 *
 * \return the current data timestamp offset
 *
 * \see SetDataTimeStamp, PutDataTimeStamp, GetDataTimeStamp
 */
TINE_EXPORT double GetDataTimeStampOffset(void);

extern double gDataTimeStamp;
extern double gServerDataTimeStamp;
extern double gDataTimeStampOffset;
#define UTCtimestamp ((SINT32)gDataTimeStamp)
#define UTCtimestampMSEC (int)((gDataTimeStamp - (SINT32)gDataTimeStamp) * 1000)

#define tval2ts(s,us) ((double)s + ((double)us)/1000000.0)
/* RequireAcknowledgments */
/**
 * \brief Determines whether acknowledgements are expected following data changes where contracts use CM_DATACHANGE mode
 *
 * Under CM_TIMER link conditions the server is not directly interested as to whether the client received its data
 * packet or not, since the client is expecting data to arrive at the polling interval and will complain if it is
 * missing.  However,when a client establishes a data link and requests CM_DATACHANGE mode, data are are only returned to the client
 * when they have changed, or when the link heartbeat (1 minute) has expired.  A server will under these conditions
 * require an acknowledgement from the client if data do in fact change.  A missing acknowledgement will cause
 * the server to re-issue the data transfer.  This default behavior will of course increase the network traffic
 * when data are frequently changing.  Under some circumstance it is undesireable to require acknowledgments and
 * this feature can be turned off by setting this value to FALSE.
 *
 * \note Macro definition RequireAcknowledgments to provide compatibility with pre - refactoring usage
 *
 * \b Default: TRUE
 *
 */
extern TINE_EXPORT int gRequireAcknowledgments;
#define RequireAcknowledgments gRequireAcknowledgments

#define DEFAULT_REMOTE_MANAGMENT FALSE
extern TINE_EXPORT int gAllowRemoteExit;
extern TINE_EXPORT int gAllowRemoteReset;
extern TINE_EXPORT int gAllowRemoteInit;
extern int gDieOnAddrInUse;

/* SetDieOnAddressInUse */
/**
 * \brief Determines whether a server will exit() if an 'address in use' message
 * is delivered by the ENS upon registration.
 *
 * \param value is the desired setting (default = TRUE)
 *
 * \see GetDieOnAddressInUse()
 */
TINE_EXPORT void SetDieOnAddressInUse(int value);

/* GetDieOnAddressInUse */
/**
 * \brief Returns the current setting for this feature.
 *
 * If TRUE a server will exit() if an 'address in use' message
 * is delivered by the ENS upon registration.
 *
 * \return the current setting for this feature
 *
 * \see SetDieOnAddressInUse()
 */
TINE_EXPORT int GetDieOnAddressInUse(void);

extern int gDieOnSocketErrors;
/**
 * \brief Determines whether a server will exit() if there are
 * continuous socket errors on important server sockets.
 *
 * \param value is the desired setting (default = TRUE)
 *
 * \see GetDieOnSocketError()
 */
TINE_EXPORT void SetDieOnSocketError(int value);

/**
 * \brief Returns the current setting for this feature.
 *
 * If TRUE a server will exit() if there are
 * continuous socket errors on important server sockets.
 *
 * \return the current setting for this feature
 *
 * \see SetDieOnSocketError()
 */
TINE_EXPORT int GetDieOnSocketError(void);

extern int gDieOnFecIsAlias;

/* SetDieOnFecIsAlias */
/**
 * \brief Determines whether a server will exit() if a 'FEC has an alias' message
 * is delivered by the ENS upon registration.
 *
 * \param value is the desired setting (default = FALSE)
 *
 * \see GetDieOnFecIsAlias()
 */
TINE_EXPORT void SetDieOnFecIsAlias(int value);
/* GetDieOnFecIsAlias */
/**
 * \brief Returns the current setting for this feature.
 *
 * If TRUE a server will exit() if a 'FEC has an alias' message
 * is delivered by the ENS upon registration.
 *
 * \return the current setting for this feature
 *
 * \see SetDieOnFecIsAlias()
 */
TINE_EXPORT int GetDieOnFecIsAlias(void);

typedef void(*DIEFCNP)(const char *);
/**
* \brief Sets a dispatch routine to call if and when the FEC
* process dies due to e.g. 'address in use'
*
*/
TINE_EXPORT void SetDieFunction(DIEFCNP fcn);

extern int gDieAnotherDay;
/**
* \brief Revokes an 'exit' due to a 'die on' condition
*
* \param value is the desired setting.
*
* \note: the default (FALSE) is restablished following a call to SetDieAnotherDay
*/
TINE_EXPORT void SetDieAnotherDay(int value);

extern int gCatchConsoleBreak;
/**
 * \brief Determines whether an application will catch an interrupt
 * or break signal (such as control-C) and issue a graceful exit
 * or not.
 *
 * \param value is the desired setting (default = TRUE)
 *
 * \see GetCatchConsoleBreak()
 */
TINE_EXPORT void SetCatchConsoleBreak(int value);

/**
 * \brief Returns the current setting for this feature.
 *
 * If TRUE an application will catch an interrupt
 * or break signal (such as control-C) and issue a graceful exit.
 *
 * \return the current setting for this feature
 *
 * \see SetCatchConsoleBreak()
 */
TINE_EXPORT int GetCatchConsoleBreak(void);

/* SetAllowRemoteManagement */
/**
 * \fn TINE_EXPORT void SetAllowRemoteManagement(int value)
 * \brief Determines whether remote management via stock properties is possible
 *
 * The stock properties "SRVEXIT", "SRVINIT", "SRVRESET" will return 'not_allowed'
 * unless remote managment is turned on (default OFF)
 *
 * \param value is the desired setting
 *
 * \see GetAllowedRemoteManagement()
 */
TINE_EXPORT void SetAllowRemoteManagement(int value);

/* GetAllowedRemoteManagement */
/**
 * \fn TINE_EXPORT int GetAllowedRemoteManagement(void)
 * \brief Returns whether remote management via stock properties is possible
 *
 * The stock properties "SRVEXIT", "SRVINIT", "SRVRESET" will return 'not_allowed'
 * unless remote managment is turned on (default OFF)
 *
 * \return the desired setting
 *
 * \see SetAllowRemoteManagement()
 */
TINE_EXPORT int GetAllowedRemoteManagement(void);

/* SetScanForUsersFiles */
/**
 * \fn TINE_EXPORT int SetScanForUsersFiles(const char *eqm)
 * \brief Instructs the initialization process to look for device and property
 * specific 'users' files.
 *
 * Normally this routine is not necessary.  However for some systems such as
 * VxWorks, where directory scanning only works in conjunction with NFS mounts
 * it is then useful to explicitly intruct the initialization process to look
 * for device and property specific ACL files (such as 'device1-users.csv', etc.).
 *
 * \return 0 upon success
 *
 * \see SetScanForNetsFiles()
 */
TINE_EXPORT int SetScanForUsersFiles(const char *eqm);

/* SetScanForNetsFiles */
/**
 * \fn TINE_EXPORT int SetScanForNetsFiles(const char *eqm)
 * \brief Instructs the initialization process to look for device and property
 * specific 'ipnets' files.
 *
 * Normally this routine is not necessary.  However for some systems such as
 * VxWorks, where directory scanning only works in conjunction with NFS mounts
 * it is then useful to explicitly intruct the initialization process to look
 * for device and property specific ACL files (such as 'device1-ipnets.csv', etc.).
 *
 * \return 0 upon success
 *
 * \see SetScanForUsersFiles()
 */
TINE_EXPORT int SetScanForNetsFiles(const char *eqm);

#define LOG_SCAN_NONE 0
#define LOG_SCAN_FEC  1
#define LOG_SCAN_FULL 2

extern int gLogfileAllowScan;
extern char *logFileScanText[3];

/* SetLogFileAllowScan */
/**
 * \fn TINE_EXPORT void SetLogFileAllowScan(int value)
 * \brief Sets the allowed range of log file retrieval from a FEC process
 * (via stock property SRVLOGFILE).
 *
 * A running FEC process will attempt to read and return the contents of
 * the file name input to stock property SRVLOGFILE.  The FEC process must
 * have read permission in order to succeed. The setting of
 * the LogFileAllowScan will further restrict the allowed directories from
 * which files may be read.  LogFileAllowScan = LOG_SCAN_FEC will only return
 * the contents of files within the FEC_HOME or FEC_LOG directories and their
 * subdirectories.  LogFileAllowScan = LOG_SCAN_FULL will itself allow the
 * entire file system.  LogFileAllowScan = LOG_SCAN_NONE will not allow any
 * files to be retrieved other than fec.log.
 *
 * \param value is the desired scan range, one of: LOG_SCAN_NONE,
 * LOG_SCAN_FEC, LOG_SCAN_FULL (default = LOG_SCAN_FEC).
 *
 * \see GetLogFileAllowScan()
 */
TINE_EXPORT void SetLogFileAllowScan(int value);

/* GetLogFileAllowScan */
/**
 * \fn TINE_EXPORT void SetLogFileAllowScan(int value)
 * \brief Gets the allowed range of log file retrieval from a FEC process
 * (via stock property SRVLOGFILE).
 *
 * A running FEC process will attempt to read and return the contents of
 * the file name input to stock property SRVLOGFILE.  The FEC process must
 * have read permission in order to succeed. The setting of
 * the LogFileAllowScan will further restrict the allowed directories from
 * which files may be read.  LogFileAllowScan = LOG_SCAN_FEC will only return
 * the contents of files within the FEC_HOME or FEC_LOG directories and their
 * subdirectories.  LogFileAllowScan = LOG_SCAN_FULL will itself allow the
 * entire file system.  LogFileAllowScan = LOG_SCAN_NONE will not allow any
 * files to be retrieved other than fec.log.
 *
 * \return the current scan range setting, one of: LOG_SCAN_NONE,
 * LOG_SCAN_FEC, LOG_SCAN_FULL (default = LOG_SCAN_FEC).
 *
 * \see SetLogFileAllowScan()
 */
TINE_EXPORT int GetLogFileAllowScan(void);

extern short SychronizeFecClock;
extern WORD watchdogBaseAddress;
extern UINT32 gSystemCycleDeadband;
extern TINE_EXPORT unsigned short MinPollingRate;
extern TINE_EXPORT unsigned short MaxPollingRate;

/* SetSystemCycleDeadband */
/**
 * \fn TINE_EXPORT void SetSystemCycleDeadband(int value)
 * \brief Sets the system cycle deadband
 *
 * A TINE client or server requires a 'cycler' to run either on separate thread
 * (see SystemStartCycleTimer() where available) or as part of a user defined engine
 * (running SystemCycle() in an infinite loop).
 * The cycler will block on incoming network traffic for a duration up to the
 * given system cycle deadband (typically 10 msec). Either a network request wakes
 * the cycler up or the deadband timeout expires.  Following this other background
 * activity will take place (such as processing contracts, checking local histories,
 * and alarms, etc.).
 * The system cycle deadband can be 'pinned' with this API call.
 *
 * \param value is the desired system cycle deadband (default: OS dependent, typically 10 msec)
 *
 * \see GetSystemCycleDeadband()
 */
TINE_EXPORT void SetSystemCycleDeadband(int value);

/* GetSystemCycleDeadband */
/**
 * \fn TINE_EXPORT int GetSystemCycleDeadband(void)
 * \brief Gets the system cycle deadband
 *
 * A TINE client or server requires a 'cycler' to run either on separate thread
 * (see SystemStartCycleTimer() where available) or as part of a user defined engine
 * (running SystemCycle() in an infinite loop).
 * The cycler will block on incoming network traffic for a duration up to the
 * given system cycle deadband (typically 10 msec). Either a network request wakes
 * the cycler up or the deadband timeout expires.  Following this other background
 * activity will take place (such as processing contracts, checking local histories,
 * and alarms, etc.).
 * The system cycle deadband can be read with this API call.
 *
 * \return the current system cycle deadband (default: OS dependent, typically 10 msec)
 *
 * \see SetSystemCycleDeadband()
 */
TINE_EXPORT int GetSystemCycleDeadband(void);

/* SetMinimumAllowedPollingInterval */
/**
 *\fn TINE_EXPORT void SetMinimumAllowedPollingInterval(int value)
 * \brief Sets the minimum polling interval in milliseconds a server will allow.
 *
 * A server will not honor polling rates smaller than this value.  An attempt to poll a property a 1 msec
 * for instance will be serviced at the MinPollingRate instead.
 * If it is known that the server should support very fast polling rates,
 * then this value should be set accordingly at initialization time.
 *
 * \param value is the desired minimum polling interval (Default: 20 )
 *
 * \see GetMinimumAllowedPollingInterval()
 */
TINE_EXPORT void SetMinimumAllowedPollingInterval(int value);

/* GetMinimumAllowedPollingInterval */
/**
 *\fn TINE_EXPORT int GetMinimumAllowedPollingInterval(void)
 * \brief returns the minimum polling rate in milliseconds a server will allow.
 *
 * A server will not honor polling rates smaller than this value.  An attempt to poll a property a 1 msec
 * for instance will be serviced at the MinPollingRate instead.
 * If it is known that the server should support very fast polling rates,
 * then this value should be set accordingly at initialization time.
 *
 * \return the current setting for this value
 *
 * \see SetMinimumAllowedPollingInterval
 */
TINE_EXPORT int GetMinimumAllowedPollingInterval(void);

/* SetSystemSynchronizeContracts */
/**
 * \brief Establishes the setting for general contract synchronization at the server
 *
 * A server will try to schedule contracts at the regular intervals
 * necessary to satisfy its clients' needs.
 * If 'contract synchronization' is in effect, the server will attempt
 * to call the relevant equipment module dispatch routines in a tight loop
 * in order to collect and disperse information in the smallest number of
 * data transfers (contract packing).
 * You can control this behavior with this API call.
 *
 * \param value is the desired setting for contract synchronization
 * (default = TRUE);
 *
 * \see GetSystemSynchronizeContracts()
 */
TINE_EXPORT void SetSystemSynchronizeContracts(int value);

/* GetSystemSynchronizeContracts */
/**
 * \brief Returns the setting for general contract synchronization at the server
 *
 * A server will try to schedule contracts at the regular intervals
 * necessary to satisfy its clients' needs.
 * If 'contract synchronization' is in effect, the server will attempt
 * to call the relevant equipment module dispatch routines in a tight loop
 * in order to collect and disperse information in the smallest number of
 * data transfers (contract packing).
 * You can ascertain the current setting with this API call.
 *
 * \return the current setting for contract synchronization
 * (default = TRUE);
 *
 * \see SetSystemSynchronizeContracts()
 */
TINE_EXPORT int GetSystemSynchronizeContracts(void);

extern int defaultClientTimeout;
#define defaultClientTimeoutMsec defaultClientTimeout
/* MaxNumContracts */
/**
 * \brief Determines the maximum number of contracts a server will service.
 *
 * A server manages both a contract table and a client table, which are independent of one another
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number
 * contracts, then this value should be set accordingly at initialization time.
 *
 * \b Default: 1000 (Or #define CONTRACTLIST_CAPACITY in project.def)
 *
 * \see SetContractListCapacity(), GetContractListCapacity()
 */
extern TINE_EXPORT int MaxNumContracts;
extern int ContractListCapacity;
/* SetContractListCapacity */
/**
 * \brief Sets the maximum number of contracts a server will service.
 *
 * A server manages both a contract table and a client table, which are independent of one another
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number
 * contracts, then this value should be set accordingly at initialization time.
 *
 * \param value is the requested contract table size
 *
 * \b Default: 1000 (Or #define CONTRACTLIST_CAPACITY in project.def)
 *
 * \see GetContractListCapacity()
 */
TINE_EXPORT void SetContractListCapacity(int value);
/* GetContractListCapacity */
/**
 * \brief Gets the maximum number of contracts a server will service.
 *
 * A server manages both a contract table and a client table, which are independent of one another
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number
 * contracts, then this value should be set accordingly at initialization time. The current setting can be obtained
 * with this call
 *
 * \return the current contract table size
 *
 * \see SetContractListCapacity()
 */
TINE_EXPORT int GetContractListCapacity(void);
/* MaxNumClients */
/**
 * \brief Determines the maximum number of clients a server will service.
 *
 * A server manages both a contract table and a client table, which are independent of one another
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number
 * clients, then this value should be set accordingly at initialization time.
 *
 * \b Default: 50 (Or #define CLIENTLIST_CAPACITY in project.def)
 *
 * \see SetClientListCapacity(), GetClientListCapacity()
 */
extern TINE_EXPORT int MaxNumClients;
extern int ClientListCapacity;
/* GetClientListCapacity */
/**
 * \brief Gets the maximum number of clients a server will service.
 *
 * A server manages both a contract table and a client table, which are independent of one another
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number
 * clients, then this value should be set accordingly at initialization time. The current setting can be obtained
 * with this call
 *
 * \return the current client table size
 *
 * \see SetClientListCapacity()
 */
TINE_EXPORT int GetClientListCapacity(void);
/* SetClientListCapacity */
/**
 * \brief Sets the maximum number of clients a server will service.
 *
 * A server manages both a contract table and a client table, which are independent of one another
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number
 * clients, then this value should be set accordingly at initialization time.
 *
 * \param value is the requested client table size
 *
 * \b Default: 100 (Or #define CLIENTLIST_CAPACITY in project.def)
 *
 * \see GetClientListCapacity()
 */
TINE_EXPORT void SetClientListCapacity(int value);
/* gConTblCapacity */
/**
 * \brief Determines the maximum number of entries in the connection table.
 *
 * A client's connections are managed and maintained in a connection table.  The size of this table
 * is pre-allocated at initialization time.  This allows for fast lookups, since a connection ID is
 * simply an entry into the table.  If it is known that the client will need a large number of simultaneous
 * links then this value should be set accordingly at initialization time.
 *
 * \b Default: 1000 (Or #define CONTBL_CAPACITY in project.def)
 *
 * \see SetConnectionTableCapacity(), GetConnectionTableCapacity()
 */
extern TINE_EXPORT int gConTblCapacity;
#define MaxNumConnections gConTblCapacity
/* SetConnectionTableCapacity */
/**
 * \brief Sets the maximum number of entries in the connection table.
 *
 * A client's connections are managed and maintained in a connection table.  The size of this table
 * is pre-allocated at initialization time.  This allows for fast lookups, since a connection ID is
 * simply an entry into the table.  If it is known that the client will need a large number of simultaneous
 * links then this value should be set accordingly at initialization time.
 *
 * \param value is the requested client connection table size
 *
 * \b Default: 1000 (Or #define CLIENTLIST_CAPACITY in project.def)
 *
 * \see GetConnectionTableCapacity()
 */
TINE_EXPORT int SetConnectionTableCapacity(int value);
/* GetConnectionTableCapacity */
/**
 * \brief Gets the maximum number of entries allowed in the connection table.
 *
 * A client's connections are managed and maintained in a connection table.  The size of this table
 * is pre-allocated at initialization time.  This allows for fast lookups, since a connection ID is
 * simply an entry into the table.  If it is known that the client will need a large number of simultaneous
 * links then this value should be set accordingly at initialization time.
 *
 * \see SetConnectionTableCapacity()
 */
TINE_EXPORT int GetConnectionTableCapacity(void);

/* FlushContractTable */
/**
 * \brief Flushes the current contract and client entry tables
 *
 * Calling this routine will effectively disconnect all attached clients and remove all
 * persistent contracts.  This will appear to any connected client to be a server 'time-out' and
 * attached clients will attempt to re-establish any lost connections.  Any link callbacks a client
 * might have will receive 'link_timeout' status messages during the reconnection interval.
 * Thus this routine can effectively be used to signal a 'server restart' during server runtime.
 */
TINE_EXPORT void FlushContractTable(void);

/* MaxSystemTransportSize */
/**
 * \brief Determines the maximum size in bytes of any call supported by the server
 *
 * Any call which attempts to access more than the MaxSystemTransportSize will fail with
 * 'invalid transport size'.  If it is known before hand that the server should support calls for
 * very large amounts of data, and there is available main memory on the server, this value should be
 * set accordingly prior to the call to SystemInit().
 *
 * \b Default: 64 Kb for most operating systems.  Exceptions: Win16 (32 Kbyte).
 *
 */
extern TINE_EXPORT UINT32 MaxSystemTransportSize;
#define MaxRPCTransportSize MaxSystemTransportSize
extern UINT16 gPortOffset;
/* FeclogDepth */
/**
 * \brief Determines the approximate size of the server's log file in lines.
 *
 * The server's log file, if located on the local disk, is a file called 'fec.log'.  When it's length
 * surpasses the FecLogDepth, this file is moved to 'fec.bak' and a new 'fec.log' file is started.  A
 * call for the server's log file will always be able to scan both the 'fec.bak' file and 'fec.log', and
 * will thus always be able to scan at least FeclogDepth entries.
 *
 * \b Default: 100
 *
 */
extern TINE_EXPORT int FeclogDepth;
extern int leftFeclogTime;
/* gStartupDebug */
/**
 * \brief Determines whether server initialization messages are displayed or not
 *
 * A running server is quiet regarding information displayed on the standard output. An
 * initializing server is however by default not quiet, and will echo various steps of the
 * initialization process on the screen, including whether of not optional configuration files
 * are found or not, etc.  This information can be surpressed by setting gStartupDebug to FALSE
 * prior to calling SystemInit().
 *
 * \note Critical errors will be displayed on the screen in all circumstances.
 *
 * \note Macro definition StartupDebug to provide compatibility with pre-refactoring usage
 *
 * \b Default: TRUE
 *
 */
extern TINE_EXPORT int gStartupDebug;
#define StartupDebug gStartupDebug
extern int queryDeviceNamesFromProperties;

/* LastCompletionDataSize */
/**
 * \brief Supplies the data size of the most recent link
 *
 * A global variable which supplies the data size of the most recent link.
 *
 * \note This is not thread safe.  In multithreaded applications, use
 * GetCompletionDataSize() or GetCompletionDataSizeFromCallbackId()
 *
 * \see GetCompletionDataSize(), GetCompletionDataSizeFromCallbackId()
 *
 */
extern UINT32 LastCompletionDataSize;

/* LastCompletionDataType */
/**
 * \brief Supplies the data type of the most recent link
 *
 * A global variable which supplies the data type of the most recent link.
 *
 * \note This is not thread safe.  In multithreaded applications, use
 * GetCompletionDataType() or GetCompletionDataTypeFromCallbackId()
 *
 * \see GetCompletionDataType(), GetCompletionDataTypeFromCallbackId()
 *
 */
extern int LastCompletionDataType;
/* putCommandsInFeclog */
/**
 * \brief Determines whether all in-coming WRITE access calls are automatically included in the server's log file
 *
 * \b Default: TRUE
 *
 */
extern TINE_EXPORT int putCommandsInFeclog;

/* SetPutCommandsInFeclog */
/**
 * \brief Determines whether all in-coming WRITE access calls are automatically included in the server's log file
 *
 * \param value is the desired setting (default = TRUE).
 *
 * \note This is an exported routine which simply sets the global variable putCommandsInFeclog
 *
 * \see GetPutCommandsInFeclog
 */
TINE_EXPORT void SetPutCommandsInFeclog(int value);

/* GetPutCommandsInFeclog */
/**
 * \brief Returns the current setting for putCommandsInFeclog
 *  (which determines whether all in-coming WRITE access calls are automatically included in the server's log file)
 *
 * \return the current setting for putCommandsInFeclog
 *
 * \see SetPutCommandsInFeclog
 */
TINE_EXPORT int GetPutCommandsInFeclog(void);

/* nofeclog */
/**
 * \brief Determines whether a server is to keep a log file on the local disk
 *
 * The location of the log file is
 * determined by the FECDB environment variable or, if not set, the current working directory is used.
 * If set to TRUE, then no log file will be maintained on the local disk.  A log file will nonetheless be
 * maintained in a local ring-buffer, with the caveat that all entries will disappear upon server restart.
 * To eliminate the log file entirely, 'nofeclog' should be set to TRUE and 'FeclogDepth' should be set to 0.
 *
 * \b Default: FALSE for most operating systems.  Exceptions: VxWorks, NIOS.
 *
 */
extern TINE_EXPORT int nofeclog;
extern int gVerboseCommands;

TINE_EXPORT void SetPutCommandsInCmdlog(int value);
TINE_EXPORT int GetPutCommandsInCmdlog(void);
TINE_EXPORT void SetVerboseCommandsInCmdLog(int value);
TINE_EXPORT int GetVerboseCommandsInCmdLog(void);

/* IsStandAlone() */
/**
 * \brief Determines whether a client or server process is running in stand-alone mode.
 *
 * Stand-alone mode is generally set by setting the environment variable TINE_STANDALONE to TRUE prior
 * starting the client or server process.  If TRUE, no attempt is made to contact the TINE
 * equipment name server.  Instead the local address CACHE is consulted.  A server updates its
 * own entry by supplying the registered port offset and the loopback address.
 *
 * \b Default: FALSE.
 *
 */
TINE_EXPORT int IsStandAlone(void);

/* SetStandAlone() */
/**
 * \brief Sets stand-alone mode.
 *
 * In lieu of setting the environment variable TINE_STANDALONE to TRUE prior
 * starting a server process, this API call can set stand-alone mode if used prior to SystemInit().
 * If TRUE, no attempt is made to contact the TINE
 * equipment name server.  Instead the local address CACHE is consulted.  A server updates its
 * own entry by supplying the registered port offset and the loopback address.
 *
 * \b Default: FALSE.
 *
 * \return 0 upon success or 'already_assigned' if called after SystemInit().
 *
 */
TINE_EXPORT int SetStandAlone(int value);

/* SetSystemLogging */
/**
 * \brief Sets system logging (output to fec.log) to TRUE or FALSE
 *
 * \param value should be TRUE (non zero) if logging is desired (DEFAULT)
 * or FALSE if not.
 *
 */
TINE_EXPORT void SetSystemLogging(int value);
/* GetSystemLogging */
/**
 * \brief Returns the current system logging setting (TRUE or FALSE)
 *
 * \see SetSystemLogging().
 *
 */
TINE_EXPORT int GetSystemLogging(void);

/* useGlobalSynchronization */
/**
 * \brief Determines whether data timestamps are to be externally synchronized
 *
 * If set to TRUE, data timestamps will be synchronized to an externally supplied timestamp.  This
 * requires a TINE Time Server to be in operation and sending timestamps at regular intervals either as
 * multicast or as broadcast.
 *
 * \b Default: TRUE
 *
 */
extern TINE_EXPORT int useGlobalSynchronization;

/**
 * \brief Determines whether data timestamps are to be externally synchronized
 *
 * \param value If set to TRUE, data timestamps will be synchronized to an externally supplied timestamp.  This
 * requires a TINE Time Server to be in operation and sending timestamps at regular intervals either as
 * multicast or as broadcast (default TRUE);
 *
 * \see GetUseGlobalSynchronization()
 */
TINE_EXPORT void SetUseGlobalSynchronization(int value);

/**
 * \brief Returns whether data timestamps are to be externally synchronized
 *
 * \return TRUE or FALSE
 *
 * \see SetUseGlobalSynchronization()
 */
TINE_EXPORT int GetUseGlobalSynchronization(void);
extern time_t appmktime;

/* SetAppDate */
/**
 * \brief Sets the compilation date of the current running server process
 *
 * The server's application creation date is queriable from a TINE server.
 * However this information needs to be passed to the server engine at initialization
 * time via this call.
 *
 * \param appdate is the creation time as a UNIX time variable.
 *
 * \see SetAppVersion()
 */
TINE_EXPORT void SetAppDate(time_t appdate);

extern char appmkversion[64];
/* SetAppVersion */
/**
 * \brief Sets the application version of the current running server process
 *
 * The server's application version number is queriable from a TINE server.
 * However this information needs to be passed to the server engine at initialization
 * time via this call.
 *
 * \param major is the major version number of the application
 * \param minor is the minor version number of the application
 * \param revision is the revision version number of the application
 *
 * \see SetAppVersion()
 */
TINE_EXPORT void SetAppVersion(int major,int minor,int revision);

/**
 * \brief Enables/Disables automatic link watchdogs.
 *
 * On-Data-Change or Event monitor links can sometimes fail to notice a missing server
 * for up to the system heartbeat of 60 seconds.  With this flag enabled (the default)
 * all such links to a given server are coupled to a watchdog link which will fire the
 * appropriate link notifications when a server goes down. If this feature is not
 * desireable, it can be turned off with this call.
 *
 * \param value is either TRUE to turn this feature on (the default) or FALSE to turn it
 * off
 *
 */
TINE_EXPORT void SetAutoLinkWatchdogs(int value);

/**
 * \brief Sets the default timeout in milliseconds to be used in simple calls to ExecLink().
 *
 * A call to ExecLink() does not take a timeout parameter and hence will use the current
 * system default timeout value.  A call to ExecLinkEx() does take a timeout parameter
 * which overrides the system default timeout value.
 *
 * \param timeoutMilliseconds is the timeout value to be applied to simple synchronous
 * links using ExecLink().
 *
 * \see ExecLink(), ExecLinkEx().
 */
TINE_EXPORT void SetClientTimeout(int timeoutMilliseconds);
#define setClientTimeout(a1) SetClientTimeout(a1)

extern double gSyncTolerance;

/* setSynchronizationTolerance */
/**
 * \brief Sets the tolerance used in deciding whether to apply a timestamp offset or not.
 *
 * If a TINE time server is running and supplying a system time stamp, and this server
 * is accepting this to synchronize its local clock when applying data time stamps,
 * the tolerance used in applying this offset can be set by this routine.
 *
 * \param toleranceInSeconds is the maximum time difference allowed before setting the
 * data timestamp offset. (Default: 0.5 seconds)
 *
 *
 * \see getSynchronizationTolerance()
 */
TINE_EXPORT void SetSynchronizationTolerance(double toleranceInSeconds);

/* getSynchronizationTolerance */
/**
 * \brief Gets the tolerance used in deciding whether to apply a timestamp offset or not.
 *
 * If a TINE time server is running and supplying a system time stamp, and this server
 * is accepting this to synchronize its local clock when applying data time stamps,
 * the tolerance used in applying this offset can be read by this routine.
 *
 * \return toleranceInSeconds is the maximum time difference allowed before setting the
 * data timestamp offset.
 *
 * \see setSynchronizationTolerance()
 */
TINE_EXPORT double GetSynchronizationTolerance(void);

TINE_EXPORT int hasDevicePropertyList(char *eqm,char *device);
TINE_EXPORT int hasPropertyDeviceList(char *eqm,char *property);

/**
 * \brief Fills in the given DTYPE data object with all data stamp information.
 *
 * If the link ID is known, a reference to a DTYPE object will be filled with the
 * current time and data stamps as well as the transfer reason. The remaining
 * elements of the given DTYPE object are not touched.
 *
 * \param linkId (input) is the link ID (see AttachLink()) of the data link.
 * \param dout (output) is a reference to the DTYPE data object to be filled in.
 *
 * \return 0 or a TINE error code.
 *
 * \see UpdateDataStampsFromCallbackId()
 */
TINE_EXPORT int UpdateDataStampsFromLinkId(int linkId,DTYPE *dout);

/**
 * \brief Fills in the given DTYPE data object with all data stamp information.
 *
 * If the user-supplied callback identifier is unambiguous, a reference to a
 * DTYPE object will be filled with the
 * current time and data stamps as well as the transfer reason. The remaining
 * elements of the given DTYPE object are not touched.
 *
 * \param id (input) is the callback id supplied with the data link.
 * \param dout (output) is a reference to the DTYPE data object to be filled in.
 *
 * \return 0 or a TINE error code.
 *
 * \see UpdateDataStampsFromCallbackId()
 */
TINE_EXPORT int UpdateDataStampsFromCallbackId(int id,DTYPE *dout);

/* GetDataFromLinkId */
/**
 * \brief Supplies the DTYPE data object and call status for the link ID in question.
 *
 * If the link ID is known, the output data information pertaining the the link can be
 * obtained via this call. A reference to a DTYPE object will be filled with the
 * current data parameters, including all time and data stamps, transfer reason, data
 * format, etc.
 *
 * \param linkId (input) is the link ID (see AttachLink()) of the data link.
 * \param dout (output) is a reference to the DTYPE data object to be filled in.
 * \param status (output) is a reference to an unsigned short integer to receive the
 * call's return code.
 *
 * \return 0 or a TINE error code.
 *
 * \see GetDataFromCallbackId()
 */
TINE_EXPORT int GetDataFromLinkId(int linkId, DTYPE *dout,UINT16 *status);

/* GetDataFromCallbackId */
/**
 * \brief Supplies the DTYPE data object and call status for the callback ID in question.
 *
 * If the user-supplied callback identifier is unambiguous, the output data information
 * pertaining the the link can be obtained via this call. A reference to a DTYPE object
 * will be filled with the current data parameters, including all time and data stamps,
 * transfer reason, data format, etc.
 *
 * \param id (input) is the callback id of the link supplied with the data link.
 * \param dout (output) is a reference to the DTYPE data object to be filled in.
 * \param status (output) is a reference to an unsigned short integer to receive the
 * call's return code.
 *
 * \return 0 or a TINE error code.
 *
 * \see GetDataFromLinkId()
 */
TINE_EXPORT int GetDataFromCallbackId(int id,DTYPE *dout,UINT16 *status);

/* GetLinkCallbackReference */
/**
 * \brief Returns the supplied callback reference for the link ID in question.
 *
 * If the user-supplied callback reference is given in the doubly extended AttachLink call
 * to AttachLinkEx2(), this reference pointer can be retrieved by using this routine.
 *
 * \param id (input) is the callback id of the link for which the current completion code is desired.
 *
 * \return NULL or the reference pointer supplied in the call to AttachLinkEx2().
 *
 * \see AttachLinkEx2()
 */
TINE_EXPORT void *GetLinkCallbackReference(int id);

/* SetSystemDataStamp() */
/**
 * \brief Sets the global system data stamp with which to tag outbound data sets
 *
 * A server can set a global data stamp value (e.g. cycle id or run number, etc.)
 * which will accompany all data transactions.
 *
 * \param value is the integer value to use as the system data stamp.
 *
 * \return void
 *
 * \see SetDataTimeStamp
 */
TINE_EXPORT void SetSystemDataStamp(int value);

/* MakeSystemDataStamp() */
/**
 * \brief Returns the current valid global system data stamp
 *
 * A server can set a global data stamp value (e.g. cycle id or run number, etc.)
 * which will accompany all data transactions.
 *
 * \param value is the integer value to use as the system data stamp.
 *
 * \return void
 *
 * \see MakeDataTimeStamp()
 */
TINE_EXPORT int MakeSystemDataStamp(void);

/* GetSystemDataStamp() */
/**
 * \brief Gets the system data stamp with which the incoming data set has been tagged
 *
 * A server can set a global data stamp value (e.g. cycle id or run number, etc.)
 * which will accompany all data transactions. The caller can retrieve this value with this call.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).  If -1
 * is passed, then the currently active value for the data stamp is returned.
 *
 * \return The data stamp if the call succeeds or -1.
 *
 */
TINE_EXPORT int GetSystemDataStamp(int linkId);

/* GetSystemDataStampFromCallbackId() */
/**
 * \brief Gets the system data stamp with which the incoming data set has been tagged
 *
 * A server can set a global data stamp value (e.g. cycle id or run number, etc.)
 * which will accompany all data transactions. The caller can retrieve this value with this call.
 *
 * \param id is the callback id (supplied with a call to AttachLinkEx()).  If -1
 * is passed, then the currently active value for the data stamp is returned.
 *
 * \return The data stamp if the call succeeds or -1.
 *
 */
TINE_EXPORT int GetSystemDataStampFromCallbackId(int id);

/* GetGlobalsSystemStamp() */
/**
 * \brief Gets the system data stamp with which the incoming globals data set has been tagged
 *
 * A server can set a global data stamp value (e.g. cycle id or run number, etc.)
 * which will accompany all data transactions. The caller can retrieve this value with this call.
 *
 * \param glbId is the globals table link id (returned from a call to recvNetGlobal()).  If -1
 * is passed, then the currently active value for the data stamp is returned.
 *
 * \return The data stamp if the call succeeds or -1.
 *
 */
TINE_EXPORT int GetGlobalsSystemStamp(int glbId);

/* GetDataStamp() */
/**
 * \brief Gets the data stamp with which the incoming data set has been tagged
 *
 * A server can set an integer data stamp with any transaction
 * which will accompany all data transactions. The caller can retrieve this value with this call.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).  If -1
 * is passed, then the currently active value for the data stamp is returned.
 *
 * \return The data stamp if the call succeeds or -1.
 *
 */
TINE_EXPORT int GetDataStamp(int linkId);

/* GetDataStampFromCallbackId() */
/**
 * \brief Gets the global system data stamp with which the incoming data set has been tagged
 *
 * A server can set an integer data stamp with any transaction
 * which will accompany all data transactions. The caller can retrieve this value with this call.
 *
 * \param id is the callback id (supplied with a call to AttachLinkEx()).  If -1
 * is passed, then the currently active value for the data stamp is returned.
 *
 * \return The data stamp if the call succeeds or -1.
 *
 */
TINE_EXPORT int GetDataStampFromCallbackId(int id);

/* GetGlobalsDataStamp() */
/**
 * \brief Gets the data stamp with which the incoming globals data set has been tagged
 *
 * A server can set an integer data stamp with any transaction
 * which will accompany all data transactions. The caller can retrieve this value with this call.
 *
 * \param glbId is the connection table link id (returned from a call to recvNetGlobal()).  If -1
 * is passed, then the 0 is returned.
 *
 * \return The data stamp if the call succeeds or -1.
 *
 */
TINE_EXPORT int GetGlobalsDataStamp(int glbId);

/* getLastContractTableId() */
/**
 * \brief Returns the last contract table identifier.
 *
 * A server can retieve the last contract table identifier at any time.  This
 * routine is most usefull however when used within a server's equipment module.
 *
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
TINE_EXPORT int GetLastContractTableId(void);

/* getLastContractTableId() */
/**
 * \brief Returns the access rate (interval) associated with the given contract
 *
 * A server can retieve the access rate of the specified contract.  If a negative
 * number is passed, then the most recently invoked contract is assumed.
 * This routine is most usefull however when used within a server's equipment module.
 *
 * \param id is the contract table indentifier whose access rate is desired
 *
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
TINE_EXPORT int GetContractAccessRate(int id);
#define getContractAccessRate(a1) GetContractAccessRate(a1)

/* GetTransferFlag() */
/**
 * \brief Gets the data transfer flag for the given link ID
 *
 * The caller can obtain an examine the data transfer reason associated with
 * the last data transfer for the link ID given.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).  If -1
 * is passed, then the current value of the transfer flag is returned.
 *
 * \return The data transfer flag.
 *
 */
TINE_EXPORT UINT16 GetTransferFlag(int linkId);

/* GetTransferFlagFromCallbackId() */
/**
 * \brief Gets the data transfer flag for the give callback identifier
 *
 * The caller can obtain an examine the data transfer reason associated with
 * the last data transfer for the callback identifier given.
 *
 * \param id is the callback id (supplied with a call to AttachLinkEx()).  If -1
 * is passed, then the current value of the transfer flag is returned.
 *
 * \return The data transfer flag.
 *
 */
TINE_EXPORT UINT16 GetTransferFlagFromCallbackId(int id);

/* ReassignLinkData() */
/**
 * \brief Reassigns the bound data buffer for the link table id given.
 *
 * A monitored link is often 'bound' to a specified buffer which is to
 * maintain a representation of the corresponding data (i.e. if the
 * output DTYPE object passed in the call references a non-null data
 * pointer). The caller can reassign this buffer by using this call.
 *
 * \param id is the link table id (e.g. returned from the original call to AttachLink()).
 * \param buffer is a pointer to the new buffer;
 * \param length is the length in bytes of the new buffer.  This is only used as
 *  a 'control'.  The call is only successful if the length passed is at least as
 *  big as the number of bytes required in the original link.
 *
 * \note If the monitored link uses an 'assigned' buffer (i.e. a NULL pointer
 * was originally specified in the call) then the call will fail with the error code
 * 'already_assigned',
 *
 * \note If the monitored link makes use of exotic data types such as CF_IMAGE or CF_MDA
 * then the caller must ensure the integrity of the buffer passed with respect to the
 * original buffer.
 *
 * \return 0 upon success or a TINE error code.
 *
 */
TINE_EXPORT int ReassignLinkData(int id,void *buffer,UINT32 length);

/* GetClnRecvQueueDepth() */
/**
 * \brief Gets the default client-side receive queue depth for all client links.
 *
 * Asynchronous data acquisition implies a callback handler which accepts and processes incoming data.
 * Under most circumstances the callback handler should process the incoming data quickly and return
 * prior to the next data acquisition (otherwise the data rate is set too high to be useful for the client).
 * However circumstances arrive where either bursts of data arrive or the client side is otherwise busy
 * processing an atypical event, etc.  In these cases queueing the incoming data can be beneficial.
 * You can get the default queue depth for initializing links with this routine.
 *
 * \return the current setting for the default receive queue depth or a negative TINE return code upon error.
 *
 */
TINE_EXPORT int GetClnRecvQueueDepth(void);

/* GetLinkQueueDepth() */
/**
 * \brief Gets the client-side receive queue depth for the link in question
 *
 * Asynchronous data acquisition implies a callback handler which accepts and processes incoming data.
 * Under most circumstances the callback handler should process the incoming data quickly and return
 * prior to the next data acquisition (otherwise the data rate is set too high to be useful for the client).
 * However circumstances arrive where either bursts of data arrive or the client side is otherwise busy
 * processing an atypical event, etc.  In these cases queueing the incoming data can be beneficial.
 * You can get the current queue depth for a specific link using this routine.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).
 *
 * \return the current queue depth setting for the link in question.
 */
TINE_EXPORT int GetLinkQueueDepth(int linkId);

/* SetClnRecvQueueDepth() */
/**
 * \brief Sets the default client-side receive queue depth for all client links.
 *
 * Asynchronous data acquisition implies a callback handler which accepts and processes incoming data.
 * Under most circumstances the callback handler should process the incoming data quickly and return
 * prior to the next data acquisition (otherwise the data rate is set too high to be useful for the client).
 * However circumstances arrive where either bursts of data arrive or the client side is otherwise busy
 * processing an atypical event, etc.  In these cases queueing the incoming data can be beneficial.
 * You can set the default queue depth for initializing links with this routine.
 *
 * \param depth is the requested queue depth (values less than zero are ignored). A value of zero
 * implies no queue will be maintained.
 *
 */
TINE_EXPORT void SetClnRecvQueueDepth(int depth);

/* SetLinkQueueDepth() */
/**
 * \brief Sets the client-side receive queue depth for the link in question
 *
 * Asynchronous data acquisition implies a callback handler which accepts and processes incoming data.
 * Under most circumstances the callback handler should process the incoming data quickly and return
 * prior to the next data acquisition (otherwise the data rate is set too high to be useful for the client).
 * However circumstances arrive where either bursts of data arrive or the client side is otherwise busy
 * processing an atypical event, etc.  In these cases queueing the incoming data can be beneficial.
 * You can set the queue depth for a specific link using this routine.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).
 * \param depth is the requested queue depth (values less than zero are ignored). A value of zero
 * implies no queue will be maintained.
 *
 */
TINE_EXPORT void SetLinkQueueDepth(int linkId,int depth);

/**
 * \brief Sets a client-side callback notification
 *
 * Asynchronous data acquisition implies a callback handler which accepts and processes incoming data.
 * In some exotic circumstances it might be desireable to delay a callback notification for some
 * specified number of TINE kernel cycles.  For instance a 'master link' should needs to fire its
 * callback prior to examining data in other link callbacks.
 * One can use this call to accomplish this.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).
 * \param cyclesToDelay is the number of cycles to delay prior to firing the link callback. A value of zero
 * (DEFAULT) implies no delay.
 *
 * \see GetLinkCallbackDelay
 */
TINE_EXPORT void SetLinkCallbackDelay(int linkId,int cyclesToDelay);

/**
 * \brief Gets the client-side callback notification
 *
 * Asynchronous data acquisition implies a callback handler which accepts and processes incoming data.
 * In some exotic circumstances it might be desireable to delay a callback notification for some
 * specified number of TINE kernel cycles.  For instance a 'master link' should needs to fire its
 * callback prior to examining data in other link callbacks.
 * One can use this call to examine the current setting.
 *
 * \param linkId is the connection table link id (returned from a call to AttachLink()).
 *
 * \return the current delay setting (DEFAULT = 0).
 *
 * \see SetLinkCallbackDelay
 */
TINE_EXPORT int GetLinkCallbackDelay(int linkId);

/* GetServerWaiting() */
/**
 * \brief Gets the server's waiting state
 *
 * A Server can be set in a wait state prior to initialization by passing a non-zero value as
 * argument.  In this state, the server will respond to all contract requests with the
 * status code 'not_initialized'.
 * This is useful, if the server needs to undergo a long initialization precedure before
 * allowing clients access.  The call retrieves the current wait state (0 = not waiting,
 * non-zero = waiting).
 * The server's wait state is, of course, FALSE (not waiting) by
 * default.
 *
 * \return the value of the wait state.
 *
 * \see SetServerWaiting()
 */
TINE_EXPORT int GetServerWaiting(void);

/* SetServerWaiting() */
/**
 * \brief Sets the server's waiting state to the value given
 *
 * A Server can be set in a wait state prior to initialization by passing a non-zero value as
 * argument.  In this state, the server will respond to all contract requests with the
 * status code 'not_initialized'.
 * This is useful, if the server needs to undergo a long initialization precedure before
 * allowing clients access.  Setting the wait state to TRUE before stating the server's cycle()
 * and then later to FALSE when all properties and equipment modules are ready is a safe way
 * to prevent premature access.  The server's wait state is, of course, FALSE (not waiting) by
 * default.
 *
 * \param value is the value of the wait state.
 *
 * \note This call does not target a specific equipment module and hence applies to all registered
 * equipment modules on the FEC.
 *
 * \see GetServerWaiting()
 */
TINE_EXPORT void SetServerWaiting(int value);

/* GetServerIdleState() */
/**
 * \brief Gets the server's idle state
 *
 * A Server can be set in an idle state where no background activity occurs and the server's
 * equipment function is temporarily disabled.  This call returns the current idle state of the
 * equipment module specified.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \return the value of the idle state.
 *
 * \see SetServerIdleState()
 */
TINE_EXPORT int GetServerIdleState(char *eqm);

/* SetServerIdleState() */
/**
 * \brief Sets the server's idle state to the value given
 *
 * A Server can be set in an idle state where no background activity occurs and the server's
 * equipment function is temporarily disabled.  Calls to the equipment function will receive the
 * status code 'server_idle' if this is the case.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param value is the value of the idle state. Any non-zero sets the idle state to TRUE. A zero
 * value sets the idle state to FALSE.
 *
 * \note This call targets a specific equipment module and hence only applies to the
 * specified server module.
 *
 * \see GetServerIdleState()
 */
TINE_EXPORT void SetServerIdleState(char *eqm,int value);

extern int gEqmInitializedIdleState;

/* SetInitializeIdle() */
/**
 * \brief When set to TRUE, all equipment modules are initialized in an 'idle' state.
 *
 * If a Server is in an idle state no background activity occurs and the server's
 * equipment function is temporarily disabled.  Calls to the equipment function will receive the
 * status code 'server_idle' if this is the case.  This is sometime useful during initialization,
 * so that no activity (from for example the local history subsystem) can occur.  If this flag
 * is set to TRUE, then all equipment modules must be activated following the initialization
 * procedure by calling SetServerIdleState(FALSE);
 *
 * \param value is the value of the default idle state. Any non-zero sets the default idle state to TRUE. A zero
 * value sets the default idle state to FALSE.
 *
 *
 * \see SetServerIdleState()
 */
TINE_EXPORT void SetInitializeIdle(int value);

/**
 * \brief Determines whether TINE configuration files will be scanned or not.
 *
 * By default the TINE kernel will look for configuration files covering
 * all ranges of server initialization as well as client initialization
 * (such as the cshost.csv file giving a list of configured Equipment Name
 * Servers).  Some server builds might make use entirely and intentionally
 * of API calls to establish all free configuration parameters.  And in
 * such cases it is possibly not desireable to even scan for the existence
 * of such files (e.g. VxWorks CPUs which want to avoid querying the
 * host parent's file system).  By passing a value of TRUE this
 * default scanning will be turned off.
 *
 * \param value is the desired setting (default: FALSE);
 *
 * \note: If set to TRUE then at least one ENS address should be set via
 * a call to SetNameServerAddress() or FindNameServerOnNetwork().
 *
 * \note: The default value can also be changed by building the TINE library
 * with -DTINE_CONFIGURATION_CODED=TRUE or by editing project.def to
 * include the line #define TINE_CONFIGURATION_CODED TRUE
 *
 * \see GetConfigurationCoded()
 */
TINE_EXPORT void SetConfigurationCoded(int value);

/**
 * \brief Returns whether TINE configuration files will be scanned.
 *
 *
 * \return the current setting
 *
 * \see SetConfigurationCoded()
 */
TINE_EXPORT int GetConfigurationCoded(void);

#define TINE_CONFIGURATION_CODED FALSE
extern int gTineConfigurationCoded;

/* SetAlwaysRetry() */
/**
 * \brief Sets the 'Always Retry' flag to the value given.
 *
 * Client Links can be instructed to always apply the CA_RETRY flag upon initialization
 * by setting this flag to TRUE.  If TRUE all link timeouts will be retried up to three times
 * before issuing a link timeout notification. The default value is TRUE.
 *
 * \param value is the value of the retry flag.
 *
 * \see GetAlwaysRetry()
 */
TINE_EXPORT void SetAlwaysRetry(int value);

/* GetAlwaysRetry() */
/**
 * \brief Gets the current setting of the 'Always Retry' flag.
 *
 * Client Links can be instructed to always apply the CA_RETRY flag upon initialization
 * by setting this flag to TRUE.  If TRUE all link timeouts will be retried up to three times
 * before issuing a link timeout notification. The default value is TRUE.
 *
 * \return the current value of the retry flag.
 *
 * \see SetAlwaysRetry()
 */
TINE_EXPORT int GetAlwaysRetry(void);

/* GetSystemErrorString() */
/**
 * \brief Gets the system error code in plain text
 *
 * This routine only returns the corresponding string for TINE system error codes.  User
 * defined error codes should use the routine GetLastLinkError().
 *
 * \param cc is the completion code or status code for which the internal system error string is desired.
 * \param errstr is a string buffer capable of handling up to 32 characters.
 *
 * \return a pointer to the string buffer 'errstr' given (unless NULL).
 *
 * \see GetLastLinkError()
 */
TINE_EXPORT char *GetSystemErrorString(short cc,char *errstr);

/**
 * \brief Sets the link watchdog polling interval to the value given
 *
 * Data-change or event style monitor links will receive data updates only when the
 * examined data have changed or an event has been declared.
 * In addition a systematic heartbeat is received
 * at approximately once per minute per monitor link and a systematic watch dog
 * link per target server is established (if enabled) at the minimum of the desired remote
 * polling interval and the value given by this call (default = 1000 msec).
 * Thus a link timeout notification will be sent to the link callback with
 * minimum latency.  The default link watchog polling interval can be set with this function
 * call.
 *
 * \param value is the desired link watchdog polling interval in milliseconds.
 *
 */
TINE_EXPORT int SetLinkWatchdogPollingInterval(int value);

/**
 * \brief Sets the require acknowledgments flag to the value given
 *
 * A server normally requires an acknowledgment from clients being sent changed
 * data if the client link is a data-change style monitor.  Under some circumstances
 * this extra acknowledgment network traffic might be unnecessary or undesireable and
 * can be disabled with the function call.
 *
 * \param value is the desired setting (default = TRUE).
 *
 * \see GetSystemRequireAcknowledgments()
 */
TINE_EXPORT void SetSystemRequireAcknowledgments(int value);

/**
 * \brief Gets the require acknowledgments flag to the value given
 *
 * A server normally requires an acknowledgment from clients being sent changed
 * data if the client link is a data-change style monitor.  Under some circumstances
 * this extra acknowledgment network traffic might be unnecessary or undesireable.
 * The current setting can be acquired with this function call.
 *
 * \return the current setting for the require acknowledgments flag.
 *
 * \see SetSystemRequireAcknowledgments()
 */
TINE_EXPORT int GetSystemRequireAcknowledgments(void);

/* GetContractId() */
/**
 * \brief Returns the last contract identifier for the equipment module specified.
 *
 * A server can retieve the current contract identifier when the equipment module dispatch
 * routine is being called. If this routine is called outside of the dispatch routine,
 * the value returned will only represent the contract ID valid when the equipment module
 * was last called.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 *
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if there is an error (parameter is invalid).
 *
 */
TINE_EXPORT int GetContractId(const char *eqm);

/* GetLastContractId() */
/**
 * \brief Returns the last contract identifier.
 *
 * A server can retieve the last contract identifier at any time.
 *
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
TINE_EXPORT int GetLastContractId(void);

/* AddFieldToBitField() */
/**
 * \brief Adds a field definition to a registered bitfield
 *
 * \param srv is the orginating server declaring the bitfield.  A server
 * declaring a bit field should pass a NULL or otherwise a simple tag such as "this".
 * \param tag is the bitfield tag
 * \param mask is the mask to be applied to the bitfield value to obtain the field element
 * \param field is the field name assigned to the masked value.
 *
 * \return a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \see OpenBitField(), GetBitfieldAsString(), GetFieldFromBitfield()
 *
 * \b Example:
 *
 * \include eg_addFieldToBitField.c
 */
TINE_EXPORT int AddFieldToBitField(char *srv,char *tag,UINT32 mask,char *field);
#define addFieldToBitField(a1,a2,a3,a4) AddFieldToBitField(a1,a2,a3,a4)

/* OpenBitField() */
/**
 * \brief Declares a bit field type and registers with the bitfield registry
 *
 * \param srv is the orginating server declaring the bitfield.  A server
 * declaring a bit field should pass a NULL or otherwise a simple tag such as "this".
 * \param tag is the bitfield tag
 * \param fmt is the bit field format (one of CF_BITFIELD8, CF_BITFIELD16, or CF_BITFIELD32).
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to
 * GetLastLinkError().
 *
 * \see AddFieldToBitField(), GetBitfieldAsString(), GetFieldFromBitfield()
 */
TINE_EXPORT int OpenBitField(char *srv,char *tag,int fmt);
#define openBitField(a1,a2,a3) OpenBitField(a1,a2,a3)

/* GetFieldFromBitfield() */
/**
 * \brief Retrieves the requested field value from a bit field
 *
 * \param srv is the orginating server declaring the bitfield.  A server
 * declaring a bit field should pass a NULL or otherwise a simple tag such as "this".
 * \param tag is the bitfield tag
 * \param field is the field for which the values is desired
 * \param value is the full value of the bitfield (cast as a 32 bit integer)
 *
 * \return the bitfield value from the supplied value or a negative tine return code.
 *
 * \see OpenBitField(), AddFieldToBitField(), GetBitfieldAsString()
 */
TINE_EXPORT int GetFieldFromBitfield(char *srv,char *tag,char *field,UINT32 value);
#define getFieldFromBitfield(a1,a2,a3,a4) GetFieldFromBitfield(a1,a2,a3,a4)

/* GetBitfieldAsString() */
/**
 * \brief Retrieves the requested field value from a bit field as a string value
 *
 * \param srv (input) is the orginating server declaring the bitfield.  A server
 * declaring a bit field should pass a NULL or otherwise a simple tag such as "this".
 * \param tag (input) is the bitfield tag
 * \param field (input) is the field for which the values is desired
 * \param value (input) is the full value of the bitfield (cast as a 32 bit integer)
 * \param strbuf (output) is the string buffer which should contain the converted string value
 * \param buflen (input) is the length of the supplied string buffer
 *
 * \return a tine return code.
 *
 * \see OpenBitField(), AddFieldToBitField(), GetFieldFromBitfield()
 */
TINE_EXPORT int GetBitfieldAsString(char *srv,char *tag,UINT32 value,char *strbuf,int buflen);
#define getBitfieldAsString(a1,a2,a3,a4,a5) GetBitfieldAsString(a1,a2,a3,a4,a5)

/* AcquireAndRegisterBitfield() */
/**
 * \brief Acquires the bitfield specified by the bitfield tag from the specified device server.
 *
 * This routine is largly useful for generic applications which need to display results.  Logic
 * which requires knowing 'which field does what' will in general also require knowing a priori
 * the bitfield fields.
 *
 * \param devName is the targeted device server where the bitfield is registered.
 * \param tag is the bitfield tag
 *
 * \return a tine return code.
 */
TINE_EXPORT int AcquireAndRegisterBitfield(const char *devName, char *tag);
int acquireAndRegisterBitfield(const char *devname, char *tag);
TINE_EXPORT short GetBitfieldFormat(short fmt);
TINE_EXPORT short GetIntegerFormat(short fmt);

/* AcquireAndRegisterStruct() */
/**
 * \brief Acquires the structure specified by the structure tag from the specified device server.
 *
 * This routine is largly useful for generic applications which need to display results.  Logic
 * which requires knowing 'which field does what' will in general also require knowing a priori
 * the structure fields.
 *
 * \param devName is the targeted device server where the structure is registered.
 * \param tag is the structure tag
 * \param num is the maximum number of structure objects of this type which can be stord locally.
 * Note that due to alignment, byte-swapping and memory issues, the TINE kernel needs to reserve
 * enough space to prepare incoming structures for the caller.
 *
 * \return a tine return code.
 *
 * \see SealTaggedStruct(), AddFieldToStruct()
 */
TINE_EXPORT int AcquireAndRegisterStruct(char *devName, char *tag,int num);
int acquireAndRegisterStruct(char *devName, char *tag,int num);

/* SetAccessLock() */
/**
 * \brief Acquires an access lock to the server specified.
 *
 * A client application can obtain an access lock to the given server provided the caller
 * has WRITE privileges himself.  Once the lock is obtained, no other process will be
 * allowed WRITE access regardless of any other access control lists.  A lock may be
 * preemptive (LOCK_PREEMTIVE) or persistent (LOCK_PERSISTENT).  Preemptive locks may be
 * aborted (LOCK_ABORT) by other callers with WRITE privileges. Persistent locks may not.
 *
 * \param context is the targeted context of the server
 * \param server is the targeted device server
 * \param lockType is the desired lock type: one of LOCK_CANCEL, LOCK_PREEPMTIVE, LOCK_PERSISTENT or
 * LOCK_ABORT.
 * \param lockDuration is the requested duration of the lock in seconds.  This applies only to
 * preemptive locks.  Persistent locks will ignore this parameter and continue to renew the lock
 * until the caller cancels the lock (graceful) or disappears (maximum 60 second wait).
 *
 * \return a tine return code.
 *
 * Example:
 *
 * \include eg_SetAccessLock.c
 *
 * \see FreeAccessLock(), GetAccessLockInformation(), GetAccessLockStatus()
 */
TINE_EXPORT int SetAccessLock(char *context, char *server, int lockType,int lockDuration);

/* FreeAccessLock() */
/**
 * \brief Frees an access lock on the server specified.
 *
 * A client application can obtain an access lock to the given server provided the caller
 * has WRITE privilileges himself.  Once the lock is obtained, no other process will be
 * allowed WRITE access regardless of any other access control lists.  A lock may be
 * preemptive (LOCK_PREEMTIVE) or persistent (LOCK_PERSISTENT).  Preemptive locks may be
 * aborted (LOCK_ABORT) by other callers with WRITE privilileges. Persistent locks may not.
 * A client can free a lock he has obtained by calling this method (nominally equivalent to
 * SetAccessLock() + LOCK_CANCEL.
 *
 * \param context is the targeted context of the server
 * \param server is the targeted device server
 *
 * \return a tine return code.
 *
 * Example:
 *
 * \include eg_SetAccessLock.c
 *
 * \see SetAccessLock(), GetAccessLockInformation(), GetAccessLockStatus()
 */
TINE_EXPORT int FreeAccessLock(char *context, char *server);

/* GetAccessLockStatus() */
/**
 * \brief Acquires the current access lock status on a client's access lock.
 *
 * A client application can obtain an access lock to the given server provided the caller
 * has WRITE privilileges himself.  Once the lock is obtained, no other process will be
 * allowed WRITE access regardless of any other access control lists.  A lock may be
 * preemptive (LOCK_PREEMTIVE) or persistent (LOCK_PERSISTENT).  Preemptive locks may be
 * aborted (LOCK_ABORT) by other callers with WRITE privilileges. Persistent locks may not.
 * A client can of course hold different access locks to different servers, and can obtain
 * the current lock status for any of them by using this function.
 *
 * \param context is the targeted context of the server
 * \param server is the targeted device server
 *
 * \return a tine return code indicating the status of the access lock.
 *
 * \see SetAccessLock(), FreeAccessLock(), GetAccessLockInformation()
 */
TINE_EXPORT int GetAccessLockStatus(char *context, char *server);

/* GetAccessLockInformation() */
/**
 * \brief Acquires access lock information from the server specified.
 *
 * A client application can obtain an access lock to the given server provided the caller
 * has WRITE privilileges himself.  Once the lock is obtained, no other process will be
 * allowed WRITE access regardless of any other access control lists.  A lock may be
 * preemptive (LOCK_PREEMTIVE) or persistent (LOCK_PERSISTENT).  Preemptive locks may be
 * aborted (LOCK_ABORT) by other callers with WRITE privilileges. Persistent locks may not.
 * A client can obtain the current access lock information via this call. This will
 * provide information as to the caller's user name, network address, and time remaining.
 * If there is no access lock on the server specified, then the caller's name will be empty,
 * the network address will be "0.0.0.0" and the time remaining will be "0 seconds remaining".
 *
 * \param context is the targeted context of the server
 * \param server is the targeted device server
 * \param callerName is a NAME32 object to hold the user name of the holder of the access lock.
 * \param callerIp is a NAME32 object to hold the network address of the holder of the access lock.
 * \param timeLeft is a NAME32 object to hold a string given the seconds remaining on the access lock.
 *
 * \return a tine return code.
 *
 * Example:
 *
 * \include eg_GetAccessLock.c
 *
 * \see SetAccessLock(), FreeAccessLock(), GetAccessLockStatus()
 */
TINE_EXPORT int GetAccessLockInformation(char *context,char *server,NAME32 *callerName,NAME32 *callerIp,NAME32 *timeLeft);

/* SetFailoverPollingInterval() */
/**
 * \brief Sets the server failover interval to the value given
 *
 * If a server has declared itself as a failover slave then it monitors a designated master
 * at the failover polling interval.  The value of the polling interval can be set via this
 * routine.
 *
 * \param pollingInterval is the interval in milliseconds to check for connectivity with the
 * targeted master (default: 1000 msec).
 *
 * \return a tine return code.
 *
 * \note The 'polling' is an asynchronous link to the stock property "SRVSTARTTIME".
 *
 * \see SetFailoverThreshold()
 */
TINE_EXPORT void SetFailoverPollingInterval(int pollingInterval);

/* SetFailoverThreshold() */
/**
 * \brief Sets the server failover threshold to the value given
 *
 * If a server has declared itself as a failover slave then it monitors a designated master
 * at the failover polling interval.  If consecutive errors acrue such that the failover threshold
 * is exceeded, then a failover slave will assume the roll of master.
 *
 * \param errorCounts is the number of consecutive errors which triggers a role reversal of slave to
 * master (default: 5).
 *
 * \return a tine return code.
 *
 * \see SetFailoverPollingInterval()
 */
TINE_EXPORT void SetFailoverThreshold(int timeoutCounts);

/* SetFailoverMaster() */
/**
 * \brief Sets the designated server as a failover master
 *
 * A server can declared itself as a failover master, which will register a 'master' server name
 * as an additional alias for the equipment module specified. A server master will always overwrite
 * any prior entries for the master name within the ENS database.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param masterName is the desired name of the server master.  Ideally this is a distinct name
 * different from the offically registered export name.  A failover slave will use this same
 * master name.
 *
 * \return a tine return code.
 *
 * \note It is frequently better to make use of the "MASTER" column within the fecid.csv or the
 * "MASTER" tag within the fec.xml database files to establish a failover master.  Most often the
 * identical server binary is simply run on different hardware, which means that hardcoding an API
 * call such as this will require examining either the command line or local environment in order to
 * deterimine whether the booting server is to be a master or slave.
 *
 * \see SetFailoverSlave()
 */
TINE_EXPORT int SetFailoverMaster(char *eqm,char *masterName);

/* SetFailoverSlave() */
/**
 * \brief Declares the server a failover slave and targets the designated server.
 *
 * A server can declared itself as a failover slave, which will monitor the designated
 * 'master' and only assume the role master if the monitored server no longer responds.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param masterName is the desired name of the server master.
 * \param slaveMaster is the targeted name of the slave monitor.  This will be the officially
 * registered name of the server designated as the master (hopefully distinct from the 'master' name).

 *
 * \return a tine return code.
 *
 * \note It is frequently better to make use of the "MASTER" and "SLAVE_MASTER" columns within
 * the fecid.csv or the "MASTER" and "SLAVE_MASTER" tags within the fec.xml database files
 * to establish a failover master.  Most often the
 * identical server binary is simply run on different hardware, which means that hardcoding an API
 * call such as this will require examining either the command line or local environment in order to
 * deterimine whether the booting server is to be a master or slave.
 * For instance, the master server exports itself officially as "BPM1" and declares a master name of "BPM".
 * The slave server exports itself as "BPM2" and declares a master name of "BPM" and a slave master of "BPM1".
 *
 * \see SetFailoverMaster()
 */
TINE_EXPORT int SetFailoverSlave(char *eqm,char *masterName,char *slaveMaster);

/* GetCurrentFailoverState() */
/**
 * \brief Returns current master slave state of server
 *
 * If configured for master slave operation, the current state of a server is either master or slave. As this is
 * defined for EQMs it could be possible to misconfigure a server such that one EQM operates as master and another
 * as a slave. Until we can think of a meaningful usecase for such a configuration, we return an
 * indicator for an inconsistent configuration in such a case.
 * We use the Definitions: FAILOVER_NONE (0), FAILOVER_MASTER (1), FAILOVER_SLAVE (2),
 *   FAILOVER_MASTER | FAILOVER_SLAVE (3), ~FAILOVER_NONE (-1)
 *
 * \return 0 (no master-slave config), 1 (master), 2 (slave), 3 (slave-as-master), -1 (inconsistent config)
 *
 * \see SetFailoverMaster()
 */
TINE_EXPORT int GetCurrentFailoverState( void);

extern int gRejectOverloadedMetaProperties;
/* SetRejectOverloadedMetaProperties */
/**
 * \brief Enables/Disables overloaded meta properties
 *
 * Meta-properties such as .HIST or .NAM can be 'overloaded' by registering the appropriately
 * named property.  This will prevent the systematically available meta-property from being called.
 * This is by default the case (if a property ABC.HIST is itself registered it will be called instead
 * of the TINE local history).  Explicitly calling SetRejectOverloadedMetaProperties(TRUE) will
 * disable this feature.
 *
 * \param value if non-zero will reject any meta-property overloads. (default: FALSE)
 *
 * \see GetRejectOverloadedMetaProperties()
 */
TINE_EXPORT void SetRejectOverloadedMetaProperties(int value);

/* GetRejectOverloadedMetaProperties */
/**
 * \brief returns the current setting
 *
 * Meta-properties such as .HIST or .NAM can be 'overloaded' by registering the appropriately
 * named property.  This will prevent the systematically available meta-property from being called.
 * This is by default the case (if a property ABC.HIST is itself registered it will be called instead
 * of the TINE local history).  Explicitly calling SetRejectOverloadedMetaProperties(TRUE) will
 * disable this feature.
 *
 * \return the current setting (default: FALSE -> overloaded meta properties are allowed and will not be rejected)
 *
 * \see SetRejectOverloadedMetaProperties()
 */
TINE_EXPORT int GetRejectOverloadedMetaProperties(void);

/* AppendRegisteredUsers */
/**
 * \brief appends the current users access list with the name list given
 *
 * The users access list assigned to an equipment module is often read in via configuration file
 * (see users.csv or fec.xml).  However users can also be assigned via this API call.  The list
 * given will be appended to any existing list of allowed users.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param userlist is a list of users to be appended to the current users access list.  If a
 * user in the input is already in the current access list, nothing happens.
 * \param listsize is the size of the input list.
 *
 * \return 0 or a TINE error code.
 *
 * \see RemoveRegisteredUser()
 */
TINE_EXPORT int AppendRegisteredUsers(const char *eqm,NAME16 *userlist,int listsize);
TINE_EXPORT int AppendRegisteredUserList(NAME16 *userlist,int listsize);
#define appendRegisteredUserList(a1,a2) AppendRegisteredUserList(a1,a2)

/* RemoveRegisteredUser */
/**
 * \brief removes those users in the name list given from the current users access list.
 *
 * The users access list assigned to an equipment module is often read in via configuration file
 * (see users.csv or fec.xml).  However users can also be removed via this API call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param userlist is a list of users to be removed from the current users access list.  If a
 * user in the input is not in the current access list, nothing happens.
 * \param listsize is the size of the input list.
 *
 * \return 0 or a TINE error code.
 *
 * \see AppendRegisteredUsers()
 */
TINE_EXPORT int RemoveRegisteredUser(char *eqm,NAME16 *userlist,int listsize);

/* GetRegisteredUsers */
/**
 * \brief retrieves the currently assigned users with WRITE permissions on the equipment module in question.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param userlist [output] is NAME16 array buffer to receive the list of users with WRITE privileges.
 * \param *listsize [input,output] is a pointer reference to the initial size of the 'userlist' buffer.
 * When the call completes *listsize will contain the actual number of users in the equipment module's
 * user list.
 *
 * \return 0 upon success or a TINE error code.
 *
 * \see AppendRegisteredUsers()
 */
TINE_EXPORT int GetRegisteredUsers(const char *eqm,NAME16 *userlist,int *listsize);

extern int gNumFecAdmins;
extern NAME16 *gFecAdmins;

/* AssertIsAdministrator */
/**
 * \brief returns TRUE if the given user is a registered administrator
 *
 * An addional class of uses (applicable to the entire FEC process) can be registered by
 * supplying an 'admins.csv' file in analogy with a users.csv file.
 * Users in this category are merely classified as 'administrators' and as such have no
 * systematic meaning within the TINE library.  However an equipment module displatch
 * routine (the equipment function) can make use of this routine in order to limit
 * specific WRITE operations to administrators only.  Typically one checks the caller's username
 * and upon a return of 'FALSE' from AssertIsAdministrator() then one returns 'unauthorized_action'
 * from the equipment function.
 *
 * \param usr is the users name whose administrative credentials are being verified.
 *
 * \return TRUE if the given user is an administrator OR if there is NO administors list
 * (in which case everyone is classified as an administrator).
 *
 * \see GetCaller(), GetCallerInformation()
 */
TINE_EXPORT int AssertIsAdministrator(const char *usr);

/* AppendRegisteredNetsList */
/**
 * \brief appends the current network access list with the name list given
 *
 * The network access list assigned to FEC is often read in via configuration file
 * (see ipnets.csv or fec.xml).  However networks can also be assigned via this API call.  The list
 * given will be appended to any existing list of allowed network addresses.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param iplist is a list of network addresses to be appended to the current network access list.  If a
 * network address in the input is already in the current access list, nothing happens.
 * \param listsize is the size of the input list.
 *
 * \return 0 or a TINE error code.
 *
 * \see RemoveRegisteredNetsList()
 */
TINE_EXPORT int AppendRegisteredNetsList(const char *eqm,NAME64 *iplist,int listsize);
#define appendRegisteredNetsList(a1,a2,a3) AppendRegisteredNetsList(a1,a2,a3)

/**
 * \deprecated
 */
TINE_EXPORT int AppendRegisteredIPNetsList(NAME16 *iplist,int listsize);
#define appendRegisteredIPNetsList(a1,a2) AppendRegisteredIPNetsList(a1,a2)

/* AppendRegisteredBCastNetsList */
/**
 * \brief appends the current network broadcast list with the name list given
 *
 * The network broadcast list assigned to FEC is often read in via configuration file
 * (see ipbcast.csv or fec.xml).  However networks can also be assigned via this API call.  The list
 * given will be appended to any existing list of allowed network addresses.  Generally use of this
 * API call or configuration files is unncessary, as most modern operating systems will be able to
 * receive TINE multicasts.  Some legacy systems will not be able to do this however.  In such
 * cases, applying a broadcast list will re-send all multicasted transmission as UDP broadcasts to
 * those networks specified in the list.
 *
 * \param iplist is a list of network addresses to be appended to the current broadcast list.  If a
 * network address in the input is already in the current access list, nothing happens.
 * \param listsize is the size of the input list.
 *
 * \return 0 or a TINE error code.
 *
 * \see RemoveRegisteredBCastNetsList()
 */
TINE_EXPORT int AppendRegisteredBCastNetsList(NAME64 *iplist,int listsize);
#define appendRegisteredBCastNetsList(a1,a2) AppendRegisteredBCastNetsList(a1,a2)

/* RemoveRegisteredNetsList */
/**
 * \brief removes those networks in the name list given from the current network address access list.
 *
 * The network access list assigned to FEC is often read in via configuration file
 * (see ipnets.csv or fec.xml).  However networks can also be removed via this API call.
 * Items in the list given will be removed from the existing list of allowed network addresses.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param iplist is a list of network addresses to be removed from the current network access list.  If a
 * network address in the input is not in the current access list, nothing happens.
 * \param listsize is the size of the input list.
 *
 * \return 0 or a TINE error code.
 *
 * \see AppendRegisteredNetsList()
 */
TINE_EXPORT int RemoveRegisteredNetsList(const char *eqm,NAME64 *iplist,int listsize);
/**
 * \deprecated
 */
TINE_EXPORT int RemoveRegisteredIPNetsList(NAME16 *iplist,int listsize);

/* RemoveRegisteredBCastNetsList */
/**
 * \brief removes those networks in the name list given from the current broadcast list.
 *
 * The network broadcast list assigned to FEC is often read in via configuration file
 * (see ipbcast.csv or fec.xml).  However networks can also be removed via this API call.
 * Items in the list given will be removed from the existing list of broadcasted network addresses.
 *
 * \param iplist is a list of network addresses to be removed from the current broadcast list.  If a
 * network address in the input is not in the current access list, nothing happens.
 * \param listsize is the size of the input list.
 *
 * \return 0 or a TINE error code.
 *
 * \see AppendRegisteredBCastNetsList()
 */
TINE_EXPORT int RemoveRegisteredBCastNetsList(NAME64 *iplist,int listsize);

/**
 * \brief retrieves network list of ip addresses and subnets for which WRITE access is allowed
 *
 * \param list (output) is a buffer to receive a list of relevant network addresses and subnets
 * \param listsize (input/output) is the size of the input list.
 *
 * \return 0 or a TINE error code.
 */
TINE_EXPORT int GetRegisteredNetworks(const char *eqm, NAME64 *list, int *listsize);

/* SetRetardSingleContractRemoval */
/**
 * \brief sets the retard contract removal state
 *
 * When all clients disappear from a contract the contract is then scheduled for
 * removal at the server.  This can either happen 'immediately' or be 'retarded' for
 * a duration of 1 second.  Typically a synchronous READ or WRITE command establishes
 * a transient contract which if allowed to persist for this extra second can handle
 * 'RETRY' requests by returning the results of the contract.  However this strategy
 * also occupies a slot in the servers contract table until the contract is officially
 * removed, when can effectively lock up the contract table if a large number of synchronous
 * requests are issued (clients will receive a 'resources_exhausted' return code in such
 * cases.  If a server can tolerate idempotent requests (the same request twice, so to
 * speak) then it might be advisable to turn this feature off.
 *
 * \param value is the desired setting. (default = TRUE).
 *
 *
 * \see GetRetardSingleContractRemoval()
 */
TINE_EXPORT void SetRetardSingleContractRemoval(int value);

/* GetRetardSingleContractRemoval */
/**
 * \brief returns the current setting of this feature.
 *
 * When all clients disappear from a contract the contract is then scheduled for
 * removal at the server.  This can either happen 'immediately' or be 'retarded' for
 * a duration of 1 second.  Typically a synchronous READ or WRITE command establishes
 * a transient contract which if allowed to persist for this extra second can handle
 * 'RETRY' requests by returning the results of the contract.  However this strategy
 * also occupies a slot in the servers contract table until the contract is officially
 * removed, when can effectively lock up the contract table if a large number of synchronous
 * requests are issued (clients will receive a 'resources_exhausted' return code in such
 * cases.  If a server can tolerate idempotent requests (the same request twice, so to
 * speak) then it might be advisable to turn this feature off.
 *
 * \return the current setting of this feature.
 *
 * \see SetRetardSingleContractRemoval()
 */
TINE_EXPORT int GetRetardSingleContractRemoval(void);

extern int ignoreCommonDbFiles;

/* SetIgnoreCommonFiles */
/**
 * \brief turns searching for common database files in the FEC_HOME directory on or off
 *
 * When set to TRUE, attempts to read an exports.csv or related equipment module files
 * which directly reside in the FEC_HOME directory (or working directory, if not set) are
 * blocked.  Any related files within the equipment module 'local name' sub-directory
 * will still be scanned and read.
 *
 * \param value is the desired setting. (default = FALSE).
 *
 * \see GetIgnoreCommonFiles()
 */
TINE_EXPORT void SetIgnoreCommonFiles(int value);

/* GetIgnoreCommonFiles */
/**
 * \brief returns the current setting (default = FALSE)
 *
 * When set to TRUE, attempts to read an exports.csv or related equipment module files
 * which directly reside in the FEC_HOME directory (or working directory, if not set) are
 * blocked.  Any related files within the equipment module 'local name' sub-directory
 * will still be scanned and read.
 *
 * \return the current setting of this feature.
 *
 * \see SetIgnoreCommonFiles()
 */
TINE_EXPORT int GetIgnoreCommonFiles(void);

/* SetAllowDependentLinks */
/**
 * \brief turns the ability to manage identical (dependent) links ON or OFF
 *
 * When set to TRUE, attempts to establish (multiply) identical links to the same
 * target are allowed.  In this case the original link is desigated the 'parent link'
 * and constitutes the actual link (client to server).  Other client-side links
 * to the same target (have the same contract) will then obtain their results from
 * the parent link.
 *
 * \note A contract is deemed 'identical' if the device name and property parameters are
 * identical and the input and output data objects have identical characteristics
 * (including the data tag, format, and size).  The link acquisition mode
 * will be adjusted to the proper hierarchy level according to link dependencies.
 * Likewise, the link polling rate will also be adjusted according to the link
 * dependencies.
 *
 * \note The default setting is 'TRUE'.  Under some circumstances it could be
 * desireable to set the value to FALSE (e.g. the central archiver sets this to FALSE).
 *
 * \param value is the desired setting. (default = TRUE).
 *
 * \see GetAllowDependentLinks()
 */
TINE_EXPORT void SetAllowDependentLinks(int value);

/* GetAllowDependentLinks */
/**
 * \brief returns the setting for this value
 *
 * When set to TRUE, attempts to establish (multiply) identical links to the same
 * target are allowed.  In this case the original link is desigated the 'parent link'
 * and constitutes the actual link (client to server).  Other client-side links
 * to the same target (have the same contract) will then obtain their results from
 * the parent link.
 *
 *
 * \return the current setting of this feature. (default: TRUE)
 *
 * \see GetAllowDependentLinks()
 */
TINE_EXPORT int GetAllowDependentLinks(void);

/* SetSuspendCallbacks */
/**
 * \brief set suspends asynchronous callback notification ON or OFF
 *
 * When set to TRUE, all asynchronous callbacks will be suspended
 * until the call is reissued changing the setting to FALSE.
 *
 * \param value is the desired setting
 *
 * \see GetSuspendCallbacks()
 */
TINE_EXPORT void SetSuspendCallbacks(int value);

/* GetSuspendCallbacks */
/**
 * \brief returns the suspends asynchronous callback notification setting
 *
 * When set to TRUE, all asynchronous callbacks will be suspended
 * until the call is reissued changing the setting to FALSE.
 *
 * \return the current setting of this feature. (default: FALSE)
 *
 * \see SetSuspendCallbacks()
 */
TINE_EXPORT int GetSuspendCallbacks(void);

/* ResetSuspendCallbacks */
/**
 * \brief stops suspending callbacks and resets the reference counter
 *
 * When set to TRUE, all asynchronous callbacks will be suspended
 * until the call is reissued changing the setting to FALSE.
 *
 * \see SetSuspendCallbacks()
 */
TINE_EXPORT void ResetSuspendCallbacks(void);

extern int ttimeprobe;
TINE_EXPORT void SystemTimeProbe(const char *tag);
#define ProbeTimeElapsed(t) if (ttimeprobe) SystemTimeProbe(t)

TINE_EXPORT int GetErrorListSize(void);
TINE_EXPORT int GetLogLevelListSize(void);

TINE_EXPORT int LinkDataMemcpy(void *dst,void *src,size_t sizeInBytes);
TINE_EXPORT int LockLinkData(void);
TINE_EXPORT int UnlockLinkData(void);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* TINETYPE_H_ */
